/* 
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Id: espda_interp.c 234066 2018-12-19 14:55:18Z gcupani $
 * $Author: gcupani $
 * $Date: 2018-12-19 14:55:18 +0000 (Wed, 19 Dec 2018) $
 * $Revision: 234066 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include "espda_interp.h"

/*----------------------------------------------------------------------------*/
/**
  @defgroup  espda_interp  Interpolation

  Linear and bi-linear interpolation is provided, for fast estimation of 1-d and
  2-d function from tabulated values.
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Perform bi-linear interpolation.
  @param    x_a   1st abscissa.
  @param    x_b   2nd abscissa.
  @param    y_a   1st ordinate.
  @param    y_b   2nd ordinate.
  @param    v_aa  The image of (@a x_a, @a y_a).
  @param    v_ba  The image of (@a x_b, @a y_a).
  @param    v_ab  The image of (@a x_a, @a y_b).
  @param    v_bb  The image of (@a x_b, @a y_b).
  @param    x_0   The interpolation abscissa.
  @param    y_0   The interpolation ordinate.
  @return   The image of (@a x_0, @a y_0)
   
  This function computes the bi-linear interpolation across four points (@a x_a,
  @a y_a, @a v_aa), (@a x_b, @a y_a, @a v_ba), (@a x_a, @a y_b, @a v_ab), and
  (@a x_b, @a y_b, @a v_bb) in a 3-d space, given the interpolation position 
  (@a x_0, @a y_0). It is meant to be used to compute the value of a tabulated 
  2-d function at an arbitrary position.

  The bi-linear interpolation formula is @f[
    v_0 = \frac{v_{aa}(x_b - x_0)(y_b - y_0) +
                v_{ba}(x_0 - x_a)(y_b - y_0) +
                v_{ab}(x_b - x_0)(y_0 - y_a) +
                v_{bb}(x_0 - x_a)(y_0 - y_a)}
                     {(x_b - x_a)(y_b - y_a)}.
  @f]  
  It reduces to a linear interpolation when @a x_a = @a x_b or @a y_a = @a y_b.
 */
/*----------------------------------------------------------------------------*/

double espda_interp_bilin(int x_a, int x_b, int y_a, int y_b, double v_aa, 
    double v_ba, double v_ab, double v_bb, double x_0, double y_0) {
    //extern int v_level;
    double v_0;
    if (x_b == x_a) {
        v_0 = v_aa + (v_ab - v_aa) * (y_0 - y_a) / (y_b - y_a);
    }
    else if (y_b == y_a) {
        v_0 = v_aa + (v_ba - v_aa) * (x_0 - x_a) / (x_b - x_a);
    }
    else {
        v_0 = 1 / (x_b - x_a) * 1 / (y_b - y_a) *
           (v_aa * (x_b - x_0) * (y_b - y_0) +
            v_ba * (x_0 - x_a) * (y_b - y_0) +
            v_ab * (x_b - x_0) * (y_0 - y_a) +
            v_bb * (x_0 - x_a) * (y_0 - y_a)); 
    }
    return v_0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Resample the instrument profile across a spectrum
  @param    spec       Spectrum
  @param    inst_prof  Instrument profile
  @param    size       Size of the instrument profile
  @param    resol      Resolution of the instrument profile
  @param    kappa      Number of standard deviations used to truncate the 
                       instrument profile.
  @param    prof       Table with the interpolated instrument profile
  @return   CPL_ERROR_NONE iff OK
   
  This function resamples the instrument profile at the wavelengths of a 
  spectrum. At each wavelength of the spectrum, the instrument profile is 
  linearly interpolated across the range of wavelengths covered by the size of 
  the profile. The results are saved as an array column in table with the same
  wavelengths of the input spectrum. 

  NB: This function is still provisional. For the time being it handles only a 
      gaussian instrument profile defined with the function espda_model_inst 
      (parameters size, resol, and kappa are the same used for the latter 
      function).
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_interp_inst(cpl_table  *spec,
                                 cpl_table  *inst_prof,
                                 double      size,
                                 double      resol,
                                 double      kappa,
                                 cpl_table **prof)

{
    extern int v_level;
    cpl_errorstate prestate = cpl_errorstate_get();

    if (spec != NULL) {
        cpl_size   spec_size = cpl_table_get_nrow(spec);
        cpl_size   prof_size = 0;
        cpl_array *wavel_min = cpl_array_new(spec_size, CPL_TYPE_DOUBLE);
        cpl_array *wavel_max = cpl_array_new(spec_size, CPL_TYPE_DOUBLE);
        cpl_array *prof_min  = cpl_array_new(spec_size, CPL_TYPE_INT);
        cpl_array *prof_max  = cpl_array_new(spec_size, CPL_TYPE_INT);

	//espda_struct_sort(spec, 1, "WAVEL", FALSE);

/*
 * Instrument profile is resampled on the spectrum wavelengths. The output table
 * contains the following columns:
 * 
 *     WAVEL:     wavelengths, same as in the input spectrum;
 *     WAVELMIN:  minimum wavelength of the resampled instrument profile;
 *     WAVELMAX:  maximum wavelength of the resampled instrument profile;
 *     PROFMIN:   row number of the input spectrum at WAVELMIN;
 *     PROFMAX:   row number of the input spectrum at WAVELMAX;
 *     PROFNORM:  array with the resampled instrument profile, with size
 *                PROFMAX - PROFMIN.
 */

        // Loop over wavelengths
	int ii_temp = 0;
        int prof_min_i = 0, prof_max_i = 0;
        for (int i = 0; i < (int)spec_size; i++) { 
            if ((i + 1) % (int)ceil(spec_size/10.0) == 0 && i < spec_size) {
                espda_print_verb(1, cpl_msg_info(cpl_func, " %-*i row scanned "
                    "in the input spectrum; %-*i left.", 
                    (int)ceil(log10(spec_size)), i + 1, 
                    (int)ceil(log10(spec_size)), (int)spec_size - i - 1));
            }

            // Select the wavelenght range for resampling
            double wavel_cen_i = cpl_table_get_double(spec, "WAVEL", i, NULL);
            double wavel_min_i = wavel_cen_i * (1 + cpl_table_get_double(inst_prof, 
                                 "VELRED", 0,        NULL));
            double wavel_max_i = wavel_cen_i * (1 + cpl_table_get_double(inst_prof, 
                                 "VELRED", size - 1, NULL));
/*
            cpl_table_unselect_all(spec);
            cpl_table_or_selected_double(spec, "WAVEL", CPL_LESS_THAN, wavel_min_i);
            int prof_min_i_old = (int)cpl_table_count_selected(spec);
            cpl_table_unselect_all(spec);
            cpl_table_or_selected_double(spec, "WAVEL", CPL_NOT_GREATER_THAN, 
                                         wavel_max_i);
            int prof_max_i_old = (int)cpl_table_count_selected(spec);
*/

	    for (int ii = ii_temp; ii < (int)spec_size; ii++) {
	        if (cpl_table_get_double(spec, "WAVEL", ii, NULL) < wavel_min_i) {
                    prof_min_i = ii + 1;
                    ii_temp++;
                }
	        if (cpl_table_get_double(spec, "WAVEL", ii, NULL) < wavel_max_i) {
                    prof_max_i = ii + 1;
                }
		else {
		    break;
		}
            }
/*
	    if (prof_min_i == prof_min_i_old || prof_max_i == prof_max_i_old) {
	    printf("%i %i\n", prof_min_i, prof_min_i_old);
	    printf("%i %i\n", prof_max_i, prof_max_i_old);
	    }
*/
//	    int prof_min_i = prof_min_i_new;
//	    int prof_max_i = prof_max_i_new;

            // Define arrays for the output table
            cpl_array_set_double(wavel_min, i, wavel_min_i); 
            cpl_array_set_double(wavel_max, i, wavel_max_i); 
            cpl_array_set_int   (prof_min,  i, prof_min_i);  
            cpl_array_set_int   (prof_max,  i, prof_max_i);
            cpl_size temp_size = prof_max_i - prof_min_i;
            if (temp_size > prof_size) {
                prof_size = temp_size;
            }
        }

        // Create output table
        *prof = cpl_table_new(spec_size);
        cpl_table_new_column(*prof, "WAVEL",    CPL_TYPE_DOUBLE);
        cpl_table_new_column(*prof, "WAVELMIN", CPL_TYPE_DOUBLE);
        cpl_table_new_column(*prof, "WAVELMAX", CPL_TYPE_DOUBLE);
        cpl_table_new_column(*prof, "PROFMIN",  CPL_TYPE_INT);
        cpl_table_new_column(*prof, "PROFMAX",  CPL_TYPE_INT);
        cpl_table_copy_data_double(*prof, "WAVELMIN", 
                                  cpl_array_get_data_double(wavel_min));
        cpl_table_copy_data_double(*prof, "WAVELMAX",
                                  cpl_array_get_data_double(wavel_max));
        cpl_table_copy_data_int   (*prof, "PROFMIN",
                                  cpl_array_get_data_int   (prof_min));
        cpl_table_copy_data_int   (*prof, "PROFMAX",
                                  cpl_array_get_data_int   (prof_max));
        cpl_table_new_column_array(*prof, "PROFNORM", CPL_TYPE_DOUBLE, 
                                   prof_size);
        cpl_table_copy_data_double(*prof, "WAVEL",
                                   cpl_table_get_data_double(spec, "WAVEL"));

	//cpl_array_dump(prof_min, 100000, 10, NULL);
	//cpl_array_dump(prof_max, 100000, 10, NULL);

        // Loop over wavelengths
        for (int i = 0; i < (int)spec_size; i++) {
            if ((i + 1) % (int)ceil(spec_size/10.0) == 0 && i < spec_size) {
                espda_print_verb(1, cpl_msg_info(cpl_func, " %-*i rows filled "
                    " in the output profile; %-*i left.", 
                    (int)ceil(log10(spec_size)), i + 1, 
                    (int)ceil(log10(spec_size)), (int)spec_size - i - 1));
            }

 	    //if (i < 1200 == 0) printf("%i of %i\n", i, spec_size);
            int ii_min           = cpl_array_get_int(prof_min, i, NULL);
            int ii_max           = cpl_array_get_int(prof_max, i, NULL);
            cpl_array *prof_i    = cpl_array_new(prof_size, CPL_TYPE_DOUBLE);
            double     norm_fact = 0;

	    //printf("%i %i %i\n", i, ii_min, ii_max);
            // Resample the instrument profile with linear interpolation
            for (int ii = ii_min; ii < ii_max; ii++) {
                double ii_0 = (cpl_table_get_double(*prof, "WAVEL", ii, NULL) /
                               cpl_table_get_double(*prof, "WAVEL", i,  NULL) -
                               1.0) * 
                               size * resol / kappa + 0.5 * (size - 1.0);
                double ii_1 = (double)((int)(ii_0 * 1e6) * 1e-6);
                int    ii_a = (int)floor(ii_1);
                int    ii_b = (int)ceil (ii_1); 
	        if (ii_a >= 0 && ii_b >= 0) {
                double prof_a = cpl_table_get_double(inst_prof, "PROF", ii_a, 
		    NULL);
                double prof_b = cpl_table_get_double(inst_prof, "PROF", ii_b, 
                    NULL);
		double prof_0;
                prof_0 = espda_interp_lin(ii_a, ii_b, prof_a, prof_b, ii_1);
		cpl_array_set_double(prof_i, ii - ii_min, prof_0);
		norm_fact += prof_0;
                }
            }

            cpl_array *norm_i = cpl_array_duplicate(prof_i);
            cpl_array_delete(prof_i);
//printf("%i %i\n", cpl_array_get_size(norm_i), cpl_table_get_column_depth(*prof, "PROFNORM"));
            if (norm_fact != 0) {
	        cpl_array_divide_scalar(norm_i, norm_fact);
		cpl_table_set_array (*prof, "PROFNORM", i, norm_i);
	    }



            cpl_array_delete(norm_i);
        }
        cpl_array_delete(wavel_min);
        cpl_array_delete(wavel_max);
        cpl_array_delete(prof_min);
        cpl_array_delete(prof_max);
    }    
    if (!cpl_errorstate_is_equal(prestate)) {
        return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to interpolate the instrument profile.");
    }

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Perform linear interpolation.
  @param    x_a  1st abscissa.
  @param    x_b  2nd abscissa.
  @param    v_a  The image of @a x_a.
  @param    v_b  The image of @a x_b.
  @param    x_0  The interpolation abscissa.
  @return   The image of @a x_0.
   
  This function computes the linear interpolation across two points (@a x_a, 
  @a v_a) and (@a x_b, @a v_b) in a 2-d space, given the interpolation position 
  @a x_0. It is meant to be used to compute the value of a tabulated 1-d 
  function at an arbitrary position.

  The linear interpolation formula is @f[
    v_0 = v_a + (v_b - v_a) \frac{x_0 - x_a}{x_b - x_a}.
  @f]  
  It reduces to @a v_0 = @a v_a = @a v_b when @a x_a = @a x_b.
 */
/*----------------------------------------------------------------------------*/

double espda_interp_lin(int x_a, int x_b, double v_a, double v_b, double x_0) {
    extern int v_level;
    double v_0;
    if (x_b == x_a || x_b - x_a == 0) {
        v_0 = v_a;
    }
    else {
        v_0 = v_a + (v_b - v_a) * (x_0 - x_a) / (x_b - x_a);
    }
    return v_0;
}

/**@}*/

