 /* $Id: espda_syst.c 270655 2022-02-25 09:12:45Z amodigli $
 * $Name:  $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include "espda_syst.h"

/*----------------------------------------------------------------------------*/
/**
  @defgroup  espda_syst  Lists of absorption systems
  
  A list of absorption systems is a special list of lines (LINE_SYST) produced
  from a list of redshift by applying specific rules for existence or 
  acceptance. It contains lines associated with absorption systems (@a SYSTID
  column is filled).
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief   Apply acceptance rules.
  @param   syst       A list of absorption systems (LINE_SYST).
  @param   syst_id    The ID of the selected system.
  @param   row_start  The starting row of the selected system.
  @param   row_num    Number of rows of the selected system.
  @param   disc_flag  Flag to notify that lines have been discarded.
  @return  CPL error code.
  
  This function evaluates an absorption system from a list @a syst against a set
  of acceptance rules and discards all lines that do not comply. The system is 
  defined by its ID @a syst_id and by the starting row @a row_start and number 
  of rows @a row_num in @a syst (as obtained when @a syst is sorted by ascending 
  @a SYSTID). If lines are discarded, the flag @a disc_flag is set to 1.
  
  If @a syst is NULL, an error is issued. If lines are discarded and the 
  verbosity level is greater than 0, a warning is issued.
  
  N.B. Acceptance rules have not been implemented yet.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_syst_accept(cpl_table *syst, char *syst_id,
    int *row_start, int *row_num, int *disc_flag) {
    extern int v_level;
    espda_util_error(syst == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    espda_struct_sort(syst, 1, "SYSTID", FALSE);
    *disc_flag = 0;
    char msg[LEN_STRING];
    for (int m = *row_start; m < *row_start + *row_num; m++) {
        //double line_z = cpl_table_get_double(syst, "LINEZ",  m, NULL);
        //const char *line_id = cpl_table_get_string(syst, "LINEID", m);
        int accept_flag = 1;
        
        // Provisional instructions to artificially discard one line
        /*
        if (!strcmp(syst_id, "c") && 
            !strcmp(cpl_table_get_string(syst, "LINEID", m), "OI_1302")) {
            accept_flag = 0;
            sprintf(msg, "Artificially discarded.");
        }
        */
        
        // Real rules to discard lines (TBC)

        // Lines are discarded
        if (accept_flag == 0 && 
            strcmp(cpl_table_get_string(syst, "SYSTID", m), INV_STRING)) {
            espda_print_verb(1, cpl_msg_warning(cpl_func, " Line %i from "
						"candidate %s was not "
						"accepted! %s",
						(int)(m - *row_start), syst_id, 
						msg));
            cpl_ensure_code(espda_syst_update(syst, NULL, NULL, NULL, 0.0, 
                "remove", m, 1) == CPL_ERROR_NONE, cpl_error_get_code());
            *disc_flag = 1;
        }
    }
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Add lines to absorption systems.
  @param    syst      A list of absorption systems (LINE_SYST).
  @param    noid      A list of non-identified lines (LINE_SYST).
  @param    ion       A list of ionic transitions (ION).
  @param    syst_num  Number of systems.
  @param    syst_dv    The half-width of the window to define coincidences.
  @return   CPL error code.
  
  This function scans a list of non-identified lines @a noid, computes the 
  possible redshifts of the lines by comparison with a list of ionic transitions
  @a ion, and find coincidences between these redshifts and those of a list 
  @a syst of @a syst_num absorption systems, within a window of half-width 
  @a syst_dv. The lines with coincident redshifts are added to @a syst and 
  removed from @a noid.
  
  The redshift @f$ z @f$ of a line is said to be coincident with the redshift 
  @f$ \bar{z} @f$ of a sistem if @f[
    \left|z-\bar{z}\right|<\frac{\mathit{add\_width}}{\lambda_\mathrm{trans}},
  @f]
  where @f$ \lambda_\mathrm{trans} @f$ is the wavelength of the transition 
  associated to the line (from column LINEID).

  If @a syst, @a noid, or @a ion are NULL, an error is issued.
 */   
/*----------------------------------------------------------------------------*/

cpl_error_code espda_syst_add(cpl_table *syst, cpl_table *line_pre, 
    cpl_table *ion, int syst_num, double syst_dv, double z_em) {
    //cpl_errorstate prestate = cpl_errorstate_get();
    extern int v_level;
    extern char syst_id1, syst_id2;
    espda_util_error(syst == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    espda_util_error(line_pre == NULL, CPL_ERROR_NULL_INPUT, 
        cpl_error_get_code());
    espda_util_error(ion == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    cpl_table *ion_trim = espda_redsh_trim(ion, z_em);  // Probably not needed
    cpl_table *noid = cpl_table_duplicate(line_pre);
    cpl_size noid_size = cpl_table_get_nrow(noid);
    cpl_size ion_size  = cpl_table_get_nrow(ion_trim);
    espda_print_verb(1, cpl_msg_info(cpl_func, "Checking redshift "
        "coincidences..."));
    espda_struct_sort(syst, 1, "SYSTID", FALSE);
    espda_syst_init();
    for (int n = 0; n < syst_num; n++) {
        cpl_table_fill_column_window_string(noid, "SYSTID", 0, noid_size,
            INV_STRING);
        char syst_id[4];
        espda_syst_adv(syst_id);
        int row_start, row_num;
        espda_syst_find(syst, syst_id, &row_start, &row_num);
        double z_ave, null_errave, null_rms;
        espda_util_func(espda_spec_ave(syst, "LINEZ", NULL, NULL, row_start, 
            row_start + row_num, &z_ave, &null_errave, &null_rms), SPEC_AVE, 0);
        int add_count = 0;
        for (int j = 0; j < ion_size; j++) {
	  double ion_wavel = cpl_table_get_double(ion_trim, "WAVEL",  j, NULL);
            const char *lineid = cpl_table_get_string(ion_trim, "LINEID", j);
            const char *multid = cpl_table_get_string(ion_trim, "MULTID", j);
            for (int i = 0; i < noid_size; i++) {
	        double wavel = cpl_table_get_double(noid, "WAVEL", i, NULL);
                double linez = wavel/ion_wavel - 1;
                double check = fabs(linez-z_ave) * ion_wavel/wavel * LIGHTSPEED;

		// Coincidence is found
                if (check < syst_dv) {
		printf("%f %f %f %f %f\n", check, linez, z_ave, ion_wavel, syst_dv);
                    espda_syst_update(noid, syst_id, lineid, multid, linez, 
                        "add", i, 1);
                    add_count++;
                }
            }
        }
        if (add_count > 0) {
            espda_print_verb(1, cpl_msg_info(cpl_func, " %i lines added to "
                "system %s.", add_count, syst_id));
        }
        cpl_table *newid = espda_struct_extract(noid, 1, "SYSTID", 
            CPL_NOT_EQUAL_TO, INV_STRING);
        if (cpl_table_get_nrow(newid) > 0) {
            cpl_table_insert(syst, newid, 0);
        }
        cpl_table_delete(newid);
    }
    espda_syst_init();
    cpl_table_delete(ion_trim);
    
    // Redshift values exceeding the emission redshift are removed
    /*
    cpl_table_unselect_all(noid);
    cpl_table_or_selected_double(noid, "LINEZ", CPL_GREATER_THAN, zem);
    cpl_table_erase_selected(noid);
    */
    /*
    if (!cpl_errorstate_is_equal(prestate)) {
        cpl_error_set_message(cpl_func, cpl_error_get_code(), "Unable to "
            "add lines values to a system.");
    }
    */
    cpl_table_delete(noid);
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Adjust the number of components in multiplets.
  @param   syst       A list of absorption systems.
  @param   syst_num   Number of systems.
  @return  CPL error code.
  
  This functions scans a list @a syst of @a syst_num absorption systems, detects
  multiplet transitions and adjusts the number of components for all multiplet 
  members (i.e., lines are added until all the multiplet members have as many 
  components as the richest member). The multiplets currently considered are 
  NV @f$ \lambda\lambda @f$1238-1242 (doublet), SiII 
  @f$ \lambda\lambda @f$1260-1526 (triplet), SiIV 
  @f$ \lambda\lambda @f$1393-1402 (doublet), CIV @f$ \lambda\lambda @f$1548-1550
  (doublet), FeII @f$ \lambda\lambda @f$2344-2600 (quadruplet), MgII 
  @f$ \lambda\lambda @f$2796-2803 (doublet), 
  CaII @f$ \lambda\lambda @f$3934-3969 (doublet), and NaI 
  @f$ \lambda\lambda @f$5891-5897 (doublet).
  
  If @a syst is NULL, an error is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_syst_adjust(cpl_table *syst, cpl_table *ion,
				 cpl_size syst_num, double mult_dz) {
    //cpl_errorstate prestate = cpl_errorstate_get();
    extern int v_level;
    extern char syst_id1, syst_id2;
    espda_struct_sort_old(syst, "WAVEL", FALSE);
    espda_syst_init();
    
    cpl_table *prova = NULL;
    for (int n = 0; n < syst_num; n++) {
      //for (int n = 0; n < 1; n++) {
        
        // The number of component of each transition is determined
        char syst_id[4];
        espda_syst_adv(syst_id);
	/*	espda_syst_adv(syst_id);
        espda_syst_adv(syst_id);
        espda_syst_adv(syst_id);
        espda_syst_adv(syst_id);
	printf("%s\n", syst_id);
	*/
        // CPL_EQUAL_TO cannot be used, otherwise the string are treated as
        // regular expressions
        cpl_table *syst_temp = espda_struct_extract(syst, 1, "SYSTID", 
            CPL_NOT_LESS_THAN, syst_id);
        cpl_table *syst_sel = espda_struct_extract(syst_temp, 1, "SYSTID", 
            CPL_NOT_GREATER_THAN, syst_id);
        cpl_table_delete(syst_temp);
        cpl_size line_num = cpl_table_get_nrow(syst_sel);
        cpl_array *line_id_list  = cpl_array_new(line_num, CPL_TYPE_STRING);
        cpl_array *line_num_list = cpl_array_new(line_num, CPL_TYPE_INT);
        cpl_array *mult_id_list  = cpl_array_new(line_num, CPL_TYPE_STRING);
        cpl_array *mult_num_list = cpl_array_new(line_num, CPL_TYPE_INT);
        char *line_id_old = malloc(sizeof(char) * LEN_STRING);
        char *mult_id_old = malloc(sizeof(char) * LEN_STRING);
        strcpy(line_id_old, "");
        strcpy(mult_id_old, "");
        int l_list = 0;
        int m_list = 0;
        for (int l = 0; l < line_num; l++) {
            const char *line_id = cpl_table_get_string(syst_sel, "LINEID", l);
            if (strcmp(line_id, line_id_old)) {
                cpl_table *line_sel = espda_struct_extract(syst_sel, 1, 
                    "LINEID", CPL_EQUAL_TO, line_id);
                cpl_array_set_string(line_id_list,  l_list, line_id);
                cpl_array_set_int   (line_num_list, l_list, 
                    cpl_table_get_nrow(line_sel));
                cpl_table_delete(line_sel);
                strcpy(line_id_old, line_id);
                l_list++;
            } 
            const char *mult_id = cpl_table_get_string(syst_sel, "MULTID", l);
            if (strcmp(mult_id, mult_id_old)) {
                cpl_table *mult_sel = espda_struct_extract(syst_sel, 1, 
                    "MULTID", CPL_EQUAL_TO, mult_id);
                cpl_array_set_string(mult_id_list,  m_list, mult_id);
                cpl_array_set_int   (mult_num_list, m_list, 
                    cpl_table_get_nrow(mult_sel));
                cpl_table_delete(mult_sel);
                strcpy(mult_id_old, mult_id);
	        m_list++;
	    } 
        } 
	cpl_free(mult_id_old);

        if (l_list < line_num) {
            cpl_array_erase_window(line_id_list,  l_list, line_num - l_list);
            cpl_array_erase_window(line_num_list, l_list, line_num - l_list);
        }
        if (m_list < line_num) {
            cpl_array_erase_window(mult_id_list,  m_list, line_num - m_list);
            cpl_array_erase_window(mult_num_list, m_list, line_num - m_list);
        }
	/*
	printf("l_list = %i\n", l_list);
	cpl_array_dump(line_id_list, 0, 10, NULL);
	cpl_array_dump(line_num_list, 0, 10, NULL);
	printf("m_list = %i\n", m_list);
	cpl_array_dump(mult_id_list, 0, 10, NULL);
	cpl_array_dump(mult_num_list, 0, 10, NULL);
	*/
	//cpl_array_dump(mult_id_list, 0, 10, NULL);
        for (int m = 0; m < cpl_array_get_size(mult_id_list); m++) {

	    const char *multid = cpl_array_get_string(mult_id_list, m);
	    //cpl_table *multid_t = espda_struct_extract(syst, 1, "SYSTID", 
	    //					       CPL_EQUAL_TO, syst_id);
	    cpl_table *multid_sel = espda_struct_extract(syst_sel, 1, "MULTID",
							 CPL_EQUAL_TO, multid);

	    //cpl_table_dump(multid_sel, 0, 10, NULL);
	    cpl_size multid_num = cpl_table_get_nrow(multid_sel);
	    espda_struct_sort(multid_sel, 1, "LINEZ", FALSE);
	    //cpl_array *peak = cpl_array_new(multid_num, CPL_TYPE_DOUBLE);
	    cpl_array *linez = cpl_array_new(multid_num, CPL_TYPE_DOUBLE);
	    cpl_array *linez_min = cpl_array_new(multid_num, CPL_TYPE_DOUBLE);
	    cpl_array *linez_max = cpl_array_new(multid_num, CPL_TYPE_DOUBLE);
	    cpl_array *linez_diff = cpl_array_new(multid_num, CPL_TYPE_DOUBLE);

	    espda_struct_sort(syst, 1, "LINEZ", FALSE);
	    cpl_table *comp = cpl_table_duplicate(syst_sel);
	    cpl_table_erase_window(comp, 0, cpl_table_get_nrow(comp));

	    double linez_old = 0.0;
	    int z_num = 0;
	    for (int s = 0; s < multid_num; s++) {
	        //double peak_temp = cpl_table_get_double(multid_sel, "PEAK", s,
		//					NULL);
	        double linez_temp = cpl_table_get_double(multid_sel, "LINEZ", s,
							 NULL);
	        double linez_min_temp = cpl_table_get_double(multid_sel, 
                    "LINEZMIN", s, NULL);
	        double linez_max_temp = cpl_table_get_double(multid_sel, 
                    "LINEZMAX", s, NULL);

		
	        if (fabs(linez_temp - linez_old) > mult_dz) {
		  //cpl_array_set_double(peak, z_num, peak_temp);
		    cpl_array_set_double(linez, z_num, linez_temp);
		    cpl_array_set_double(linez_min, z_num, linez_min_temp);
		    cpl_array_set_double(linez_max, z_num, linez_max_temp);
		    cpl_array_set_double(linez_diff, z_num, 
			0.5 * (linez_max_temp - linez_min_temp));

		    cpl_table *sel = cpl_table_extract(multid_sel, s, 1);
		    cpl_table_insert(comp, sel, z_num);
		    
		    z_num++;
		    linez_old = linez_temp;
		}
	    }
	    if (z_num < multid_num) {
	        cpl_array_erase_window(linez, z_num, multid_num - z_num);
	        cpl_array_erase_window(linez_min, z_num, multid_num - z_num);
	        cpl_array_erase_window(linez_max, z_num, multid_num - z_num);
	        cpl_array_erase_window(linez_diff, z_num, multid_num - z_num);
	    }

	    //cpl_array_dump(linez_diff, 0, 10, NULL);
	    //cpl_table_dump(comp, 0, 10, NULL);

	    for (int l = 0; l < cpl_array_get_size(line_id_list); l++) {
	    //for (int l = 0; l < 1; l++) {
		char *same = strstr(cpl_array_get_string(line_id_list, 
                    l), cpl_array_get_string(mult_id_list, m));
		if (same != NULL) {
		    const char *lineid = cpl_array_get_string(line_id_list, l);
		    //cpl_table *lineid_t = espda_struct_extract(syst, 1, 
	   	    //    "SYSTID", CPL_EQUAL_TO, syst_id);
		    //cpl_table_dump(temp, 0, 100, NULL);
		    cpl_table *lineid_sel = espda_struct_extract(syst_sel, 1,
			"LINEID", CPL_EQUAL_TO, lineid);
		    //cpl_table_dump(lineid_sel, 0, 100, NULL);
		    cpl_table *ion_sel = espda_struct_extract(ion, 1, "LINEID",
			CPL_EQUAL_TO, lineid);
		    double ion_wavel = cpl_table_get_double(ion_sel, "WAVEL", 0,
							    NULL);
		    for (int s = 0; s < z_num; s++) {
		        double linez_temp = 0.0;  
			if (s < cpl_table_get_nrow(lineid_sel)) {
		            linez_temp = cpl_table_get_double(lineid_sel,
                               "LINEZ", s, NULL);
                        }
			double linez_ref = cpl_array_get_double(linez, s, NULL);
			//printf("%f %f %f\n", linez_temp, linez_ref, mult_dz);
			if (fabs(linez_temp - linez_ref) > mult_dz) {
			  //double peak_sel = cpl_array_get_double(peak, s, 
			  //				   NULL);
			    double linez_sel = cpl_array_get_double(linez, s, 
								    NULL);
			    double linez_min_sel = cpl_array_get_double(
				linez_min, s, NULL);
			    double linez_max_sel = cpl_array_get_double(
				linez_max, s, NULL);
			    double wavel_sel = (1.0 + linez_sel) * ion_wavel;
			    double wavel_min_sel = (1.0 + linez_min_sel) 
			        * ion_wavel;
			    //wavel_sel - 
			    //    cpl_array_get_double(linez_diff, s, NULL);
			    double wavel_max_sel = (1.0 + linez_max_sel) 
			        * ion_wavel;
			      //wavel_sel + 
			      //cpl_array_get_double(linez_diff, s, NULL);
			    cpl_table_insert_window(lineid_sel, s, 1);
			    //printf("%f %f %f\n", cpl_array_get_double(linez_diff, s, NULL), wavel_min_sel, wavel_max_sel);
			    /*
			    cpl_table_insert_window(syst, 0, 1);
			    cpl_table_set_double(syst, "PEAK", 0, peak_sel);
			    cpl_table_set_double(syst, "LINEZ", 0, linez_sel);
			    cpl_table_set_double(syst, "LINEZMAX", 0, 
						 linez_max_sel);
			    cpl_table_set_double(syst, "LINEZMIN", 0, 
						 linez_min_sel);
			    cpl_table_set_double(syst, "WAVEL", 0, wavel_sel);
			    cpl_table_set_double(syst, "WAVELMAX", 0, 
						 wavel_max_sel);
			    cpl_table_set_double(syst, "WAVELMIN", 0, 
						 wavel_min_sel);
			    cpl_table_set_string(syst, "LINEID", 0, lineid);
			    cpl_table_set_string(syst, "MULTID", 0, multid);
			    cpl_table_set_string(syst, "SYSTID", 0, 
				cpl_table_get_string(multid_sel, "SYSTID", 0));
				*/
			    cpl_table *add = cpl_table_extract(comp, s, 1);
			    cpl_table_set_double(add, "LINEZ", 0, linez_sel);
			    cpl_table_set_double(add, "LINEZMIN", 0, 
						 linez_min_sel);
			    cpl_table_set_double(add, "LINEZMAX", 0, 
						 linez_max_sel);
			    cpl_table_set_double(add, "WAVEL", 0, wavel_sel);
			    cpl_table_set_double(add, "WAVELMIN", 0, 
						 wavel_min_sel);
			    cpl_table_set_double(add, "WAVELMAX", 0, 
						 wavel_max_sel);
			    cpl_table_set_string(add, "LINEID", 0, lineid);
			    cpl_table_set_int(add, "GRPID", 0, INV_INT);
			    
			    //cpl_table_dump(add, 0, 100, NULL);
			    cpl_table_insert(syst, add, 0);
			    //cpl_table_dump(syst, 0, 10, NULL);
			    //printf("%i\n", cpl_table_get_nrow(syst));
			}
		    }
		}
	    }
	    espda_struct_sort(syst, 2, "LINEID", FALSE, "LINEZ", FALSE);
	}
	cpl_free(line_id_old);
    }
    prova = cpl_table_duplicate(syst);
    //cpl_table_erase_column(prova, "PEAK");
    cpl_table_erase_column(prova, "WAVELERR");
    //cpl_table_erase_column(prova, "WAVELMIN");
    //cpl_table_erase_column(prova, "WAVELMAX");
    //cpl_table_erase_column(prova, "LINEID");
    //cpl_table_erase_column(prova, "MULTID");
    //cpl_table_erase_column(prova, "GRPID");
    //cpl_table_erase_column(prova, "LINEZ");
    cpl_table_erase_column(prova, "LINEZERR");
    cpl_table_erase_column(prova, "LINEZMIN");
    cpl_table_erase_column(prova, "LINEZMAX");
    //cpl_table_erase_column(prova, "LINEZCON");
    //cpl_table_erase_column(prova, "COLDEN");
    cpl_table_erase_column(prova, "COLDENERR");
    cpl_table_erase_column(prova, "COLDENMIN");
    cpl_table_erase_column(prova, "COLDENMAX");
    cpl_table_erase_column(prova, "COLDENCON");
    //cpl_table_erase_column(prova, "THERB");
    cpl_table_erase_column(prova, "THERBERR");
    cpl_table_erase_column(prova, "THERBMIN");
    cpl_table_erase_column(prova, "THERBMAX");
    cpl_table_erase_column(prova, "THERBCON");
    cpl_table_erase_column(prova, "TURBB");
    cpl_table_erase_column(prova, "TURBBERR");
    cpl_table_erase_column(prova, "TURBBMIN");
    cpl_table_erase_column(prova, "TURBBMAX");
    cpl_table_erase_column(prova, "TURBBCON");
    cpl_table_erase_column(prova, "USE");
    cpl_table_erase_column(prova, "CONT");
    //cpl_table_erase_column(prova, "SYSTID");
    cpl_table_erase_column(prova, "ORDER");
    //cpl_table_dump(prova, 0, 1000, NULL);
    //if (1 == 1) break;
/*
            cpl_array *ref_multid = NULL;
            cpl_array *ref_linez = NULL;
            cpl_array *ref_linez_min = NULL;
            cpl_array *ref_linez_max = NULL;
	    cpl_array *ref_linez_red = NULL; 
            double ref_diff;
            for (int l = 0; l < l_list; l++) {
                char *found_mult = strstr(cpl_array_get_string(line_id_list, 
                    l), cpl_array_get_string(mult_id_list, m));
                cpl_size line_num_adj = cpl_array_get_int(line_num_list, l, 
                    NULL);
                cpl_size mult_num_adj = cpl_array_get_int(mult_num_list, m, 
                    NULL);
                if (found_mult != NULL && line_num_adj == mult_num_adj) {
                    cpl_table *ref_temp = espda_struct_extract(syst_sel, 1, 
                        "LINEID", CPL_NOT_LESS_THAN, 
                        cpl_array_get_string(line_id_list, l));
                    cpl_table *ref_sel = espda_struct_extract(ref_temp, 1, 
                        "LINEID", CPL_NOT_GREATER_THAN, 
                        cpl_array_get_string(line_id_list, l));
                    ref_multid = cpl_array_wrap_string(
                        cpl_table_get_data_string(ref_sel, "MULTID"), 
                        cpl_table_get_nrow(ref_sel));
                    ref_linez = cpl_array_wrap_double(
                        cpl_table_get_data_double(ref_sel, "LINEZ"), 
                        cpl_table_get_nrow(ref_sel));
                    ref_linez_min = cpl_array_wrap_double(
                        cpl_table_get_data_double(ref_sel, "LINEZMIN"), 
                        cpl_table_get_nrow(ref_sel));
                    ref_linez_max = cpl_array_wrap_double(
                        cpl_table_get_data_double(ref_sel, "LINEZMAX"), 
                        cpl_table_get_nrow(ref_sel));
                    ref_diff = (cpl_table_get_double(ref_sel, "WAVELMAX", 
                        0, NULL) - cpl_table_get_double(ref_sel, 
                        "WAVELMIN", 0, NULL)) * 0.5;
                }
	    }
	    if (m == 4 || m == 5) cpl_array_dump(ref_linez, 0, 100, NULL);

            for (int l = 0; l < l_list; l++) {
                char *found_mult = strstr(cpl_array_get_string(line_id_list, 
                    l), cpl_array_get_string(mult_id_list, m));
                cpl_size line_num_adj = cpl_array_get_int(line_num_list, l, 
                    NULL);
                cpl_size mult_num_adj = cpl_array_get_int(mult_num_list, m, 
                    NULL);

                // Lines are added to match the number of components
                if (found_mult != NULL && line_num_adj != mult_num_adj) {
                    int found_count = 0;

                    cpl_table_unselect_all(syst_sel);
                    cpl_table_or_selected_string(syst_sel, "LINEID", 
                        CPL_EQUAL_TO, cpl_array_get_string(line_id_list,
                        l));
                    //cpl_table *line_sel = 
                    //    cpl_table_extract_selected(syst_sel);
                    cpl_table *line_check = 
                        cpl_table_extract_selected(syst_sel);
                    //cpl_table *syst2 = 
                    //    cpl_table_extract_selected(syst);

		    // Find the rows in the table where lines are to be added
		    //cpl_table_new_column(line_check, "LINEZDIFF", 
		    //		     CPL_TYPE_DOUBLE);
		    //cpl_table_copy_data_double(line_check, "LINEZDIFF",
		    //    cpl_array_get_data_double(ref_linez_red));
		    //cpl_table_subtract_columns(line_check, "LINEZDIFF", 
		    //			   "LINEZ");
		    //cpl_table_abs_column(line_check, "LINEZDIFF");
		    //cpl_table_dump(line_check, 0, 1000, NULL);
		    for (int l_sel = 0; l_sel < cpl_array_get_size(ref_linez);
			 l_sel++) {
  		        double diff_abs = fabs(cpl_table_get_double(line_check, 
                            "LINEZ", l_sel, NULL) -
	                    cpl_array_get_double(ref_linez, l_sel, NULL)); 
			printf("%f %f %f", cpl_table_get_double(line_check,  "LINEZ", l_sel, NULL),
			       cpl_array_get_double(ref_linez, l_sel, NULL), diff_abs);
			if (diff_abs > 1e-4) {
			  printf(" *");
			    cpl_table_insert_window(line_check, l_sel, 1);
			    cpl_table_insert_window(syst, 0, 1);
			    cpl_table_set_double(syst, "LINEZ", 0, 
				cpl_array_get_double(ref_linez, l_sel, NULL));
			    cpl_table_set_string(syst, "LINEID", 0, 
                          	cpl_array_get_string(line_id_list, l));
			    cpl_table_set_string(syst, "SYSTID", 0, 
				cpl_table_get_string(line_check, "SYSTID", 0));
			}
			printf("\n");
		    }
		    //cpl_table_delete(line_check);
		    espda_struct_sort(syst, 2, "LINEID", FALSE, "LINEZ", FALSE);
		    //cpl_table_dump(syst, 0, 100, NULL);
		    //syst = cpl_table_duplicate(syst2);

		    
                    //for (int l_adj = 0; l_adj < mult_num_adj - line_num_adj;
                    //    l_adj++) {
                    //    found_count++;
                    //    cpl_table_unselect_all(syst_sel);
                    //    cpl_table_or_selected_string(syst_sel, "LINEID", 
                    //       CPL_EQUAL_TO, cpl_array_get_string(line_id_list,
                    //       l));
                    //    cpl_table *line_sel = 
                    //        cpl_table_extract_selected(syst_sel);

                    //    if (cpl_table_get_nrow(line_sel) > 0) {
                    //        cpl_table_and_selected_window(line_sel, 0, 1);
                    //        cpl_table *row = 
                    //            cpl_table_extract_selected(line_sel);
                    //        if (cpl_table_has_column(row, "ORDER")) {

			//        cpl_table_set_int(row, "ORDER", 0, 
                        //            (int)cpl_table_get_column_max(syst, 
                        //            "ORDER") + 1);
                        //    }
			//    cpl_table_set_invalid(row, "WAVEL", 0);
			//    cpl_table_set_invalid(row, "LINEZ", 0);
                        //    cpl_table_select_all(line_sel);
                        //    cpl_table_insert(syst, row, 
                        //        cpl_table_get_nrow(syst)); 
                        //    cpl_table_delete(line_sel);
                        //    cpl_table_delete(row);
                        //}
                    //}
                }
		} 
    prova = cpl_table_duplicate(syst);
    //    espda_struct_sort(prova, 1, "WAVEL", FALSE);
    //cpl_table_erase_column(prova, "PEAK");
    cpl_table_erase_column(prova, "WAVELERR");
    cpl_table_erase_column(prova, "WAVELMIN");
    cpl_table_erase_column(prova, "WAVELMAX");
    //cpl_table_erase_column(prova, "LINEID");
    cpl_table_erase_column(prova, "MULTID");
    cpl_table_erase_column(prova, "GRPID");
    //cpl_table_erase_column(prova, "LINEZ");
    cpl_table_erase_column(prova, "LINEZERR");
    cpl_table_erase_column(prova, "LINEZMIN");
    cpl_table_erase_column(prova, "LINEZMAX");
    //cpl_table_erase_column(prova, "LINEZCON");
    //cpl_table_erase_column(prova, "COLDEN");
    cpl_table_erase_column(prova, "COLDENERR");
    cpl_table_erase_column(prova, "COLDENMIN");
    cpl_table_erase_column(prova, "COLDENMAX");
    cpl_table_erase_column(prova, "COLDENCON");
    //cpl_table_erase_column(prova, "THERB");
    cpl_table_erase_column(prova, "THERBERR");
    cpl_table_erase_column(prova, "THERBMIN");
    cpl_table_erase_column(prova, "THERBMAX");
    cpl_table_erase_column(prova, "THERBCON");
    cpl_table_erase_column(prova, "TURBB");
    cpl_table_erase_column(prova, "TURBBERR");
    cpl_table_erase_column(prova, "TURBBMIN");
    cpl_table_erase_column(prova, "TURBBMAX");
    cpl_table_erase_column(prova, "TURBBCON");
    cpl_table_erase_column(prova, "USE");
    cpl_table_erase_column(prova, "CONT");
    //cpl_table_erase_column(prova, "SYSTID");
    cpl_table_erase_column(prova, "ORDER");
    if (m == 4 || m == 5)
    cpl_table_dump(prova, 0, 1000, NULL);

            // Redshift and wavelength are adjusted for all components
            for (int l = 0; l < l_list; l++) {
	        char *found_mult = strstr(cpl_array_get_string(line_id_list, 
                    l), cpl_array_get_string(mult_id_list, m));
                cpl_size line_num_adj = cpl_array_get_int(line_num_list, l, 
                    NULL);
                cpl_size mult_num_adj = cpl_array_get_int(mult_num_list, m, 
                    NULL);
                if (found_mult != NULL) {
                    
                    const char *ion_id = cpl_array_get_string(line_id_list, l); 
                    cpl_table *ion_temp = espda_struct_extract(ion, 1, 
                        "LINEID", CPL_NOT_LESS_THAN, ion_id);
                    cpl_table *ion_sel = espda_struct_extract(ion_temp, 1, 
                        "LINEID", CPL_NOT_GREATER_THAN, ion_id);
                    double ion_wavel = cpl_table_get_double(ion_sel, "WAVEL", 
                        0, NULL);
                    cpl_table *comp_temp_1 = espda_struct_extract(syst, 1, 
                        "SYSTID", CPL_NOT_LESS_THAN, syst_id);
                    cpl_table *comp_temp_2 = espda_struct_extract(comp_temp_1, 
                        1, "SYSTID", CPL_NOT_GREATER_THAN, syst_id);
                    cpl_table *comp_temp_3 = espda_struct_extract(comp_temp_2, 
                        1, "LINEID", CPL_NOT_LESS_THAN, 
                        cpl_array_get_string(line_id_list, l));
                    cpl_table *comp_sel = espda_struct_extract(comp_temp_3, 1,
                        "LINEID", CPL_NOT_GREATER_THAN, 
                        cpl_array_get_string(line_id_list, l));
    prova = cpl_table_duplicate(comp_sel);
    //espda_struct_sort(prova, 1, "WAVEL", FALSE);
    //cpl_table_erase_column(prova, "PEAK");
    cpl_table_erase_column(prova, "WAVELERR");
    cpl_table_erase_column(prova, "WAVELMIN");
    cpl_table_erase_column(prova, "WAVELMAX");
    cpl_table_erase_column(prova, "LINEID");
    cpl_table_erase_column(prova, "MULTID");
    cpl_table_erase_column(prova, "GRPID");
    //cpl_table_erase_column(prova, "LINEZ");
    cpl_table_erase_column(prova, "LINEZERR");
    cpl_table_erase_column(prova, "LINEZMIN");
    cpl_table_erase_column(prova, "LINEZMAX");
    cpl_table_erase_column(prova, "LINEZCON");
    //cpl_table_erase_column(prova, "COLDEN");
    cpl_table_erase_column(prova, "COLDENERR");
    cpl_table_erase_column(prova, "COLDENMIN");
    cpl_table_erase_column(prova, "COLDENMAX");
    cpl_table_erase_column(prova, "COLDENCON");
    //cpl_table_erase_column(prova, "THERB");
    cpl_table_erase_column(prova, "THERBERR");
    cpl_table_erase_column(prova, "THERBMIN");
    cpl_table_erase_column(prova, "THERBMAX");
    cpl_table_erase_column(prova, "THERBCON");
    cpl_table_erase_column(prova, "TURBB");
    cpl_table_erase_column(prova, "TURBBERR");
    cpl_table_erase_column(prova, "TURBBMIN");
    cpl_table_erase_column(prova, "TURBBMAX");
    cpl_table_erase_column(prova, "TURBBCON");
    cpl_table_erase_column(prova, "USE");
    cpl_table_erase_column(prova, "CONT");
    cpl_table_erase_column(prova, "SYSTID");
    cpl_table_erase_column(prova, "ORDER");
    if (m == 5)
      //cpl_table_dump(prova, 0, 1000, NULL);
//cpl_table_dump(comp_sel, 0, 100, NULL);
// Sometimes there are problems here, investigate 
//printf("%i %i\n", cpl_array_get_size(ref_linez), cpl_table_get_nrow(comp_sel));
                    for (int ll = 0; ll < cpl_table_get_nrow(comp_sel); ll++) {
//	        printf("   %i\n", ll);
//cpl_array_dump(ref_linez, 0, 100, NULL);
                        double linez = cpl_array_get_double(ref_linez, ll, 
                            NULL);
                        double linez_min = cpl_array_get_double(ref_linez_min, 
                            ll, NULL);
                        double linez_max = cpl_array_get_double(ref_linez_max, 
                            ll, NULL);
                        double wavel = (1.0 + linez) * ion_wavel;
		        double wavel_min = wavel - ref_diff;
                        double wavel_max = wavel + ref_diff;
			
			//double peak = 
			

//    if (!cpl_errorstate_is_equal(prestate)) {
//        return cpl_error_set_message(cpl_func, cpl_error_get_code(), "Unable to "
//            "adjust number of components in multiplets.");
//    }

//                        if (ll >= cpl_array_get_size(ref_linez_min))
                        //printf("%s %s %i %i %f\n", syst_id, cpl_array_get_string(line_id_list, l), ll, cpl_array_get_size(ref_linez_min), linez_min);
                        cpl_table_set_double(comp_sel, "LINEZ", ll, linez);
                        cpl_table_set_double(comp_sel, "LINEZMIN", ll, 
                            linez_min);
                        cpl_table_set_double(comp_sel, "LINEZMAX", ll, 
                            linez_max);
                        cpl_table_set_double(comp_sel, "WAVEL", ll, wavel);
                        cpl_table_set_double(comp_sel, "WAVELMAX", ll, 
                            wavel_max);
                        cpl_table_set_double(comp_sel, "WAVELMIN", ll, 
                            wavel_min);
                    }
    prova = cpl_table_duplicate(comp_sel);
    espda_struct_sort(prova, 1, "WAVEL", FALSE);
    //cpl_table_erase_column(prova, "PEAK");
    cpl_table_erase_column(prova, "WAVELERR");
    cpl_table_erase_column(prova, "WAVELMIN");
    cpl_table_erase_column(prova, "WAVELMAX");
    cpl_table_erase_column(prova, "LINEID");
    cpl_table_erase_column(prova, "MULTID");
    cpl_table_erase_column(prova, "GRPID");
    //cpl_table_erase_column(prova, "LINEZ");
    cpl_table_erase_column(prova, "LINEZERR");
    cpl_table_erase_column(prova, "LINEZMIN");
    cpl_table_erase_column(prova, "LINEZMAX");
    cpl_table_erase_column(prova, "LINEZCON");
    //cpl_table_erase_column(prova, "COLDEN");
    cpl_table_erase_column(prova, "COLDENERR");
    cpl_table_erase_column(prova, "COLDENMIN");
    cpl_table_erase_column(prova, "COLDENMAX");
    cpl_table_erase_column(prova, "COLDENCON");
    //cpl_table_erase_column(prova, "THERB");
    cpl_table_erase_column(prova, "THERBERR");
    cpl_table_erase_column(prova, "THERBMIN");
    cpl_table_erase_column(prova, "THERBMAX");
    cpl_table_erase_column(prova, "THERBCON");
    cpl_table_erase_column(prova, "TURBB");
    cpl_table_erase_column(prova, "TURBBERR");
    cpl_table_erase_column(prova, "TURBBMIN");
    cpl_table_erase_column(prova, "TURBBMAX");
    cpl_table_erase_column(prova, "TURBBCON");
    cpl_table_erase_column(prova, "USE");
    cpl_table_erase_column(prova, "CONT");
    cpl_table_erase_column(prova, "SYSTID");
    cpl_table_erase_column(prova, "ORDER");
    if (m == 5)
      //    cpl_table_dump(prova, 0, 1000, NULL);
		    //cpl_table_dump(comp_sel, 0, 100, NULL);
                    for (int s = 0; s < cpl_table_get_nrow(syst); s++) {
                        if (!strcmp(cpl_table_get_string(syst, "SYSTID", s), 
                            syst_id) && !strcmp(cpl_table_get_string(syst, 
                            "LINEID", s), cpl_array_get_string(line_id_list, 
                            l))) {
                            cpl_table_set_invalid(syst, "WAVEL", s);
                        }
                    }
                    if (cpl_table_get_nrow(syst) - 
                        cpl_table_count_invalid(syst, "WAVEL") > 1) {
                        cpl_table_erase_invalid(syst);
                        cpl_table_insert(syst, comp_sel, 
                        cpl_table_get_nrow(syst));
                    }
                    else {
                        cpl_table_insert(syst, comp_sel, 
                        cpl_table_get_nrow(syst));
                        cpl_table_erase_invalid(syst);                        
                    }
                }
            }
        }
        
        cpl_table_delete(syst_sel);
        cpl_array_delete(line_id_list);
        cpl_array_delete(line_num_list);
        cpl_array_delete(mult_id_list);
        cpl_array_delete(mult_num_list);
    } 
    prova = cpl_table_duplicate(syst);
    espda_struct_sort(prova, 1, "WAVEL", FALSE);
    //cpl_table_erase_column(prova, "PEAK");
    cpl_table_erase_column(prova, "WAVELERR");
    cpl_table_erase_column(prova, "WAVELMIN");
    cpl_table_erase_column(prova, "WAVELMAX");
    cpl_table_erase_column(prova, "LINEID");
    cpl_table_erase_column(prova, "MULTID");
    cpl_table_erase_column(prova, "GRPID");
    //cpl_table_erase_column(prova, "LINEZ");
    cpl_table_erase_column(prova, "LINEZERR");
    cpl_table_erase_column(prova, "LINEZMIN");
    cpl_table_erase_column(prova, "LINEZMAX");
    cpl_table_erase_column(prova, "LINEZCON");
    //cpl_table_erase_column(prova, "COLDEN");
    cpl_table_erase_column(prova, "COLDENERR");
    cpl_table_erase_column(prova, "COLDENMIN");
    cpl_table_erase_column(prova, "COLDENMAX");
    cpl_table_erase_column(prova, "COLDENCON");
    cpl_table_erase_column(prova, "THERB");
    cpl_table_erase_column(prova, "THERBERR");
    cpl_table_erase_column(prova, "THERBMIN");
    cpl_table_erase_column(prova, "THERBMAX");
    cpl_table_erase_column(prova, "THERBCON");
    cpl_table_erase_column(prova, "TURBB");
    cpl_table_erase_column(prova, "TURBBERR");
    cpl_table_erase_column(prova, "TURBBMIN");
    cpl_table_erase_column(prova, "TURBBMAX");
    cpl_table_erase_column(prova, "TURBBCON");
    cpl_table_erase_column(prova, "USE");
    cpl_table_erase_column(prova, "CONT");
    cpl_table_erase_column(prova, "SYSTID");
    cpl_table_erase_column(prova, "ORDER");
    //cpl_table_dump(prova, 0, 1000, NULL);
    */
    espda_struct_sort(syst, 1, "WAVEL", FALSE);
    espda_syst_init();       
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Advance the system ID.
  @param   syst_id  The ID of an absorption system.
  @return  CPL error code.
  
  This function advances the ID of an absorption system by one step. The ID 
  consists of a couple of characters (defined through the external variables 
  @a syst_id1, @a syst_id2) that together count from a to z and from aa to zz.
  The function also updates the string @a syst_id obtained by juxtaposing 
  @a syst_id1 and @a syst_id2.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_syst_adv(char syst_id[4]) {
    extern char syst_id1, syst_id2;
    if (syst_id1 == 'a' - 1) {
        sprintf(syst_id, "%c",   syst_id2);
    } 
    else {
        sprintf(syst_id, "%c%c", syst_id1, syst_id2);
    }
    if (syst_id2 == 'z') {
        ++syst_id1;
        syst_id2 = 'a';
    }
    else {
        ++syst_id2;
    }        
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Compute the average redshift of systems.
  @param   syst      A list of absorption systems (LINE_SYST).
  @param   syst_num  Number of systems.
  @return  The average redshift of systems.
  
  This functions scans a list @a syst of @a syst_num absorption systems and 
  computes the redshift of each one by averaging the redshifts of the lines in 
  the system.
  
  The average is computed by @p espda_spec_ave using @a LINEZ as @a data_col. 
  For details on the averaging, see @p espda_spec_ave.

  If @a syst is NULL, an error is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_array *espda_syst_ave(cpl_table *syst, cpl_size syst_num) {
    extern int v_level;
    extern char syst_id1, syst_id2;
    espda_util_error(syst == NULL, CPL_ERROR_NULL_INPUT, NULL);
    espda_syst_init();
    cpl_array *syst_z = cpl_array_new(syst_num, CPL_TYPE_DOUBLE);
    for (int n = 0; n < syst_num; n++) {
        char syst_id[4];
	if (syst_num > 1) {
	    espda_syst_adv(syst_id);
	}
	else {
	    const char *syst_id_arr = cpl_table_get_string(
		syst, "SYSTID", cpl_table_get_nrow(syst)-1);
	    strcpy(syst_id, syst_id_arr);
	}
        int row_start, row_num;
        espda_syst_find(syst, syst_id, &row_start, &row_num);
        double z_ave = 0.0, null_errave, null_rms;

        // Line redshifts are averaged
        espda_spec_ave(syst, "LINEZ", NULL, NULL, row_start,
		       row_start + row_num, &z_ave, &null_errave, &null_rms);
        cpl_array_set_double(syst_z, n, z_ave);
    }
    if (syst_num > 1) {
        espda_syst_init();
    }
    return syst_z;
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Associate blended lines to an absorption system.
  @param   syst      A list of absorption systems.
  @param   syst_id   ID of the system.
  @return  The selected system with the associated blended lines.
  
  This functions selects an absorption system with ID @a syst_id from a list 
  @a syst and finds all the lines that, though not belonging to the system, are 
  blended with its lines. A line with @a SYSTID different from @a syst_id and 
  @a GRPID equal to one or more line of the system is considered a blended line 
  (because the wavelength range used to fit it overlaps the wavelength range 
  used to fit one or more lines of the system). An expanded systems with the 
  associated blended lines is produced, and @a syst_id is advanced for iteration
  purpose.

  If @a syst is NULL, an error is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_table *espda_syst_blend(cpl_table *syst, char syst_id[4]) {
    extern int v_level;
    extern char syst_id1, syst_id2;
    espda_util_error(syst == NULL, CPL_ERROR_NULL_INPUT, NULL);
    
    // System is extracted
    espda_syst_adv(syst_id);
    // CPL_EQUAL_TO cannot be used, otherwise the string are treated as
    // regular expressions
    cpl_table *syst_temp = espda_struct_extract(syst, 1, "SYSTID", 
        CPL_NOT_LESS_THAN, syst_id);
    cpl_table *syst_sel = espda_struct_extract(syst_temp, 1, "SYSTID", 
        CPL_NOT_GREATER_THAN, syst_id);
    //cpl_table *syst_sel = espda_struct_extract(syst, 1, "SYSTID", CPL_EQUAL_TO, 
    //    syst_id);
    // CPL_EQUAL_TO cannot be used, otherwise the string are treated as
    // regular expressions
    cpl_table_unselect_all(syst);
    cpl_table *syst_not = espda_struct_extract(syst, 1, "SYSTID", 
        CPL_NOT_EQUAL_TO, syst_id);
    cpl_table_delete(syst_temp);
    espda_struct_check(syst_sel, espda_struct_sort(syst_sel, 1, "WAVEL", 
	FALSE));

    int grp_max = 0;

    if (cpl_table_get_nrow(syst_sel) > 0) {
        grp_max = (int)cpl_table_get_column_max(syst_sel, "GRPID");
    }

    //printf("%i\n", cpl_table_get_nrow(syst_sel));
    
    // Line groups are scanned
    for (int g = 0; g < grp_max; g++) {
        cpl_table *grp = espda_struct_extract(syst_sel, 1, "GRPID", 
            CPL_EQUAL_TO, g + 1);
            
        // If lines not belonging to the systems are find in a group, they are
        // associated to the system 
        if (cpl_table_get_nrow(grp) != 0) {
            cpl_table *blend = espda_struct_extract(syst_not, 1, "GRPID", 
                CPL_EQUAL_TO, g + 1);
            if (cpl_table_get_nrow(blend) != 0) {
                for (int h = 0; h < cpl_table_get_nrow(grp); h++) {
                    for (int b = 0; b < cpl_table_get_nrow(blend); b++) {
                        if (cpl_table_get_double(grp,   "WAVEL", h, NULL) ==
                            cpl_table_get_double(blend, "WAVEL", b, NULL)) {
                            cpl_table_set_double(blend, "WAVEL", b, INV_DOUBLE);
                        }
                    }
                }
                espda_struct_erase(blend, 1, "WAVEL", CPL_EQUAL_TO, INV_DOUBLE);
                espda_struct_insert(syst_sel, blend, 
                    cpl_table_get_nrow(syst_sel)); 
            }
            cpl_table_delete(blend);
        }
        cpl_table_delete(grp);
    }
    cpl_table_delete(syst_not);
    espda_struct_check(syst_sel, espda_struct_sort(syst_sel, 1, "WAVEL", FALSE));
    return syst_sel;
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Confirm the absorption systems.
  @param   syst      A list of absorption systems (LINE_SYST).
  @param   line      The list of lines where the systems were identified (LINE).
  @param   comp_num  Minimum number of components.
  @param   cand_num  Number of candidate systems.
  @param   syst_num  Number of confirmed systems.
  @param   noid      The list of non-identified lines (LINE_SYST).
  @return  CPL error code.
  
  This functions scans a list of lines @a syst where @a cand_num candidate 
  absorption systems were already identified, and applies the existence and 
  acceptance rules (depending on the minimum number of components @a comp_num to
  confirm a system). The lines that do not pass the test are removed from 
  @a syst and the number of confirmed system @a syst_num is computed. 
  Concurrently, the lines that do pass the test are removed from the list 
  @a line where the systems were originally identified, to produce a list 
  @a noid of non-identified lines.
  
  If @a syst or @a line are NULL, an error is issued. If @a syst_num is 0 and 
  the verbosity level is greater than 0, a warning is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_syst_conf(cpl_table *syst, cpl_table *line, int comp_num,
    int cand_num, int *syst_num, cpl_table **noid) {
    extern int v_level;
    extern char syst_id1, syst_id2;
    espda_util_error(syst == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    espda_util_error(line == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    *syst_num = cand_num;
    //cpl_table *syst_bck = cpl_table_duplicate(syst);
    espda_struct_sort(syst, 1, "SYSTID", FALSE);
    espda_print_verb(1, cpl_msg_info(cpl_func, "Applying existence and "
        "acceptance rules..."));

    // Candidate systems are confirmed after existence and acceptance checks
    espda_syst_init();
    for (int n = 0; n < cand_num; n++) {
        char syst_id[4];
        espda_syst_adv(syst_id);
        int exist_flag = 1, disc_flag = 1;
        while (exist_flag == 1 && disc_flag == 1) {
            espda_struct_sort(syst, 1, "SYSTID", FALSE);
            int row_start, row_num;
            espda_syst_find(syst, syst_id, &row_start, &row_num);

            // Existence check
            espda_syst_exist(syst, syst_id, comp_num, &row_start, &row_num, 
                &exist_flag);
            if (exist_flag == 0) {
                --*syst_num;
            }
            else {

                // Acceptance check
                espda_syst_accept(syst, syst_id, &row_start, &row_num, 
                    &disc_flag);                
            }
        }
    }
    espda_syst_init();
    if (*syst_num > 0) {
        espda_print_verb(1, cpl_msg_info(cpl_func, " Confirmed %i candidates "
            "(%1.0f percent).", *syst_num, 100.0 * (double) * syst_num / 
            (double)cand_num));
        espda_print_verb(2, espda_dfs_call(0,
            espda_print_syst(syst, cand_num)));
    }
    else {
        espda_print_verb(1, cpl_msg_warning(cpl_func, " No candidate "
            "confirmed."));
    }
    espda_struct_erase(syst, 1, "SYSTID", CPL_EQUAL_TO, INV_STRING);
    espda_struct_sort(syst, 1, "LINEZ", FALSE);
    
    // Discarded lines are extracted from the line list
    *noid = cpl_table_duplicate(line);
    for (int m = 0; m < cpl_table_get_nrow(*noid); m++) {
        for (int n = 0; n < cpl_table_get_nrow(syst); n++) {
            if (cpl_table_get_double(*noid, "WAVEL", m, NULL) ==
                cpl_table_get_double(syst,  "WAVEL", n, NULL)) {
                cpl_table_set_invalid(*noid, "WAVEL", m);
            }
        }
    }
    cpl_table_erase_invalid(*noid);
    if (!cpl_table_has_column(*noid, "WAVEL") ||
        cpl_table_get_nrow(*noid) == 0) {
        *noid = NULL;
        espda_print_verb(1, cpl_msg_info(cpl_func, " All lines associated with "
            "a system."));
    }
    else {
        espda_print_verb(1, cpl_msg_info(cpl_func, " %i lines not associated "
            "with any system (%1.0f percent).", (int)cpl_table_get_nrow(*noid), 
            100.0 * (double)cpl_table_get_nrow(*noid) / 
            (double)cpl_table_get_nrow(line)));
    }     
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Put constraints on multiplet components.
  @param   syst  A list of absorption systems (LINE_SYST).
  @param   ion   A list of ionic transitions (QSO_ION).
  @return  CPL error code.
  
  This functions takes a list of absorption systems @a syst, detects the lines 
  identified as components of a multiplet, and set constraints on the value of 
  their Voigt parameters, so that each multiplet member can be fitted with the 
  same redshift @f$ z @f$, column density @f$ N @f$, and thermal broadening 
  @f$ b @f$ of the matching members. A list of ionic transition @a ion is used 
  to identify the multiplet components (all lines whose @a LINEID contains one 
  of the tags from column @a MULTID of @a ion is considered as a member of a 
  multiplet). The constraints are set with a standard syntax: 
  @a mult_ID-@a par_symbol-@a count, where the @a mult_ID is the multiplet ID as
  taken from column @a MULTID of @a ion, @a par_symbol is the parameter symbol 
  (either "z", "N", or "b" for redshift, column density, and thermal broadening 
  respectively), and @a count is a 3-digit counter increased for each new 
  multiplet component.
  
  If @a syst or @a ion are NULL, an error is issued. 
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_syst_constr(cpl_table *syst, cpl_table *ion) {
    extern int v_level;
    espda_util_error(syst == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    espda_util_error(ion == NULL,  CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    espda_struct_sort(syst, 2, "SYSTID", FALSE, "LINEID", FALSE);
    char *line_id_old = malloc(sizeof(char) * LEN_STRING);
    strcpy(line_id_old, "");
    char *syst_id_old = malloc(sizeof(char) * LEN_STRING);
    strcpy(syst_id_old, "");
    int line_start = 0, line_count = 0, line_max = 0;
    espda_struct_sort(syst, 2, "SYSTID", FALSE, "WAVEL", FALSE);
    for (int i = 0; i < cpl_table_get_nrow(syst); i++) {
    
        // Multiplet components are identified
        const char *syst_id = cpl_table_get_string(syst, "SYSTID", i);
        const char *line_id = cpl_table_get_string(syst, "LINEID", i);
        if (strcmp(syst_id, syst_id_old)) {
            line_start = line_max;
            strcpy(syst_id_old, syst_id);
        }
        if (strcmp(line_id, line_id_old)) {
            line_count = line_start;
            strcpy(line_id_old, line_id);
        }
        else {
            line_count++;
        }
        cpl_size mult_flag = 0;
        // CPL_EQUAL_TO cannot be used, otherwise the string are treated as
        // regular expressions
        cpl_table *syst_temp = espda_struct_extract(syst, 1, "SYSTID", 
            CPL_NOT_LESS_THAN, syst_id);
        cpl_table *syst_sel = espda_struct_extract(syst_temp, 1, "SYSTID", 
            CPL_NOT_GREATER_THAN, syst_id);
        //cpl_table *syst_sel = espda_struct_extract(syst, 1, "SYSTID", 
        //    CPL_EQUAL_TO, syst_id);
        cpl_table *ion_sel  = espda_struct_extract(ion,  1, "LINEID", 
            CPL_EQUAL_TO, line_id);
        //const char *mult_id = malloc(sizeof(char) * LEN_STRING);
	char *mult_id = malloc(sizeof(char) * LEN_STRING);
        if (cpl_table_get_nrow(ion_sel) > 0) { 
            strcpy(mult_id, cpl_table_get_string(ion_sel, "MULTID", 0));
	    char *line_id_mult = malloc(sizeof(char) * LEN_STRING);
	    strcpy(line_id_mult, cpl_table_get_string(ion_sel, 
                "LINEID_MULT", 0));
	    char *line_id_sel = strtok(line_id_mult, ",");
            while (line_id_sel) {
                line_id_sel = strtok(NULL, ",");                
                if(strcmp(line_id_mult, "")) {
                    cpl_table_unselect_all(syst_sel);
                    cpl_table_or_selected_string(syst_sel, "LINEID", 
                        CPL_EQUAL_TO, line_id_mult);    
                    mult_flag = cpl_table_count_selected(syst_sel);
                }
            }
	    free(line_id_mult);
        } 
        espda_struct_delete(CPL_TABLE, 3, syst_temp, syst_sel, ion_sel);

        // Constraints are set
        if (mult_flag != 0) {
            char linez_con[20], colden_con[20], therb_con[20];
            sprintf(linez_con,  "%s-z-%03i", mult_id, line_count);
            sprintf(colden_con, "%s-N-%03i", mult_id, line_count);
            sprintf(therb_con,  "%s-b-%03i", mult_id, line_count);
            cpl_table_set_string(syst, "LINEZCON",  i, linez_con);
            cpl_table_set_string(syst, "COLDENCON", i, colden_con);
            cpl_table_set_string(syst, "THERBCON",  i, therb_con);
            line_max = line_count + 1;
        }
        free(mult_id);
    }

    espda_struct_sort(syst, 1, "WAVEL", FALSE);
    free(line_id_old);
    free(syst_id_old);
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Decrease the system ID.
  @return  CPL error code.

  This function decreases the ID of an absorption system by one step. The ID
  consists of a couple of characters (defined through the external variables 
  @a syst_id1, @a syst_id2) that together count from a to z and from aa to zz.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_syst_dec() {
    extern char syst_id1, syst_id2;
    if (syst_id1 == 'a' && syst_id2 == 'a') {
        --syst_id1;
        syst_id2 = 'z';
    }
    else {
        --syst_id2;
    }        
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Apply existence rules.
  @param   syst        A list of absorption systems (LINE_SYST).
  @param   syst_id     The ID of the selected system.
  @param   comp_num    Minimum number of components.
  @param   row_start   The starting row of the selected system.
  @param   row_num     Number of rows of the selected system.
  @param   exist_flag  Flag to notify that the candidate exists.
  @return  CPL error code.
  
  This function evaluates an absorption system from a list @a syst against a set 
  of existence rules and discards all systems that do not comply. The system is 
  defined by its ID @a syst_id and by the starting row @a row_start and number 
  of rows @a row_num in @a syst (as obtained when @a syst is sorted by ascending 
  @a SYSTID). The rules are:

    - The system must have more than @a comp_num components.

  If the system is discarded, the flag @a exist_flag is set to 0.
  
  If @a syst is NULL, an error is issued. If the system is discarded and the 
  verbosity level is greater than 0, a warning is issued.

  N.B. Existence rules have not been completely implemented yet.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_syst_exist(cpl_table *syst, char *syst_id, int comp_num,
    int *row_start, int *row_num, int *exist_flag) {
    extern int v_level;
    espda_util_error(syst == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    *exist_flag = 1;
    if (*row_num > 0) {

        char msg[LEN_STRING];

	// Provisional instructions to artificially discard one candidate
        /*
        if (!strcmp(syst_id, "b")) {
            *exist_flag = 0;
            sprintf(msg, "Artificially discarded.");
        }
        */

        cpl_table *temp = espda_struct_extract(syst, 1, "SYSTID", 
            CPL_NOT_LESS_THAN, syst_id);
        cpl_table *sel = espda_struct_extract(temp, 1, "SYSTID", 
            CPL_NOT_GREATER_THAN, syst_id);


	// Instructions to discard systems with non-matching line peaks
	char mult_id[][5] = {"NV", "SiII", "SiIV", "CIV", "FeII", "MgII",
			     "CaII"};
	for (int m = 0; m < (int)(sizeof(mult_id) / sizeof(mult_id[0])); m++) {
	    cpl_table *mult = espda_struct_extract(sel, 1, "MULTID",
						   CPL_EQUAL_TO, mult_id[m]);
	    espda_struct_sort(mult, 1, "WAVEL", FALSE);
	    cpl_size mult_size = cpl_table_get_nrow(mult);
	    if (mult_size > 0) {
	        if (cpl_table_get_column_stdev(mult, "NORM") > 0.5) {
		    *exist_flag = 0;
		    sprintf(msg, "Peak values of %s not matching.", mult_id[m]);
		}
	    }
	}
	
        // Instructions to discard FeII based on members' oscillator strength
        cpl_table *feii_2586 = espda_struct_extract(sel, 1, "LINEID", 
	    CPL_EQUAL_TO, "FeII_2586");
        cpl_table *feii_2344 = espda_struct_extract(sel, 1, "LINEID", 
	    CPL_EQUAL_TO, "FeII_2344");
        cpl_table *feii_2600 = espda_struct_extract(sel, 1, "LINEID", 
	    CPL_EQUAL_TO, "FeII_2600");
        cpl_table *feii_2382 = espda_struct_extract(sel, 1, "LINEID", 
	    CPL_EQUAL_TO, "FeII_2382");
	cpl_size feii_2586_num = cpl_table_get_nrow(feii_2586);
	cpl_size feii_2344_num = cpl_table_get_nrow(feii_2344);
	cpl_size feii_2600_num = cpl_table_get_nrow(feii_2600);
	cpl_size feii_2382_num = cpl_table_get_nrow(feii_2382);
	if (feii_2382_num == 0) {
	    if (feii_2586_num > 0 || feii_2344_num > 0 || feii_2600_num > 0) {
	        *exist_flag = 0;
		sprintf(msg, "FeII_2382 absent from FeII multiplet.");
            }
	}
        else {
    	    if (feii_2600_num == 0) {
    	        if (feii_2586_num > 0 || feii_2344_num > 0) {
    	            *exist_flag = 0;
    		    sprintf(msg, "FeII_2600 absent from FeII multiplet with "
			"members of lower oscillator strength.");
                }
	    }
            else {
	    	if (feii_2344_num == 0) {
	    	    if (feii_2586_num > 0) {
	    	        *exist_flag = 0;
	    		    sprintf(msg, "FeII_2344 absent from FeII multiplet "
				"with FeII_2586.");
	            }
		}
            }
	}
	espda_struct_delete(CPL_TABLE, 4, feii_2586, feii_2344, feii_2600, 
            feii_2382);

	// Instructions to discard SiII based on members' oscillator strength
        cpl_table *siii_1304 = espda_struct_extract(sel, 1, "LINEID", 
	    CPL_EQUAL_TO, "SiII_1304");
        cpl_table *siii_1526 = espda_struct_extract(sel, 1, "LINEID", 
	    CPL_EQUAL_TO, "SiII_1526");
        cpl_table *siii_1260 = espda_struct_extract(sel, 1, "LINEID", 
	    CPL_EQUAL_TO, "SiII_1260");
	cpl_size siii_1304_num = cpl_table_get_nrow(siii_1304);
	cpl_size siii_1526_num = cpl_table_get_nrow(siii_1526);
	cpl_size siii_1260_num = cpl_table_get_nrow(siii_1260);
	if (siii_1260_num == 0) {
	    if (siii_1526_num > 0 || siii_1304_num > 0) {
	        *exist_flag = 0;
		sprintf(msg, "SiII_1260 absent from SiII multiplet.");
            }
	}
        else {
    	    if (siii_1526_num == 0) {
    	        if (siii_1304_num > 0) {
    	            *exist_flag = 0;
    		    sprintf(msg, "SiII_1526 absent from SiII multiplet with "
			"SiII_1304.");
                }
	    }
	}
	espda_struct_delete(CPL_TABLE, 5, temp, sel, siii_1304, siii_1526, 
            siii_1260);


        // Real rules to discard candidates with insufficient number of lines
        if (*row_num < comp_num) {
            *exist_flag = 0;
            sprintf(msg, "%i lines are not enough.", *row_num);
        }

        // Systems are discarded
        if (*exist_flag == 0) {
            espda_print_verb(1, cpl_msg_warning(cpl_func, " Candidate %s does "
                "not exist! %s", syst_id, msg));
            espda_syst_update(syst, NULL, NULL, NULL, 0.0, "remove", 
                *row_start, *row_num);
        }
    }
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Find an absoprtion system.
  @param   syst       A list of absorption systems (LINE_SYST).
  @param   syst_id    The ID of the selected system.
  @param   row_start  The starting row of the selected system.
  @param   row_num    Number of rows of the selected system.
  @return  CPL error code.
  
  This function scans a list of absorption systems @a syst previously sorted by 
  @a SYSTID and finds the rows corresponding to the systems with ID @a syst_id. 
  The starting row @a row_start and the number of rows @a row_num are issued.
  
  If @a syst is NULL, an error is issued. 
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_syst_find(cpl_table *syst, char *syst_id, int *row_start,
    int *row_num) {
    espda_util_error(syst == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());

    int found_flag = 0;
    int i_start = 0, i_end = cpl_table_get_nrow(syst);
    *row_start = 0;
    *row_num   = 0;
    for (int i = i_start; i < i_end; i++) {
        const char *syst_id_comp = cpl_table_get_string(syst, "SYSTID", i);
        char  temp[LEN_STRING];
        strcpy(temp, syst_id_comp);
        //char *found = strstr(temp, syst_id);
        //if (found != NULL) {
        if (!strcmp(temp, syst_id)) {
            found_flag = 1;
            ++*row_num;
        }
        else if (found_flag == 0) {
            *row_start = i + 1;
        }
    }
    return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Update an absorption system.
  @param   syst       A list of absorption systems (LINE_SYST).
  @param   syst_id    The ID of the selected system.
  @param   line_id    The ID of the line.
  @param   line_z     The redshift of the line.
  @param   edit       The operation to perform.
  @param   row_start  The starting row of the selected system.
  @param   row_num    Number of rows of the selected system.
  @return  CPL error code.
  
  This functions selects an absorption system from a list @a syst and updates 
  it, either adding or removing a line. The system is defined by its ID 
  @a syst_id and by the starting row @a row_start and number of rows @a row_num 
  in @a syst (as obtained when @a syst is sorted by ascending @a SYSTID). The 
  behavior depends on @a edit, which describes the operation to perform:

    - if @a edit is "add", the line is added and the corresponding row in 
      @a syst is updated with the system ID @a syst_id, the line ID @a line_id, 
      and the line redshift @a line_z (respectively copied into columns 
      @a SYSTID, @a LINEID, and @a LINEZ);
    - if @a edit is "remove", the value of @a SYSTID in the corresponding row is 
      changed to the globally-defined void string INV_STRING.

  If @a syst is NULL, an error is issued. 
*/
/*----------------------------------------------------------------------------*/

cpl_error_code espda_syst_update(cpl_table  *syst, const char *syst_id, 
    const char *line_id, const char *mult_id, double line_z, char *edit, 
    int row_start, int row_num) {
    for (int i = row_start; i < row_start + row_num; i++) {
        const char *syst_id_comp = cpl_table_get_string(syst, "SYSTID", i);

        // Line is added
        if (!strcmp(edit, "add")) {
            if (strcmp(syst_id_comp, INV_STRING)) {
                cpl_table *row = espda_struct_extract(syst, 0, i, 1);
                cpl_table_set_string(row,  "SYSTID", 0, syst_id);
                cpl_table_set_string(row,  "LINEID", 0, line_id);
                cpl_table_set_string(row,  "MULTID", 0, mult_id);
                cpl_table_set_double(row,  "LINEZ",  0, line_z);
                cpl_table_insert(syst, row, cpl_table_get_nrow(syst));
                cpl_table_delete(row);
            }
            else {
                cpl_table_set_string(syst, "SYSTID", i, syst_id);
                cpl_table_set_string(syst, "LINEID", i, line_id);
                cpl_table_set_string(syst, "MULTID", i, mult_id);
                cpl_table_set_double(syst, "LINEZ",  i, line_z);
            }
        }

        // Line or system is removed 
        if (!strcmp(edit, "remove")) {
            cpl_table_set_string(syst, "SYSTID", i, INV_STRING);
        }
    }
    return cpl_error_get_code();
}

/**@}*/
