/* $Id: espda_coadd_spec.c 256541 2020-07-24 13:12:49Z gcupani $
 *
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: gcupani $
 * $Date: 2020-07-24 13:12:49 +0000 (Fri, 24 Jul 2020) $
 * $Revision: 256541 $
 * $Name:  $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#include <string.h>
#include <time.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>

#include "espda_check.h"
#include "espda_dfs.h"
#include "espda_frame.h"
#include "espda_model.h"
#include "espda_param.h"
#include "espda_spec.h"
#include "espda_wrap.h"

/*-----------------------------------------------------------------------------
                             Plugin registration
 -----------------------------------------------------------------------------*/

int cpl_plugin_get_info(cpl_pluginlist * list);

/*-----------------------------------------------------------------------------
                            Private function prototypes
 -----------------------------------------------------------------------------*/

static int espda_coadd_spec_create(cpl_plugin *);
static int espda_coadd_spec_exec(cpl_plugin *);
static int espda_coadd_spec_destroy(cpl_plugin *);
static int espda_coadd_spec(cpl_frameset *, const cpl_parameterlist *);

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static char espda_coadd_spec_description[] =
"This recipe combines different observations of the same object into a single "
"spectrum. Input order spectra are optionally equalized at the same flux level "
"and rebinned.\n"
"\n"
"INPUT FRAMES                             FORMAT         "
"# EXPECTED  TAGS               NOTES\n"
"1D merged spectrum                       S1D            "
"0 to 50     S1D\n"
"2D spectrum with orders                  S2D            "
"0 to 50     S2D                must be present if S1D is not present\n"
"Matrix of pixel sizes in wavel. units    DLL            "
"0 or 1      DLL_MATRIX\n"
"\n"
"OUTPUT FRAMES                            FORMAT         "
"# PRODUCED  TAGS               NOTES\n"
"Original full spectrum                   SPEC_FULL      "
"0 or 1      FSPEC_ORIG         produced only if S2D is present\n"
"Original merged spectrum                 SPEC_RMS       "
"0 or 1      RSPEC_ORIG\n"
"Full spectrum, optionally equalised      SPEC_FULL      "
"0 or 1      FSPEC_PRE          produced only if S2D is present\n"
"Rebinned spectrum, optionally equalised  SPEC_RMS       "
"1           RSPEC_PRE\n"
"Rebinned instrumental profile            PROF           "
"1           RPROF"
;

/*-----------------------------------------------------------------------------
                                Function code
 -----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module.
  @param    list    the plugin list
  @return   0 if everything is ok, 1 otherwise
  @note     Only this function is exported

  Create the recipe instance and make it available to the application using the
  interface.
 */
/*----------------------------------------------------------------------------*/

int cpl_plugin_get_info(cpl_pluginlist *list)
{
    cpl_recipe *recipe = cpl_calloc(1, sizeof *recipe );
    cpl_plugin *plugin = &recipe->interface;

    if (cpl_plugin_init(plugin, CPL_PLUGIN_API, ESPDA_BINARY_VERSION,
                        CPL_PLUGIN_TYPE_RECIPE, "espda_coadd_spec", "Combine "
                        "different spectra of the same object",
                        espda_coadd_spec_description, "Guido Cupani",
                        PACKAGE_BUGREPORT, espda_dfs_license(),
                        espda_coadd_spec_create, espda_coadd_spec_exec,
                        espda_coadd_spec_destroy)) {
        cpl_msg_error(cpl_func, "Plugin initialization failed");
        (void)cpl_error_set_where(cpl_func);
        return 1;
    }

    if (cpl_pluginlist_append(list, plugin)) {
        cpl_msg_error(cpl_func, "Error adding plugin to list");
        (void)cpl_error_set_where(cpl_func);
        return 1;
    }

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options
  @param    plugin  the plugin
  @return   0 if everything is ok

  Defining the command-line/configuration parameters for the recipe.
 */
/*----------------------------------------------------------------------------*/
static int espda_coadd_spec_create(cpl_plugin *plugin)
{
    cpl_recipe    *recipe;
    cpl_parameter *param;

    /* Do not create the recipe if an error code is already set */
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());
        return (int)cpl_error_get_code();
    }

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    /* Create the parameters list in the cpl_recipe object */
    recipe->parameters = cpl_parameterlist_new();
    if (recipe->parameters == NULL) {
        cpl_msg_error(cpl_func, "Parameter list allocation failed");
        cpl_ensure_code(0, (int)CPL_ERROR_ILLEGAL_OUTPUT);
    }

    /* Fill the parameters list */
    /*
    param = cpl_parameter_new_value("espda.espda_coadd_spec.rej_flag",
				    CPL_TYPE_BOOL,
				    "Flag to use reject bad pixels as "
				    "identified by the DRS: TRUE, FALSE",
				    "espda_coadd_spec",
				    "TRUE");
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "rej-flag");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);
    */

    param = cpl_parameter_new_value("espda.espda_coadd_spec.rebin_flag",
				    CPL_TYPE_BOOL,
				    "Flag to rebin spectra: TRUE, FALSE "
				    "(disregarded if only S1D is present)",
				    "espda_coadd_spec",
				    "TRUE");
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "rebin-flag");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_coadd_spec.air_flag",
				    CPL_TYPE_BOOL,
				    "Flag to use air wavelengths instead of "
				    "vacuum: TRUE, FALSE",
				    "espda_coadd_spec",
				    "FALSE");
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "air-flag");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_coadd_spec.equal_flag",
				    CPL_TYPE_BOOL,
				    "Flag to equalize spectra: TRUE, FALSE (if "
				    "FALSE, equal-ref is disregarded)",
				    "espda_coadd_spec",
				    "FALSE");
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "equal-flag");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_coadd_spec.equal_ref",
				    CPL_TYPE_INT,
				    "ID num2dber of the reference exposure to be "
				    "used in equalization: [0, num2dber of "
				    "exposures]",
				    "espda_coadd_spec",
				    1);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "equal-ref");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_coadd_spec.kappa",
				    CPL_TYPE_DOUBLE,
				    "Threshold for kappa-sigma clipping to be "
				    "used in rebinning: [1.0, 1000.0]",
				    "espda_coadd_spec",
				    3.0);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "kappa");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_coadd_spec.wavel_min",
				    CPL_TYPE_DOUBLE,
				    "Minimum wavelength [nm]: [300,2500]",
				    "espda.espda_coadd_spec",
				    WAVEL_START);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "wavel-min");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_coadd_spec.wavel_max",
				    CPL_TYPE_DOUBLE,
				    "Maximum wavelength [nm]: [300,2500]",
				    "espda.espda_coadd_spec",
				    WAVEL_END);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "wavel-max");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_coadd_spec.vel_step",
				    CPL_TYPE_DOUBLE,
				    "Velocity step for the SINGLEUHR "
				    "(SINGLEHR, MULTIMR) mode [km/s]: "
				    "[0.3, 1.0e2]",
				    "espda.espda_coadd_spec",
				    VEL_STEP,
				    0.5);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "vel-step");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_coadd_spec.model_resol",
				    CPL_TYPE_DOUBLE,
				    "Resolution of the instrument model "
				    "[1.0e3, 1.0e6]",
				    "espda.espda_coadd_spec",
				    6.0e4);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "model-resol");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_coadd_spec.rv",
				    CPL_TYPE_DOUBLE,
				    "Radial velocity, copied from keyword "
				    "CCF rv if present [km/s] (only for "
				    "stellar spectra)",
				    "espda.espda_coadd_spec",
				    -9999.0);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "rv");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_coadd_spec.zem",
				    CPL_TYPE_DOUBLE,
				    "Emission redshift, to be used instead of HIERARCH ESO "
            "OCS OBJ Z EM (in case a better estimation is available)",
				    "espda.espda_coadd_spec",
				    -9999.0);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "zem");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_coadd_spec.adp_columns",
				    CPL_TYPE_STRING,
				    "Columns to upload from input in the S1D ADP format (disregarded "
            "if input has a different format)",
				    "espda.espda_coadd_spec",
				    "WAVE,FLUX,ERR,QUAL");
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "adp-columns");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_coadd_spec_exec(cpl_plugin *plugin)
{
    int             recipe_status;

    cpl_recipe     *recipe;
    cpl_errorstate  initial_errorstate = cpl_errorstate_get();

    /* Return immediately if an error code is already set */
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());
        return (int)cpl_error_get_code();
    }

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    /* Verify parameter and frame lists */
    if (recipe->parameters == NULL) {
        cpl_msg_error(cpl_func, "Recipe invoked with NULL parameter list");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }
    if (recipe->frames == NULL) {
        cpl_msg_error(cpl_func, "Recipe invoked with NULL frame set");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Invoke the recipe */
    recipe_status = espda_coadd_spec(recipe->frames, recipe->parameters);

    /* call DFS-compliance of the products */
    if (cpl_dfs_update_product_header(recipe->frames)) {
        if (!recipe_status) recipe_status = (int)cpl_error_get_code();
    }

    if (!cpl_errorstate_is_equal(initial_errorstate)) {
        /* Dump the error history since recipe execution start.
           At this point the recipe cannot recover from the error */
        cpl_errorstate_dump(initial_errorstate, CPL_FALSE, NULL);
    }

    return recipe_status;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_coadd_spec_destroy(cpl_plugin *plugin)
{
    cpl_recipe * recipe;

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    cpl_parameterlist_delete(recipe->parameters);

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    frameset   the frames list
  @param    parlist    the parameters list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_coadd_spec(cpl_frameset *frameset,
                            const cpl_parameterlist *parlist) {
    extern int v_level;

/*
 * Input is checked and loaded. FITS frames are extracted into a merged spectrum
 * or an order spectrum. Redshift is also loaded, to distinguish between stellar
 * spectra (zem = 0) and quasar spectra (zem > 0). The first spectrum frame
 * (S1D or S2D) is taken as reference for redshift, assuming that all 1D or 2D
 * spectra refer to a single object.
 */


    // Check if frame is from DRS
    int drs_flag = 0;
    if (cpl_frameset_count_tags(frameset, S1D) > 0) {
        char *pipe_id = (char *)malloc(sizeof(char) * LEN_STRING);
        espda_frame_desc(frameset, "S1D", "ESO PRO REC1 PIPE ID",
			 &pipe_id);
        char *drs = strstr(pipe_id, "espdr");
        if (drs != NULL) {
            drs_flag = 1;
        }
    }
    if (cpl_frameset_count_tags(frameset, S2D) > 0) {
        char *pipe_id = (char *)malloc(sizeof(char) * LEN_STRING);
        espda_frame_desc(frameset, "S2D", "ESO PRO REC1 PIPE ID",
			 &pipe_id);
        char *drs = strstr(pipe_id, "espdr");
        if (drs != NULL) {
            drs_flag = 1;
        }
    }
    if (cpl_frameset_count_tags(frameset, S2D_SKYSUB) > 0) {
        char *pipe_id = (char *)malloc(sizeof(char) * LEN_STRING);
        espda_frame_desc(frameset, "S2D_SKYSUB", "ESO PRO REC1 PIPE ID",
			 &pipe_id);
        char *drs = strstr(pipe_id, "espdr");
        if (drs != NULL) {
            drs_flag = 1;
        }
    }

    // Check frames
    espda_dfs_call(0, espda_dfs_groups(frameset));
    int min = 0;
    int num2d = cpl_frameset_count_tags(frameset, FLUX_2D);
    int num = cpl_frameset_count_tags(frameset, FLUX_2D);
    int skysub_flag = 0;
    if (drs_flag == 1) {
        num2d = cpl_frameset_count_tags(frameset, S2D);
        num = cpl_frameset_count_tags(frameset, S2D);
	if (num2d == 0) {
   	    skysub_flag = 1;
	    num2d = cpl_frameset_count_tags(frameset, S2D_SKYSUB);
      num = cpl_frameset_count_tags(frameset, S2D_SKYSUB);
  }


	    if (num == 0) {
	        num = cpl_frameset_count_tags(frameset, S1D);
	    }

    }
    if (cpl_frameset_count_tags(frameset, S1D) == 0) {
        min = 1;
    }

    cpl_recipeconfig *cfg = cpl_recipeconfig_new();
    char *cfg_tag = S1D;
    cpl_recipeconfig_set_tag  (cfg, cfg_tag, 0, FRAME_MAX);
    cpl_table *rprof_multimr = NULL;
    if (drs_flag == 1) {
        if (skysub_flag == 0) {
            cpl_recipeconfig_set_input(cfg, cfg_tag, S2D, min, FRAME_MAX);
	}
	else {
            cpl_recipeconfig_set_input(cfg, cfg_tag, S2D_SKYSUB, min,
				       FRAME_MAX);
	}
        cpl_recipeconfig_set_input(cfg, cfg_tag, "DLL_MATRIX",  0, 1);
        cpl_recipeconfig_set_input(cfg, cfg_tag, RPROF_MULTIMR,  0, 1);
	rprof_multimr = espda_frame_load(frameset, cfg, cfg_tag, 2);
    }
    else {
        cpl_recipeconfig_set_input(cfg, cfg_tag, FLUX_2D, min, FRAME_MAX);
        cpl_recipeconfig_set_input(cfg, cfg_tag, FLUXERR_2D, num2d, num2d);
        cpl_recipeconfig_set_input(cfg, cfg_tag, WAVEL_2D, num2d, num2d);
        cpl_recipeconfig_set_input(cfg, cfg_tag, BLAZE_2D, 0, FRAME_MAX);
        cpl_recipeconfig_set_input(cfg, cfg_tag, RES_MAP, 0, 1);
    }
    espda_wrap_check(frameset, cfg, cfg_tag);

   // Load and check parameters
    int rebin_flag, air_flag, equal_flag, equal_ref;
    double kappa, wavel_min, wavel_max, vel_step, model_resol, rv, zem;
    char *adp_columns = (char *)malloc(sizeof(char) * LEN_STRING);;
    //espda_param_load(parlist, COADD_SPEC_REJ_FLAG, &rej_flag);
    espda_param_load(parlist, COADD_SPEC_REBIN_FLAG, &rebin_flag);
    espda_param_load(parlist, COADD_SPEC_AIR_FLAG, &air_flag);
    espda_param_load(parlist, COADD_SPEC_EQUAL_FLAG, &equal_flag);
    espda_param_load(parlist, COADD_SPEC_EQUAL_REF, &equal_ref, 0, num2d);
    espda_param_load(parlist, COADD_SPEC_KAPPA, &kappa, 1.0, 1000.0);
    espda_param_load(parlist, COADD_SPEC_WAVEL_MIN, &wavel_min, 300.0, 2500.0);
    espda_param_load(parlist, COADD_SPEC_WAVEL_MAX, &wavel_max, 300.0, 2500.0);
    espda_param_load(parlist, COADD_SPEC_VEL_STEP, &vel_step, 0.3, 1.e2);
    espda_param_load(parlist, COADD_SPEC_MODEL_RESOL, &model_resol, 1.e3, 1.e6);
    espda_param_load(parlist, COADD_SPEC_RV, &rv, -INFINITY, INFINITY);
    espda_param_load(parlist, COADD_SPEC_ZEM, &zem, -INFINITY, INFINITY);
    espda_param_load(parlist, COADD_SPEC_ADP_COLUMNS, &adp_columns, "", "");
    //printf("%f %f\n", kappa, wavel_max);

    cpl_array *cols = espda_param_extract(adp_columns, ",", 0, 1,
						CPL_TYPE_STRING);

    cpl_array *user_rv = cpl_array_new(num, CPL_TYPE_DOUBLE);
    cpl_array_fill_window_double(user_rv, 0, num, rv);

    if (num2d == 0) {
        if (rebin_flag == 1) {
       	    cpl_msg_info(cpl_func, "Coaddition and rebinning will be performed "
            "from already-rebinned 1D spectra because 2D order spectra were "
			 "not provided.");
        }
        if (equal_flag == 1) {
       	    equal_flag = 0;
       	    cpl_msg_warning(cpl_func, "Equalization will not be performed "
			    "(even though equal-flag=1) because 2D order "
			    "spectra were not provided.");
        }
    }
    else {
        if (rebin_flag == 0) {
       	    cpl_msg_info(cpl_func, "Rebinning will not be performed.");
        }
    }
    if (wavel_max < wavel_min) {
        double temp = wavel_max;
        wavel_max   = wavel_min;
        wavel_min   = temp;
        cpl_msg_warning(cpl_func, "The endpoints of the wavelength range were "
			"reversed and have been swapped.");
    }
    if (user_rv != NULL && rv != INV_DOUBLE) {
        cpl_msg_warning(cpl_func, "Only one value of radial velocity provided: "
			"%f. This value will be applied to all exposures.", rv);
    }
    else if (user_rv != NULL && cpl_array_get_size(user_rv) != num && num != 0){
        cpl_msg_warning(cpl_func, "The number of radial velocity values is "
			"different from the number of spectra! Parameter rv "
			"will be neglected.");
	user_rv = NULL;
    }

    // Check if data are from the DRS
    cpl_frameset *frameset_mod = cpl_frameset_duplicate(frameset);

    // Extract frames
    cpl_table *fspec_orig = NULL;
    cpl_table *rspec_orig = NULL;
    cpl_array *fspec_rv   = NULL;
    cpl_array *rspec_rv   = NULL;
    cpl_array *fspec_berv = NULL;
    cpl_array *rspec_berv = NULL;
    cpl_array *fspec_snr600 = NULL;
    cpl_array *rspec_snr600 = NULL;
    if (drs_flag == 0) {
	espda_frame_extract_old(frameset_mod, wavel_min, wavel_max,
				&fspec_orig, &rspec_orig, &fspec_rv, &rspec_rv);
    }
    else {
      /*
espda_util_func(rspec_orig =					\
			espda_frame_collect(frameset_mod, cfg, cfg_tag, -1,
					    air_flag, &rspec_rv, &rspec_berv,
					    &rspec_snr600), FRAME_COLLECT, 1);
      */
      rspec_orig =					\
			espda_frame_collect(frameset_mod, cfg, cfg_tag, -1,
					    air_flag, cols, &rspec_rv, &rspec_berv,
					    &rspec_snr600);
        fspec_orig = espda_frame_collect(frameset_mod, cfg, cfg_tag, 0,
					 air_flag, NULL, &fspec_rv, &fspec_berv,
							 &fspec_snr600);
	//cpl_table_dump(fspec_orig, 10000, 10, NULL);
        cpl_table *dll = NULL;
        cpl_array *null_rv = NULL;
        cpl_array *null_berv = NULL;
        cpl_array *null_snr600 = NULL;
	dll = espda_frame_collect(frameset_mod, cfg, cfg_tag,  1, air_flag, NULL,
				  &null_rv, &null_berv, &null_snr600);
        cpl_array_delete(null_rv);
        cpl_array_delete(null_berv);
        cpl_array_delete(null_snr600);

	// If needed, convert the flux density per pixel into flux density per
	// unit wavelength
	/*
	cpl_array *dll_arr = NULL;
	if (dll != NULL) {
	    double *dll_data = cpl_table_get_data_double(dll, "PIXSIZE");
	    dll_arr = cpl_array_new(cpl_table_get_nrow(dll), CPL_TYPE_DOUBLE);
	    cpl_array_copy_data_double(dll_arr, dll_data);
	    for (int i = 1; i < num2d; i++) {
	        cpl_array *dll_temp = cpl_array_new(cpl_table_get_nrow(dll),
						    CPL_TYPE_DOUBLE);
		cpl_array_copy_data_double(dll_temp, dll_data);
		cpl_array_insert(dll_arr, dll_temp,
				 cpl_array_get_size(dll_arr));
		cpl_array_delete(dll_temp);
	    }
	}
	*/
	if (fspec_orig != NULL) {
	      /*
	    if (dll_arr != NULL) {
 	        if (cpl_table_has_column(fspec_orig, "PIXSIZE2") != 1) {
	            cpl_table_new_column(fspec_orig, "PIXSIZE2",
					 CPL_TYPE_DOUBLE);
		}
		cpl_table_copy_data_double(fspec_orig, "PIXSIZE2",
					   cpl_array_get_data_double(dll_arr));
	      */
	  cpl_table_divide_columns(fspec_orig, "FLUX", "PIXSIZE");
	  cpl_table_divide_columns(fspec_orig, "FLUXERR", "PIXSIZE");
	  //cpl_table_divide_columns(fspec_orig, "FLUX", "PIXSIZE");
	  //cpl_table_divide_columns(fspec_orig, "FLUXERR", "PIXSIZE");
		//cpl_table_erase_column(fspec_orig, "PIXSIZE2");
		//}
	  //      cpl_table_multiply_columns(fspec_orig, "FLUX", "EFFICIENCY");
	  //      cpl_table_multiply_columns(fspec_orig, "FLUXERR", "EFFICIENCY");

	  //cpl_table_unselect_all(fspec_orig);
	  //cpl_table_or_selected_int(fspec_orig, "MASKATM", CPL_GREATER_THAN, 0);
	  //cpl_table_erase_selected(fspec_orig);
	    for (int i = 0; i < cpl_table_get_nrow(fspec_orig); i++) {
	        if (cpl_table_get_int(fspec_orig, "MASKATM", i, NULL) > 0) {
		    cpl_table_set_invalid(fspec_orig, "WAVEL", i);
		}
	    }
	}
	//cpl_table_dump(fspec_orig, 10000, 10, NULL);
    }
    cpl_recipeconfig_delete(cfg);


    cpl_table *spec_orig = NULL;
    cpl_array *spec_rv   = NULL;
    cpl_array *spec_snr600 = NULL;

    // Extract redshift, radial velocity, and SNR at 600
    if (fspec_orig != NULL) {
        spec_orig = cpl_table_duplicate(fspec_orig);
        spec_rv   = cpl_array_duplicate(fspec_rv);
        spec_snr600 = cpl_array_duplicate(fspec_snr600);
	if (zem == INV_DOUBLE) {
	    if (drs_flag == 0) {
	        espda_frame_desc(frameset_mod, "SPEC_FLUX_2D", DAS_Z_EM, &zem);
	    }
	    else {
	        if (cpl_frameset_count_tags(frameset_mod, S2D) > 0) {
	            espda_frame_desc(frameset_mod, S2D, Z_EM, &zem);
	        }
	        else if (cpl_frameset_count_tags(frameset_mod, S2D_SKYSUB) > 0){
	            espda_frame_desc(frameset_mod, S2D_SKYSUB, Z_EM, &zem);
	        }
            }
	}
    }
    else {
        spec_orig = cpl_table_duplicate(rspec_orig);
        spec_rv   = cpl_array_duplicate(rspec_rv);
        spec_snr600 = cpl_array_duplicate(rspec_snr600);
	if (zem == INV_DOUBLE) {
	    if (drs_flag == 0) {
                espda_frame_desc(frameset_mod, "S1D", DAS_Z_EM, &zem);
            }
            else {
                espda_frame_desc(frameset_mod, "S1D", Z_EM, &zem);
            }
	}
    }
    //zem = 1.713;

    if (rv != INV_DOUBLE) {
        spec_rv = user_rv;
    }

    cpl_array_delete(rspec_rv);
    cpl_array_delete(fspec_rv);
    cpl_array_delete(rspec_snr600);
    cpl_array_delete(fspec_snr600);

    double rv_mean = INV_DOUBLE;
    double snr600_rms = INV_DOUBLE;
    if (cpl_array_has_valid(spec_rv) == 1) {
	rv_mean = cpl_array_get_mean(spec_rv);
    }
    if (cpl_array_has_valid(spec_snr600) == 1) {
        cpl_array_power(spec_snr600, 2);
        snr600_rms = sqrt(cpl_array_get_mean(spec_snr600));
    }

    if (wavel_min < cpl_table_get_column_min(spec_orig, "WAVEL")) {
        wavel_min = cpl_table_get_column_min(spec_orig, "WAVEL");
    }
    if (wavel_max > cpl_table_get_column_max(spec_orig, "WAVEL")) {
        wavel_max = cpl_table_get_column_max(spec_orig, "WAVEL");
    }

/*
 * Only for stellar spectra, wavelength are shifted to the rest frame.
 */

    if (zem == 0.0) {
        cpl_msg_info(cpl_func, "*** Shift wavelengths to the rest frame ***");
        espda_util_func(espda_spec_shift(spec_orig, spec_rv), SPEC_SHIFT, 0);
    }
    cpl_array_delete(spec_rv);

/*
 * Spectra are coadded by equalising and rebinning the flux. Both operation are
 * optional: equalisation is controlled by parameter equal-flag, while rebinning
 * is controlled by parameter rebin-flag and is done only when a 2D frame is
 * given in input. An equalized spectrum table and a rebinned spectrum table
 * are consequently produced.
 */

    cpl_table *fspec_pre = NULL;
    cpl_table *rspec_pre = NULL;
    cpl_table *fprof = NULL;
    cpl_table *rprof = NULL;
    if (rprof_multimr != NULL) {
        rprof = cpl_table_duplicate(rprof_multimr);
    }
    if (rebin_flag == 1) {

        // Optional equalization
        cpl_table *spec_temp = NULL;
	int v_equal = 1;
        if (equal_flag == 1) {
            cpl_msg_info(cpl_func, "*** Coadd spectra (with equalization) ***");
            //espda_util_func(spec_temp = espda_spec_equal(spec_orig, equal_ref,
	    //	wavel_min, wavel_max), SPEC_EQUAL, 1);
        }
        else {
            cpl_msg_info(cpl_func, "*** Coadd spectra (without equalization) "
                "***");
	    equal_ref = 0;
	    v_equal = 0;
            //spec_temp = cpl_table_duplicate(spec_orig);
        }
        //cpl_table_dump(spec_orig, 0, 10, NULL);
	espda_util_func(spec_temp = espda_spec_equal(spec_orig, equal_ref,
						     wavel_min, wavel_max),
			SPEC_EQUAL, v_equal);

	// Rejection of bad pixels (taking the information from DRS)
	cpl_table_unselect_all(spec_temp);
	//cpl_table_or_selected_int(spec_temp, "QUALDATA", CPL_NOT_EQUAL_TO, 0);
	cpl_table_erase_selected(spec_temp);
	cpl_table_unselect_all(spec_temp);
        cpl_table_delete(spec_orig);

        // Optional rebinning
        if (fspec_orig != NULL) {
            fspec_pre = cpl_table_duplicate(spec_temp);
            espda_util_func(rspec_pre = espda_spec_rebin(spec_temp, kappa,
                wavel_min, wavel_max, vel_step), SPEC_REBIN, 1);
        }
        else {
          espda_util_func(rspec_pre = espda_spec_rebin(spec_temp, kappa,
              wavel_min, wavel_max, vel_step), SPEC_REBIN, 1);
        }
        cpl_table_delete(spec_temp);

	// Computation of the instrument profile (only for quasar spectra)
	if (zem != 0.0 && rprof == NULL) {


	    espda_print_verb(0, cpl_msg_info(cpl_func, "*** Creating and "
                "interpolating instrument profile ***"));
	    espda_print_verb(0, cpl_msg_info(cpl_func, "Creating the "
                "instrument profile..."));
            cpl_table *inst_prof = espda_model_inst(MODEL_SIZE, model_resol,
                MODEL_KAPPA);
	    espda_print_verb(0, cpl_msg_info(cpl_func, "Interpolating the "
	        "instrument profile..."));
            cpl_ensure_code(espda_interp_inst(rspec_pre, inst_prof, MODEL_SIZE,
                model_resol, MODEL_KAPPA, &rprof) == CPL_ERROR_NONE,
                cpl_error_get_code());
	    cpl_table_delete(inst_prof);
        }
    }

/*
 * Output is exported. QC parameters are saved. If 2D spectra were given in
 * input, an order spectrum with all their original pixels is exported. If an
 * equalized spectrum has been produced, it is also exported.
 */

    cpl_msg_info(cpl_func, "*** Export products into frames ***");

    // Create property lists
    time_t c_time;
    char* c_time_string;
    c_time = time(NULL);
    c_time_string = ctime(&c_time);

    cpl_propertylist *list = cpl_propertylist_new();
    cpl_propertylist_append_string(list, DAS_VERSION, VERSION);
    cpl_propertylist_append_string(list, DAS_DATE, c_time_string);
    cpl_propertylist_append_double(list, DAS_RV_MEAN, rv_mean);
    cpl_propertylist_append_double(list, DAS_SNR600_RMS, snr600_rms);
    cpl_propertylist_append_double(list, DAS_Z_EM, zem);
    cpl_propertylist_append_double(list, DAS_VEL_STEP, vel_step);
    cpl_propertylist *flist_orig = cpl_propertylist_duplicate(list);
    cpl_propertylist *rlist_orig = cpl_propertylist_duplicate(list);
    cpl_propertylist *flist_pre  = cpl_propertylist_duplicate(list);
    cpl_propertylist *rlist_pre  = cpl_propertylist_duplicate(list);
    cpl_propertylist *fprof_list = cpl_propertylist_duplicate(list);
    cpl_propertylist *rprof_list = cpl_propertylist_duplicate(list);
    cpl_propertylist_delete(list);

    // Create file names
    espda_frame_name(fname_orig, FSPEC_ORIG);
    espda_frame_name(rname_orig, RSPEC_ORIG);
    espda_frame_name(fname_pre,  FSPEC_PRE);
    espda_frame_name(rname_pre,  RSPEC_PRE);
    espda_frame_name(fprof_name, FPROF);
    espda_frame_name(rprof_name, RPROF);

    // Save frames
    espda_dfs_call(0, espda_frame_export(frameset, parlist, fspec_orig,
        flist_orig, "espda_coadd_spec", FSPEC_ORIG, fname_orig));
    espda_dfs_call(0, espda_frame_export(frameset, parlist, rspec_orig,
        rlist_orig, "espda_coadd_spec", RSPEC_ORIG, rname_orig));
    espda_dfs_call(0, espda_frame_export(frameset, parlist, fspec_pre,
        flist_pre,  "espda_coadd_spec", FSPEC_PRE,  fname_pre));
    espda_dfs_call(0, espda_frame_export(frameset, parlist, rspec_pre,
        rlist_pre,  "espda_coadd_spec", RSPEC_PRE,  rname_pre));
    espda_dfs_call(0, espda_frame_export(frameset, parlist, fprof,
        fprof_list, "espda_coadd_spec", FPROF,      fprof_name));
    espda_dfs_call(0, espda_frame_export(frameset, parlist, rprof,
        rprof_list, "espda_coadd_spec", RPROF,      rprof_name));

    cpl_frameset_delete(frameset_mod);
    cpl_msg_info(cpl_func, "*** Success!! ***");

    return 0;
}

/**@}*/
