/* $Id: espda_create_linelist.c 288162 2024-03-28 08:45:16Z amodigli $
 *
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: amodigli $
 * $Date: 2024-03-28 08:45:16 +0000 (Thu, 28 Mar 2024) $
 * $Revision: 288162 $
 * $Name:  $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#include <string.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>

#include "espda_check.h"
#include "espda_dfs.h"
#include "espda_eqwidth_lib.h"
#include "espda_frame.h"
#include "espda_line.h"
//#include "espda_mask.h"
#include "espda_param.h"
#include "espda_print.h"
#include "espda_spec.h"
#include "espda_struct.h"
#include "espda_wrap.h"

/*-----------------------------------------------------------------------------
                             Plugin registration
 -----------------------------------------------------------------------------*/

int cpl_plugin_get_info(cpl_pluginlist * list);

/*-----------------------------------------------------------------------------
                            Private function prototypes
 -----------------------------------------------------------------------------*/

static int espda_create_linelist_create(cpl_plugin *);
static int espda_create_linelist_exec(cpl_plugin *);
static int espda_create_linelist_destroy(cpl_plugin *);
static int espda_create_linelist(cpl_frameset *, const cpl_parameterlist *);

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static char espda_create_linelist_description[] =
"This recipe scans a spectrum tables and creates a list of found absorption "
"lines. If the spectrum is a stellar spectrum and a reference line list is "
"provided, the lines are also identified.\n"
"\n"
"INPUT FRAMES                             FORMAT  # EXPECTED  TAGS\n"
"Rebinned spectrum                        SPEC    1           RSPEC\n"
  //"list of ionic transitions                ION     0 or 1      STAR_ION\n"
"\n"
"OUTPUT FRAMES                            FORMAT  # PRODUCED  TAGS             "
"NOTES\n"
"List of absorption lines                 LINE    1           FLINE_PRE\n"
  //"list of identified absorption lines      LINE    0 or 1      FLINE_IDEN       "
  //"produced only if STAR_ION is provided\n"
  //"list of non-identified absorption lines  LINE    0 or 1      FLINE_REJ        "
  //"produced only if STAR_ION is provided\n"
"Smoothed spectrum                        SPEC    1           RSPEC_SMOOTH"
;

/*-----------------------------------------------------------------------------
                                Function code
 -----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module.
  @param    list    the plugin list
  @return   0 if everything is ok, 1 otherwise
  @note     Only this function is exported

  Create the recipe instance and make it available to the application using the
  interface.
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist *list)
{
    cpl_recipe *recipe = cpl_calloc(1, sizeof *recipe );
    cpl_plugin *plugin = &recipe->interface;

    if (cpl_plugin_init(plugin, CPL_PLUGIN_API, ESPDA_BINARY_VERSION,
                        CPL_PLUGIN_TYPE_RECIPE, "espda_create_linelist",
                        "Create a list of absorption lines found in a spectrum",
                        espda_create_linelist_description, "Guido Cupani",
                        PACKAGE_BUGREPORT, espda_dfs_license(),
                        espda_create_linelist_create,
                        espda_create_linelist_exec,
                        espda_create_linelist_destroy)) {
        cpl_msg_error(cpl_func, "Plugin initialization failed");
        (void)cpl_error_set_where(cpl_func);
        return 1;
    }

    if (cpl_pluginlist_append(list, plugin)) {
        cpl_msg_error(cpl_func, "Error adding plugin to list");
        (void)cpl_error_set_where(cpl_func);
        return 1;
    }

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options
  @param    plugin  the plugin
  @return   0 if everything is ok

  Defining the command-line/configuration parameters for the recipe.
 */
/*----------------------------------------------------------------------------*/
static int espda_create_linelist_create(cpl_plugin *plugin)
{
    cpl_recipe    *recipe;
    cpl_parameter *param;

    /* Do not create the recipe if an error code is already set */
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());
        return (int)cpl_error_get_code();
    }

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    /* Create the parameters list in the cpl_recipe object */
    recipe->parameters = cpl_parameterlist_new();
    if (recipe->parameters == NULL) {
        cpl_msg_error(cpl_func, "Parameter list allocation failed");
        cpl_ensure_code(0, (int)CPL_ERROR_ILLEGAL_OUTPUT);
    }

    /* Fill the parameters list */
    /*
    param = cpl_parameter_new_value("espda.espda_create_linelist.hwidth",
                CPL_TYPE_DOUBLE, "Halfwidth of the spectral chunks [nm] (for "
                "stellar spectra): [1.0e-2, 3.0]",
                "espda.espda_create_linelist", 1.0);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "hwidth");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_create_linelist.overlap",
                CPL_TYPE_DOUBLE, "Size of the overlapping region between "
                "adjacent chunks, as a fraction of the chunk width (for "
                "stellar spectra): [0.0, 0.7]", "espda.espda_create_linelist",
                0.5);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "overlap");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_create_linelist.iter",
                CPL_TYPE_INT, "Number of iterations to fit the local "
                "continuum (for stellar spectra); number of iterations to "
		"smooth the spectrum (for QSO spectra): [1, 1e2]",
                "espda.espda_create_linelist", 10);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "iter");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_create_linelist.thres",
                CPL_TYPE_DOUBLE, "Normalized threshold for line acceptance as "
                "a fraction of the continuum level (for stellar spectra): "
                "[1.0e-10, 1.0]", "espda.espda_create_linelist", 2.0e-2);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "thres");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_create_linelist.resol",
                CPL_TYPE_DOUBLE, "Minimum accepted separation between two "
                "lines [nm] (for stellar spectra): [1.0e-3, 1.0e-1]",
                "espda.espda_create_linelist", 1.0e-2);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "resol");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_create_linelist.smwidth",
                CPL_TYPE_INT, "Number of pixels in the boxcar to be averaged "
                "to fit the continuum (for stellar spectra): [1, 1e2]",
                "espda.espda_create_linelist", 20);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "smwidth");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_create_linelist.wdiff",
                CPL_TYPE_DOUBLE, "Maximum allowed difference between observed "
                "and laboratory wavelengths [nm] (for stellar spectra): "
                "[1.0e-3, 1.0e-1]", "espda.espda_create_linelist", 0.02);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "wdiff");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_create_linelist.kappa",
                CPL_TYPE_INT, "Threshold for kappa-sigma clipping to be used "
                "in rebinning (for QSO spectra): [3, 5]",
                "espda_create_linelist", 3);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "kappa");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);
    */

    param = cpl_parameter_new_value("espda.espda_create_linelist.vel_step_min",
				    CPL_TYPE_DOUBLE,
				    "Minimum velocity step to undersample the spectra, corresponding "
            "to the minimum FWHM of lines to be detected "
				    "[km/s] (for QSO spectra): [0.0, 10.0]",
				    "espda.espda_create_linelist",
				    2.0);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "vel-step-min");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_create_linelist.vel_step_max",
				    CPL_TYPE_DOUBLE,
				    "Maximum velocity step to undersample the spectra, corresponding "
            "to the maximum FWHM of lines to be detected "
				    "[km/s] (for QSO spectra): [10.0, 1000.0]",
				    "espda.espda_create_linelist",
				    200.0);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "vel-step-max");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_create_linelist.accept",
				    CPL_TYPE_DOUBLE,
				    "Distance between a flux minimum and the "
				    "adjacent maxima for the minimum to be "
				    "accepted as an absorption line, as a "
				    "multiple of the flux error at the "
				    "continuum level (for QSO spectra): "
				    "[3.0, 1.0e2]",
				    "espda.espda_create_linelist",
				    5.0);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "accept");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_create_linelist.add",
				    CPL_TYPE_STRING,
				    "Wavelengths of additional, manually-selected lines [nm]. "
            "Wavelengths must be separated by commas, with a trailing comma "
            "after the last wavelength.",
				    "espda.espda_create_linelist",
				    "0");
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "add");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_create_linelist_exec(cpl_plugin *plugin)
{
    int             recipe_status;

    cpl_recipe     *recipe;
    cpl_errorstate  initial_errorstate = cpl_errorstate_get();

    /* Return immediately if an error code is already set */
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());
        return (int)cpl_error_get_code();
    }

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    /* Verify parameter and frame lists */
    if (recipe->parameters == NULL) {
        cpl_msg_error(cpl_func, "Recipe invoked with NULL parameter list");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }
    if (recipe->frames == NULL) {
        cpl_msg_error(cpl_func, "Recipe invoked with NULL frame set");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Invoke the recipe */
    recipe_status = espda_create_linelist(recipe->frames, recipe->parameters);

    /* Ensure DFS-compliance of the products */
    if (cpl_dfs_update_product_header(recipe->frames)) {
        if (!recipe_status) recipe_status = (int)cpl_error_get_code();
    }

    if (!cpl_errorstate_is_equal(initial_errorstate)) {
        /* Dump the error history since recipe execution start.
           At this point the recipe cannot recover from the error */
        cpl_errorstate_dump(initial_errorstate, CPL_FALSE, NULL);
    }

    return recipe_status;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_create_linelist_destroy(cpl_plugin *plugin)
{
    cpl_recipe * recipe;

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    cpl_parameterlist_delete(recipe->parameters);

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    frameset   the frames list
  @param    parlist    the parameters list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_create_linelist(cpl_frameset            *frameset,
			         const cpl_parameterlist *parlist)
{
    extern int v_level;

/*
 * Frames are loaded and checked. Redshift is also loaded.
 */
	/* check for invalid input */
	if(espda_files_dont_exist(frameset) != CPL_ERROR_NONE) {
		return CPL_ERROR_BAD_FILE_FORMAT;
	}
    cpl_msg_info(cpl_func, "*** Load frames ***");
    cpl_ensure_code(espda_dfs_groups(frameset) == CPL_ERROR_NONE,
                    cpl_error_get_code());
    cpl_recipeconfig *cfg = cpl_recipeconfig_new();
    char *cfg_tag = RSPEC;
    cpl_recipeconfig_set_tag  (cfg, cfg_tag, 0, 1);
    //cpl_recipeconfig_set_input(cfg, cfg_tag, STAR_ION,   0, 1);
    espda_dfs_call(0, espda_check_name(frameset, cfg, cfg_tag));
    espda_dfs_call(0, espda_check_tag (frameset, cfg, cfg_tag));
    espda_dfs_call(0, espda_check_num (frameset, cfg, cfg_tag));
    cpl_table *rspec    = espda_frame_load(frameset, cfg, cfg_tag, -1);
    //cpl_table *line_ref = espda_frame_load(frameset, cfg, cfg_tag,  0);
    cpl_recipeconfig_delete(cfg);

/*
 * Redshift and velocity step of the spectrum are loaded.
 */

    double zem, vel_step;
    espda_dfs_call(0, espda_frame_desc(frameset, RSPEC,
        DAS_Z_EM, &zem));
    espda_dfs_call(0, espda_frame_desc(frameset, RSPEC,
        DAS_VEL_STEP, &vel_step));

/*
 * Parameters are loaded and checked.
 */

    cpl_msg_info(cpl_func, "*** Load parameters ***");
    /*
    espda_param_double("hwidth",   hwidth,
        cpl_parameterlist_get_first_const(parlist), 1.0e-2,  10.0);
    espda_param_double("overlap",  overlap,
        cpl_parameterlist_get_next_const(parlist),  0.0,     0.7);
    espda_param_int   ("iter",     iter,
        cpl_parameterlist_get_next_const(parlist),  1,       1e2);
    espda_param_double("thres",    thres,
        cpl_parameterlist_get_next_const(parlist),  1.0e-10, 1.0e2);
    espda_param_double("resol",    resol,
        cpl_parameterlist_get_next_const(parlist),  1.0e-3,  1.0e-1);
    espda_param_int   ("smwidth",  smwidth,
        cpl_parameterlist_get_next_const(parlist),  1,       1e2);
    espda_param_double("wdiff",    wdiff,
        cpl_parameterlist_get_next_const(parlist),  1.0e-3,  1.0e-1);
    espda_param_int   ("kappa",    kappa,
        cpl_parameterlist_get_next_const(parlist),  3,       5);
    */
    espda_param_double("vel-step-min", vel_step_min,
        cpl_parameterlist_get_first_const(parlist),  0.0,     100.0);
    espda_param_double("vel-step-max", vel_step_max,
        cpl_parameterlist_get_next_const(parlist),  10.0,    1000.0);
    espda_param_double("accept", accept,
        cpl_parameterlist_get_next_const(parlist),  1.0,     1.0e2);
    espda_param_string("add",      add,
        cpl_parameterlist_get_next_const(parlist));

    cpl_msg_info(cpl_func, "*** Creating a line list ***");

    //double wavel_max = cpl_table_get_double(rspec, "WAVEL",
    //    cpl_table_get_nrow(rspec) - 1, NULL);
    double wavel_min = cpl_table_get_column_min(rspec, "WAVEL");
    double wavel_max = cpl_table_get_column_max(rspec, "WAVEL");

    cpl_table *rspec_smooth = NULL;
    cpl_table *fline_pre    = NULL;
    cpl_table *fline_iden   = NULL;
    cpl_table *fline_rej    = NULL;
    cpl_size   fline_pre_num = 0;
    //cpl_size   fline_iden_num;
    //cpl_size   fline_rej_num;

/*
 * Stellar spectrum is processed with ARES functions. A local continuum is
 * fitted to spectral chunks of increasing wavelength and fixed wavelength
 * range, possibly overlapping. Lines are detected from the derivatives of the
 * smoothed, continuum-normalized flux, and identified by comparison with
 * known transitions.
 *
 * NB: Currently disabled
 */

    // Parameters: hwidth, overlap, iter, thres, resol, smwidth, wdiff
    if (zem == 0.0) {

      /*
        // Chunk definition
        double pos_start = wavel_min + hwidth;
        double pos_end   = wavel_max - hwidth;
        if (hwidth >= 0.5 * (wavel_max - wavel_min)) {
            hwidth = 0.5 * (wavel_max - wavel_min) -
                cpl_table_get_double(rspec, "PIXSIZE", 0, NULL);
            cpl_msg_warning(cpl_func, "Chunks size is too large w.r.t. the "
                "wavelength range of the rspectrum! Parameter hwidth "
                "redefined to %f.", hwidth);
        }
        if (wavel_min < cpl_table_get_column_min(rspec, "WAVEL")) {
            pos_start = cpl_table_get_column_min(rspec, "WAVEL") + hwidth;
        }
        if (wavel_max > cpl_table_get_column_max(rspec, "WAVEL")) {
            pos_end   = cpl_table_get_column_max(rspec, "WAVEL") - hwidth;
        }
        double pos_step = hwidth * 2.0 * (1.0 - overlap);

        int i    = 0;
        int flag = 0;

        // Line detection
        cpl_msg_info(cpl_func, "Detect lines in chunks...");
        for (double pos = pos_start; pos < pos_end; pos += pos_step) {
            i++;
            double pos_min = pos - hwidth * (1.0 - overlap);
            double pos_max = pos + hwidth * (1.0 - overlap);
            double cont_rejt = 0.995;
            cpl_table *rspec_chunk  = NULL;
            espda_dfs_call(0, espda_select_chunk(rspec, hwidth, pos,
                &rspec_chunk));
            espda_struct_check(rspec_chunk, espda_struct_sort_old(rspec_chunk,
                "WAVEL", FALSE));
            espda_dfs_call(0, espda_fit_lcont(rspec_chunk, cont_rejt, iter));
            cpl_table *fline_chunk = NULL;
            espda_dfs_call(0, espda_det_line(rspec_chunk, thres, resol, smwidth,
                &fline_chunk));
            cpl_table *temp = NULL, *part = NULL;
            espda_struct_check(fline_chunk, espda_struct_extract_old(fline_chunk,
                temp, "WAVEL", CPL_TYPE_DOUBLE, CPL_GREATER_THAN, pos_min));
            espda_struct_check(temp, espda_struct_extract_old(temp, part, "WAVEL",
                CPL_TYPE_DOUBLE, CPL_NOT_GREATER_THAN, pos_max));
            cpl_table_delete(fline_chunk);
            if (flag == 0 && cpl_table_get_nrow(part) != 0) {
                fline_pre = cpl_table_duplicate(part);
                flag = 1;
            }
            else {
                cpl_table_insert(fline_pre, part, fline_pre_num);
            }
            fline_pre_num += cpl_table_get_nrow(part);
            cpl_table_delete(rspec_chunk);
            cpl_table_delete(part);
        }

        espda_util_safe(fline_pre, espda_util_func(espda_line_iden(fline_pre,
            line_ref, wdiff, &fline_iden, &fline_rej, &fline_iden_num,
            &fline_rej_num), LINE_IDEN, 0));
      */;
    }

/*
 * QSO rspectrum is processed. Lines are detected after undersampling the
 * spectrum. The line list is then cleaned from duplicated lines.
 */

    // Parameters: kappa, vel_step, accept
    else {

      //double vel_min = 2.0;
        for (double vel_tmp = vel_step_max; vel_tmp >= vel_step_min;
	     vel_tmp=espda_max(vel_tmp-10.0, vel_step_min)) {

	  //double fact = (iter - 1 - i) / ((double)iter - 1);
	  // double vel_tmp = vel_min + (vel_step - vel_min) * fact;

	    cpl_table *fline_tmp = NULL;

	    // Spectrum undersampling
	    if (vel_tmp > 0) {
	        cpl_msg_info(cpl_func, "Smoothing length: %3.1f km s-1...",
			     vel_tmp);
		rspec_smooth = espda_spec_rebin(rspec, 3.0, wavel_min,
						wavel_max, vel_tmp);
	    }
	    else {
		cpl_msg_info(cpl_func, "Non-smoothed spectrum...");
	        rspec_smooth = cpl_table_duplicate(rspec);
	    }

	    // Rejection threshold depend on vel_tmp, because smoothing
	    // increases the SNR
	    double kappa = accept * sqrt(vel_tmp/vel_step);

	    // Line detection
	    espda_util_func(fline_tmp = espda_line_detect(rspec_smooth, rspec,
							  kappa),
			    LINE_DETECT, 1);
	    if (vel_tmp == vel_step_max) {
  	        fline_pre = cpl_table_duplicate(fline_tmp);
	    }
	    else {
	        // If fline_pre is empty, cpl_table_insert does not work
   	        if (cpl_table_get_nrow(fline_pre) == 0) {
		    fline_pre = cpl_table_duplicate(fline_tmp);
	        }
	        cpl_table_insert(fline_pre, fline_tmp, 0);
	    }
	    //double vel_clean = espda_min(vel_tmp*0.5, 20.0);
	    double vel_clean = espda_max(vel_tmp, 5.0);
	    //double vel_clean = vel_tmp*1.0;
	    espda_util_safe(fline_pre, espda_util_func(espda_line_clean(
	    	fline_pre, vel_clean), LINE_CLEAN, 1));

	    if (vel_tmp == vel_step_min) {
		break;
	    }
	}
    }

/*
 * Manually-selected lines are added to the table. The string parameter 'add' is
 * converted into line wavelengths.
 */

    espda_util_func(espda_line_add(fline_pre, add), LINE_ADD, 0);

/*
 * Products are exported into frames.
 */

    cpl_msg_info(cpl_func, "*** Export products into frames ***");

    // Create property lists
    cpl_propertylist *list = cpl_propertylist_new();
    cpl_propertylist_append_int(list, QC_DET_LINE_NUM, fline_pre_num);
    cpl_propertylist *flist_pre    = cpl_propertylist_duplicate(list);
    cpl_propertylist *flist_iden   = cpl_propertylist_duplicate(list);
    cpl_propertylist *flist_rej    = cpl_propertylist_duplicate(list);
    cpl_propertylist *rlist_smooth = cpl_propertylist_duplicate(list);
    cpl_propertylist_delete(list);

    // Create file names
    espda_frame_name(fname_pre,    FLINE_PRE);
    espda_frame_name(fname_iden,   FLINE_IDEN);
    espda_frame_name(fname_rej,    FLINE_REJ);
    espda_frame_name(rname_smooth, RSPEC_SMOOTH);

    // Save frames
    espda_dfs_call(0, espda_frame_export(frameset, parlist, fline_pre,
        flist_pre, "espda_create_linelist", FLINE_PRE,  fname_pre));
    espda_dfs_call(0, espda_frame_export(frameset, parlist, fline_iden,
        flist_iden,   "espda_create_linelist", FLINE_IDEN, fname_iden));
    espda_dfs_call(0, espda_frame_export(frameset, parlist, fline_rej,
        flist_rej,    "espda_create_linelist", FLINE_REJ,  fname_rej));
    espda_dfs_call(0, espda_frame_export(frameset, parlist, rspec_smooth,
        rlist_smooth, "espda_create_linelist", RSPEC_SMOOTH, rname_smooth));

    cpl_table_delete(rspec);
    //cpl_table_delete(line_ref);

    cpl_msg_info(cpl_func, "*** Success!! ***");
    return 0;
}

/**@}*/
