/* 
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Id: espda_fit_line.c 288163 2024-03-28 08:47:43Z amodigli $
 * $Author: amodigli $
 * $Date: 2024-03-28 08:47:43 +0000 (Thu, 28 Mar 2024) $
 * $Revision: 288163 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>

#include "espda_check.h"
#include "espda_dfs.h"
#include "espda_fit.h"
#include "espda_frame.h"
#include "espda_param.h"
#include "espda_util.h"
#include "espda_voigt.h"
#include "espda_wrap.h"

/*-----------------------------------------------------------------------------
                             Plugin registration
 -----------------------------------------------------------------------------*/

int cpl_plugin_get_info(cpl_pluginlist * list);

/*-----------------------------------------------------------------------------
                            Private function prototypes
 -----------------------------------------------------------------------------*/

static int espda_fit_line_create(cpl_plugin *);
static int espda_fit_line_exec(cpl_plugin *);
static int espda_fit_line_destroy(cpl_plugin *);
static int espda_fit_line(cpl_frameset *, const cpl_parameterlist *);

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static char espda_fit_line_description[] =
"This recipe fits a set of Voigt profiles to a list of absorption lines. Lines "
"close to each other are fitted together.\n"
"\n"
"INPUT FRAMES                        FORMAT     # EXPECTED  TAGS         "
"NOTES\n" 
"Full spectrum with continuum        SPEC       1           FSPEC_CONT\n"
"Rebinned spectrum with continuum    SPEC       1           RSPEC_CONT\n" 
"Individual absorption system        LINE       0 or 1      SLINE_IDEN\n"
"List of identified lines            LINE       0 or 1      FLINE_IDEN\n"
"List of non-identified lines        LINE       0 or 1      FLINE_REJ    not "
"considered if FLINE_IDEN is present\n"
"Full instrumental profile           PROF       0 or 1      FPROF\n"
"Rebinned instrumental profile       PROF       0 or 1      RPROF\n"
"List of ionic transitions           ION        1           QSO_ION\n"
"Reduced list of ionic transitions   ION        1           QSO_ION_RED\n"
"Table with Voigt function           VOIGT      0 or 1      VOIGT_FUNC   "
"recommended (speeds up computation)\n"
"\n"
"OUTPUT FRAMES                       FORMAT     # PRODUCED  TAGS         "
"NOTES\n" 
"Full spectrum with Voigt guess      SPEC       0 or 1      FSPEC_GUESS  "
"produced only if --fit-flag=FALSE\n"
"Rebinned spectrum with Voigt guess  SPEC       0 or 1      RSPEC_GUESS  "
"produced only if --fit-flag=FALSE\n"
"List of Voigt-guessed lines         LINE       0 or 1      SLINE_GUESS  "
"produced only if --fit-flag=FALSE\n"
"Full spectrum with Voigt fit        SPEC       0 or 1      FSPEC_VOIGT  "
"produced only if --fit-flag=TRUE\n"
"Rebinned spectrum with Voigt fit    SPEC       0 or 1      RSPEC_VOIGT  "
"produced only if --fit-flag=TRUE\n"
"List of Voigt-fitted lines          LINE       0 or 1      SLINE_VOIGT  "
"produced only if --fit-flag=TRUE\n"
;

/*-----------------------------------------------------------------------------
                                Function code
 -----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module. 
  @param    list  the plugin list
  @return   0 if everything is ok, 1 otherwise
  @note     Only this function is exported

  Create the recipe instance and make it available to the application using the 
  interface. 
 */
/*----------------------------------------------------------------------------*/

int cpl_plugin_get_info(cpl_pluginlist *list) {
    cpl_recipe *recipe = cpl_calloc(1, sizeof *recipe );
    cpl_plugin *plugin = &recipe->interface;
    if (cpl_plugin_init(plugin, CPL_PLUGIN_API, ESPDA_BINARY_VERSION,
                        CPL_PLUGIN_TYPE_RECIPE, "espda_fit_line", "Fit a Voigt "
                        "profile to absorption lines", 
                        espda_fit_line_description, "Guido Cupani",
                        PACKAGE_BUGREPORT, espda_dfs_license(),
                        espda_fit_line_create, espda_fit_line_exec,
                        espda_fit_line_destroy)) {    
        cpl_msg_error(cpl_func, "Plugin initialization failed");
        (void)cpl_error_set_where(cpl_func);                          
        return 1;                                               
    }                                                    
    if (cpl_pluginlist_append(list, plugin)) {                 
        cpl_msg_error(cpl_func, "Error adding plugin to list");
        (void)cpl_error_set_where(cpl_func);                         
        return 1;                                              
    }                                                          
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options    
  @param    plugin  the plugin
  @return   0 if everything is ok

  Defining the command-line/configuration parameters for the recipe.
 */
/*----------------------------------------------------------------------------*/
static int espda_fit_line_create(cpl_plugin *plugin) {
    cpl_recipe *recipe;                                               
    cpl_parameter *param;
                                                                     
    // Do not create the recipe if an error code is already set     
    if (cpl_error_get_code() != CPL_ERROR_NONE) {                      
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());      
        return (int)cpl_error_get_code();                              
    }                                                                  
                                                                       
    if (plugin == NULL) {                                              
        cpl_msg_error(cpl_func, "Null plugin");                        
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                 
    }                                                                  
                                                                       
    // Verify plugin type                                           
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {       
        cpl_msg_error(cpl_func, "Plugin is not a recipe");             
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);              
    }                                                                  
                                                                       
    // Get the recipe                                               
    recipe = (cpl_recipe *)plugin;                                     
                                                                       
    // Create the parameters list in the cpl_recipe object          
    recipe->parameters = cpl_parameterlist_new();                      
    if (recipe->parameters == NULL) {                                  
        cpl_msg_error(cpl_func, "Parameter list allocation failed");   
        cpl_ensure_code(0, (int)CPL_ERROR_ILLEGAL_OUTPUT);             
    }                                                                  

    // Fill the parameters list
    param = cpl_parameter_new_value(FIT_LINE_HWIDTH,
				    CPL_TYPE_DOUBLE, 
				    "Halfwidth for line grouping [nm]: "
				    "[2.5e-2, 10.0]", 
				    "espda.espda_fit_line",
				    0.1);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "hwidth");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value(FIT_LINE_PAR_RANGE,
				    CPL_TYPE_STRING, 
				    "Range of line parameters",
				    "espda.espda_fit_line",
				    "1e-4,10.0,22.0,2.0,100.0,0.0,0.0");
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "par-range");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value(FIT_LINE_EDIT_FILE,
				    CPL_TYPE_STRING, 
				    "Name of the file with instructions to "
				    "edit line table (used by ESOReflex only)",
				    "espda.espda_fit_line",
				    INV_STRING);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "edit-file");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value(FIT_LINE_ADD,
				    CPL_TYPE_STRING, 
				    "Wavelengths of additional, "
				    "manually-selected lines", 
				    "espda.espda_fit_line", "0");
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "add");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value(FIT_LINE_FIT_FLAG,
				    CPL_TYPE_BOOL, 
				    "Flag to fit the lines: ['TRUE', 'FALSE']",
				    "espda.espda_fit_line",
				    "TRUE");
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "fit-flag");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value(FIT_LINE_ITER_MAX,
				    CPL_TYPE_INT, 
				    "Maximum number of iterations: [1, 1e3]",
				    "espda.espda_fit_line",
				    10);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "iter-max");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value(FIT_LINE_GRP_SEL,
				    CPL_TYPE_INT, 
				    "Group of lines selected for fitting: "
				    "[0, number of groups], 0 meaning all "
				    "group",
				    "espda.espda_fit_line",
				    0);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "grp-sel");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value(FIT_LINE_MODEL_RESOL,
				    CPL_TYPE_DOUBLE, 
				    "Resolution of the instrument model "
				    "[1.0e3, 1.0e6]", 
				    "espda.espda_fit_line",
				    140000.0);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "model-resol");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/

static int espda_fit_line_exec(cpl_plugin *plugin) {
    int  recipe_status;

    cpl_recipe *recipe;                                                   
    cpl_errorstate  initial_errorstate = cpl_errorstate_get();              
                                                                           
    // Return immediately if an error code is already set                
    if (cpl_error_get_code() != CPL_ERROR_NONE) {                          
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",    
                      cpl_func, __LINE__, cpl_error_get_where());          
        return (int)cpl_error_get_code();                                  
    }                                                                      
                                                                           
    if (plugin == NULL) {                                                  
        cpl_msg_error(cpl_func, "Null plugin");                            
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                     
    }                                                                      
                                                                           
    // Verify plugin type                                                
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {           
        cpl_msg_error(cpl_func, "Plugin is not a recipe");                 
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);                  
    }                                                                      
                                                                           
    // Get the recipe                                                    
    recipe = (cpl_recipe *)plugin;                                         
                                                                           
    // Verify parameter and frame lists                                  
    if (recipe->parameters == NULL) {                                      
        cpl_msg_error(cpl_func, "Recipe invoked with NULL parameter list");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                     
    }                                                                      
    if (recipe->frames == NULL) {                                          
        cpl_msg_error(cpl_func, "Recipe invoked with NULL frame set");     
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                     
    }                                                                      
                                                                           
    // Invoke the recipe                                                 
    recipe_status = espda_fit_line(recipe->frames, recipe->parameters);
                                                                           
    // Ensure DFS-compliance of the products                             
    if (cpl_dfs_update_product_header(recipe->frames)) {                   
        if (!recipe_status) recipe_status = (int)cpl_error_get_code();
    }                                                                      
                                                                           
    if (!cpl_errorstate_is_equal(initial_errorstate)) {                    
        /* Dump the error history since recipe execution start.            
           At this point the recipe cannot recover from the error */       
        cpl_errorstate_dump(initial_errorstate, CPL_FALSE, NULL);          
    }
    return recipe_status;                                                  
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_fit_line_destroy(cpl_plugin *plugin) {
    cpl_recipe * recipe;                                          
                                                                 
    if (plugin == NULL) {                                         
        cpl_msg_error(cpl_func, "Null plugin");                   
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);            
    }                                                             
                                                                  
    // Verify plugin type                                       
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {  
        cpl_msg_error(cpl_func, "Plugin is not a recipe");        
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);         
    }                                                             
                                                                  
    // Get the recipe                                           
    recipe = (cpl_recipe *)plugin;                                
                                                                  
    cpl_parameterlist_delete(recipe->parameters);             
                                                                  
    return 0;                                                    
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    frameset  the frames list
  @param    parlist  the parameters list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/

static int espda_fit_line(cpl_frameset *frameset, 
			  const cpl_parameterlist *parlist) {
    extern int v_level;
    
    /* check for invalid input */
    if(espda_files_dont_exist(frameset) != CPL_ERROR_NONE) {
    	return CPL_ERROR_BAD_FILE_FORMAT;
    }

    // Load frames
    cpl_msg_info(cpl_func, "*** Load frames ***");
    espda_util_call(0, espda_dfs_groups(frameset));
    int syst_num = cpl_frameset_count_tags(frameset, SLINE_IDEN);
    cpl_recipeconfig *cfg = cpl_recipeconfig_new();
    char *cfg_tag = FSPEC_CONT;
    cpl_recipeconfig_set_tag(cfg, cfg_tag, 0, 1);
    cpl_recipeconfig_set_input(cfg, cfg_tag, RSPEC_CONT, 1, 1);
    cpl_recipeconfig_set_input(cfg, cfg_tag, FLINE_IDEN, 0, 1);
    cpl_recipeconfig_set_input(cfg, cfg_tag, SLINE_IDEN, 0, 1);
    cpl_recipeconfig_set_input(cfg, cfg_tag, FLINE_REJ, 0, 0);
    if (syst_num == 0) {
        cpl_recipeconfig_set_input(cfg, cfg_tag, FLINE_REJ, 0, 1);
    }
    cpl_recipeconfig_set_input(cfg, cfg_tag, QSO_ION, 1, 1);
    cpl_recipeconfig_set_input(cfg, cfg_tag, QSO_ION_RED, 0, 1);
    cpl_recipeconfig_set_input(cfg, cfg_tag, VOIGT_FUNC, 0, 1);
    cpl_recipeconfig_set_input(cfg, cfg_tag, FPROF, 0, 1);
    cpl_recipeconfig_set_input(cfg, cfg_tag, RPROF, 0, 1);
    espda_wrap_check(frameset, cfg, cfg_tag);

    cpl_table *fspec_cont = espda_frame_load(frameset, cfg, cfg_tag, -1);
    cpl_table *rspec_cont = espda_frame_load(frameset, cfg, cfg_tag,  0);
    cpl_table *fline = espda_frame_load(frameset, cfg, cfg_tag, 1);
    cpl_table *sline = NULL;
    if (syst_num != 0) {
        sline = espda_frame_load(frameset, cfg, cfg_tag, 2);
    }
    else {
        sline = espda_frame_load(frameset, cfg, cfg_tag, 3);
    }
    cpl_table *ion = espda_frame_load(frameset, cfg, cfg_tag, 4);
    cpl_table *ion_red = espda_frame_load(frameset, cfg, cfg_tag, 5);
    cpl_table *voigt_func = espda_frame_load(frameset, cfg, cfg_tag, 6);
    cpl_table *fprof = espda_frame_load(frameset, cfg, cfg_tag, 7);
    cpl_table *rprof = espda_frame_load(frameset, cfg, cfg_tag, 8);
    cpl_recipeconfig_delete(cfg);

    // Load parameters
    cpl_msg_info(cpl_func, "*** Load parameters ***");
    int fit_flag, iter_max, grp_sel;  
    double hwidth, model_resol;
    char *par_range = (char *)malloc(sizeof(char) * LEN_STRING);
    char *edit_file = (char *)malloc(sizeof(char) * LEN_STRING);
    char *add = (char *)malloc(sizeof(char) * LEN_STRING);
    espda_param_load(parlist, FIT_LINE_HWIDTH, &hwidth, 2.5e-3, 10.0);
    espda_param_load(parlist, FIT_LINE_PAR_RANGE, &par_range);
    espda_param_load(parlist, FIT_LINE_EDIT_FILE, &edit_file);
    espda_param_load(parlist, FIT_LINE_ADD, &add);
    espda_param_load(parlist, FIT_LINE_FIT_FLAG, &fit_flag);
    espda_param_load(parlist, FIT_LINE_ITER_MAX, &iter_max, 1, 1000);
    espda_param_load(parlist, FIT_LINE_GRP_SEL, &grp_sel, 0, GRP_MAX);
    espda_param_load(parlist, FIT_LINE_MODEL_RESOL, &model_resol, 1.0e3, 1.0e6);
    char *edit_instr = (char *)malloc(sizeof(char) * LEN_STRING); 
    FILE *file = fopen(edit_file, "r");
    if (file != NULL) {
        int null = fscanf(file, "%1000000s", edit_instr);
	if (null != 1) {
	    cpl_msg_warning(cpl_func, "Error in parsing the Voigt parameters.");
	}
        fclose(file);
    }
    else {
        strcpy(edit_instr, "");
    }

    // Load descriptors
    double z_em, syst_z;
    const char *syst_id = (char *)malloc(sizeof(char) * 4);
    espda_frame_desc(frameset, RSPEC_CONT, DAS_Z_EM, &z_em);
    if (syst_num != 0) {
        espda_frame_desc(frameset, SLINE_IDEN, "ESO QC SYSTID", &syst_id);
        espda_frame_desc(frameset, SLINE_IDEN, "ESO DAS SYSTZ", &syst_z);
    }

/*
 * Line list is prepared for Voigt profile fitting. The string parameter 
 * 'edit_instr' is converted into instructions to set up the line parameters by
 * function espda_voigt_prep().
 */

    // Parse parameter
    char *edit_repl = (char *)malloc(sizeof(char) * LEN_STRING);
    espda_dfs_ensure(espda_param_replace(edit_instr, ":", ",", edit_repl));
    cpl_free(edit_instr);
    cpl_array *edit_line = espda_param_extract(edit_repl, ",", 0, 3,
					       CPL_TYPE_INT);
    cpl_array *edit_par = espda_param_extract(edit_repl, ",", 1, 3,
					      CPL_TYPE_INT);
    cpl_array *edit_value = espda_param_extract(edit_repl, ",", 2, 3,
						CPL_TYPE_STRING);

    /*
    cpl_array_dump(edit_line, 0, 100, NULL);
    cpl_array_dump(edit_par, 0, 100, NULL);
    cpl_array_dump(edit_value, 0, 100, NULL);
    */
    
    // Create new line table based on instructions
    int j_max = 0;
    espda_util_safe(edit_line, j_max = cpl_array_get_size(edit_line));
    cpl_size edit_size = j_max;
    cpl_size sline_size = cpl_table_get_nrow(sline);
    cpl_size max_edit = 0;
    espda_util_safe(edit_line,
		    max_edit = (cpl_size)cpl_array_get_max(edit_line));
    cpl_table *sline2 = cpl_table_duplicate(sline);
    cpl_size sline_size2 = CPL_MAX(sline_size, max_edit);
    cpl_table_set_size(sline2, sline_size2);
    for (int i = 0; i < sline_size2; i++) {
        cpl_table_set_int(sline2, "ORDER", i, i + 1);
        if (i >= sline_size) {
            cpl_table_set_double(sline2, "CONT", i, 1);
        }
    }
    for (int j = 0; j < edit_size; j++) {
        int   row   = cpl_array_get_int(edit_line, j, NULL) - 1;
        int   col   = cpl_array_get_int(edit_par, j, NULL);
        const char *value = cpl_array_get_string(edit_value, j);
        char *null  = NULL;
        if (col == -1) {
            cpl_table_set_double(sline2, "PEAK", row, strtod(value, &null));
        }
        else if (col == 0) {
            cpl_table_set_string(sline2, "LINEID", row, value);
        }
        else if (col == 2) {
            cpl_table_set_double(sline2, "WAVEL", row, strtod(value, &null));
        }
    }

    // Prepare line table for fitting, using instructions to set up Voigt param.
    cpl_table *sline_pre2 = NULL;
    espda_voigt_prep(sline2, ion, hwidth, z_em, par_range, &sline_pre2);
    espda_struct_sort(sline_pre2, 1, "ORDER", FALSE);
    for (int j = 0; j < j_max; j++) {
        int row = cpl_array_get_int(edit_line, j, NULL) - 1;
        int col = cpl_array_get_int(edit_par,  j, NULL);
        char *col_name = malloc(sizeof(char) * LEN_STRING);
        cpl_type type = CPL_TYPE_DOUBLE;
	printf("%i %i %i\n", row, col, (int)cpl_error_get_code()); 
        switch (col) {
            case -2:
	        strcpy(col_name, "SYSTID");
                type = CPL_TYPE_STRING;
                break;
            case -1:
                strcpy(col_name, "PEAK");
                break;
            case 0:
                strcpy(col_name, "LINEID");
                type = CPL_TYPE_STRING;
                break;
            case 1:
                strcpy(col_name, "LINEZ");
                break;
            case 2:
                strcpy(col_name, "WAVEL");
                break;
            case 3:
                strcpy(col_name, "WAVELMIN");
                break;
            case 4:
                strcpy(col_name, "WAVELMAX");
                break;
            case 6:
                strcpy(col_name, "LINEZCON");
                type = CPL_TYPE_STRING;
                break;
            case 7:
                strcpy(col_name, "COLDEN");
                break;
            case 9:
                strcpy(col_name, "COLDENCON");
                type = CPL_TYPE_STRING;
                break;
            case 10:
                strcpy(col_name, "THERB");
                break;
            case 12:
                strcpy(col_name, "THERBCON");
                type = CPL_TYPE_STRING;
                break;
            case 13:
                strcpy(col_name, "TURBB");
                break;
            case 15:
                strcpy(col_name, "TURBBCON");
                type = CPL_TYPE_STRING;
                break;
            case 16:
                strcpy(col_name, "USE");
                type = CPL_TYPE_STRING;
                break;
            default:
                break;
        }
	printf("%i\n", (int)cpl_error_get_code()); 
	if (type == CPL_TYPE_DOUBLE) {        
            char *null;
            cpl_table_set_double(sline_pre2, col_name, row,
				 strtod(cpl_array_get_string(edit_value, j),
					&null));
        }
        else {
            cpl_table_set_string(sline_pre2, col_name, row,
				 cpl_array_get_string(edit_value, j));
        }
	cpl_free(col_name);
    }

    // Remove discarded lines from the table
    cpl_table *sline_ext2 = espda_struct_extract(sline_pre2, 1, "USE", 
						 CPL_NOT_EQUAL_TO, "N");

    cpl_table_dump(sline_ext2, 0, 100, NULL);

    if (cpl_table_get_nrow(sline_ext2) > 0) {
        espda_struct_sort(sline_pre2, 1, "WAVEL", FALSE);    
        cpl_table *sline_ext = NULL;
	espda_voigt_prep(sline_ext2, ion, hwidth, z_em, par_range, &sline_ext);

/*
 * Iteratively fit or apply Voigt profile to lines. Components are optionally 
 * added to the lines to improve the fit, depending on the number of iterations
 * set by the users (one more component per iteration).
 */
	
    	cpl_table *fspec_fit  = NULL;
    	cpl_table *rspec_fit  = NULL;
    	cpl_table *sline_out   = NULL;
    	cpl_array *rchisq_out = NULL;
    	cpl_array *iter_out   = NULL;
    	cpl_table *fspec_out  = NULL;
    	cpl_table *rspec_out  = NULL;
    	cpl_table *fspec_chisq = NULL;
    	double chisq_sline, rchisq_sline = 0.0;
    	if (fit_flag > 0) {
    	    cpl_msg_info(cpl_func, "*** Fit Voigt profiles to lines ***");
    	}
    	else {
    	    cpl_msg_info(cpl_func, "*** Apply Voigt profiles to lines ***");
    	}

    	cpl_table *sline_old = cpl_table_duplicate(sline_ext);
    	double rchisq_old = INFINITY;
    	int sline_res_num = 1;
    	int count = 0;
    	int count_max = 3;
    	cpl_msg_info(cpl_func, "Lines will be fitted in at most %i iterations.",
		     iter_max);

	// If needed, interpolate the instrument profile
    	if (rprof == NULL) {
    	    espda_print_verb(0,
			     cpl_msg_info(cpl_func, "Creating and interpolating"
					  " instrument profile..."));
    	    cpl_table *inst_prof = espda_model_inst(MODEL_SIZE, model_resol,
						    MODEL_KAPPA);
    	    cpl_ensure_code(espda_interp_inst(rspec_cont, inst_prof, MODEL_SIZE,
					      model_resol, MODEL_KAPPA, &rprof)
			    == CPL_ERROR_NONE, cpl_error_get_code());
    	    cpl_table_delete(inst_prof);
    	}

    	for (int iter_i = 1; iter_i <= iter_max && sline_res_num > 0 && 
    	    count < count_max; iter_i++) {
    	    if (iter_i == iter_max || sline_res_num == 0 || 
    	        count == count_max - 1) {
    	        cpl_msg_info(cpl_func, " Last iteration...");
    	        sline_ext = cpl_table_duplicate(sline_old);	    
    	    }
    	    else {
    	        cpl_msg_info(cpl_func, " Iteration %i...", iter_i);
    	    }

	    // Fit the lines
    	    int conv_flag = 1;
    	    espda_util_safe(sline_ext, espda_util_func(
		espda_wrap_line(fspec_cont, rspec_cont, sline_ext, ion,
				voigt_func, model_resol, 10, grp_sel,
				conv_flag, fprof, rprof, &fspec_fit, &rspec_fit,
				&sline_out, &rchisq_out, &iter_out),
		WRAP_LINE, 2));
    	    espda_struct_check(sline_out,
			       espda_struct_sort_old(sline_out, "WAVEL",
						     FALSE));
    	    espda_util_safe(fspec_cont,
			    fspec_out = cpl_table_duplicate(fspec_cont));
    	    rspec_out = cpl_table_duplicate(rspec_cont);

	    // Compute the absolute difference between fit profile and data
	    cpl_table *voigt = NULL;
    	    espda_struct_copy_old(rspec_fit, "FIT", "FLUX");
    	    voigt = cpl_table_duplicate(rspec_fit);
    	    espda_util_safe(fspec_out, espda_util_func(
		espda_fit_cspline(voigt,"FIT", fspec_out, NULL),
		FIT_CSPLINE, 0));  
    	    espda_util_func(espda_fit_cspline(voigt, "FIT", rspec_out, NULL), 
			    FIT_CSPLINE, 0);  
    	    if (fspec_out != NULL) {
    	        cpl_table_new_column(fspec_out, "DIFFMOD", CPL_TYPE_DOUBLE);
    	        for (int eqi = 0; eqi < cpl_table_get_nrow(fspec_out); eqi++) {
    	            cpl_table_set_double(fspec_out, "DIFFMOD", eqi, sqrt(pow(
    	                cpl_table_get_double(fspec_out, "FLUX", eqi, NULL) /
    	                cpl_table_get_double(fspec_out, "CONT", eqi, NULL) -
    	                cpl_table_get_double(fspec_out, "FIT", eqi, NULL), 2)));
    	        }
    	    }
    	    cpl_table_new_column(rspec_out, "DIFFMOD", CPL_TYPE_DOUBLE);
    	    for (int rebini = 0; rebini < cpl_table_get_nrow(rspec_out); 
    	        rebini++) {
    	        cpl_table_set_double(rspec_out, "DIFFMOD", rebini, sqrt(pow(
    	            cpl_table_get_double(rspec_out, "FLUX", rebini, NULL) /
    	            cpl_table_get_double(rspec_out, "CONT", rebini, NULL) -
    	            cpl_table_get_double(rspec_out, "FIT", rebini, NULL), 2)));
    	    }

	    // Compute the chi-squared
    	    if (sline_out != NULL) {
    	        for (int i = 0; i < cpl_table_get_nrow(sline_out); i++) {
    	            double wavel = cpl_table_get_double(sline_out, "WAVEL", i, 
							NULL);
    	            double hwidth = 0.5 * cpl_table_get_column_mean(rspec_out, 
								    "PIXSIZE");
    	            cpl_table *rspec_sline = NULL;
    	            while (rspec_sline == NULL) {
    	                espda_struct_window(rspec_out, "WAVEL", wavel - hwidth, 
					    wavel + hwidth, "IN", rspec_sline);
    	                hwidth = hwidth * 1.001;
    	            }
    	            cpl_table_set_double(sline_out, "PEAK", i, 
    	                cpl_table_get_double(rspec_sline, "FIT",  0, NULL) * 
    	                cpl_table_get_double(rspec_sline, "CONT", 0, NULL));
    	            cpl_table_set_double(sline_out, "CONT", i, 
    	                cpl_table_get_double(rspec_sline, "CONT", 0, NULL));
    	        }
    	    }
    	    espda_struct_norm(sline_out, "PEAK", NULL, "CONT", "NORM", NULL);
    	    fspec_chisq = cpl_table_duplicate(rspec_out);
    	    cpl_table_unselect_all(fspec_chisq);
    	    cpl_table_erase_invalid(fspec_chisq);
    	    int dof = cpl_table_get_nrow(fspec_chisq) - 
    	        cpl_table_get_nrow(sline_out) * 3;
    	    espda_model_chisq(fspec_chisq, "NORM", "NORMERR", "FIT",  dof, 
    	        chisq_sline, rchisq_sline);    
    	    if (fspec_out != NULL) {
    	        cpl_msg_info(cpl_func,
			     "Validation on the non-rebinned spectrum:");
    	        cpl_msg_info(cpl_func,
			     " Reduced chi-squared on line fitting:     "
			     "  %7.4f (size: %4i, degrees of freedom: %4i)",
			     rchisq_sline, (int)cpl_table_get_nrow(fspec_chisq),
			     dof);
    	    }

	    
    	    // If the chi-squared is getting worse, iteration is stopped
    	    int order;      
    	    order = (int)cpl_table_get_column_max(sline_out, "ORDER");
    	    if (rchisq_sline > rchisq_old) {
    	        count++;
    	    }
    	    else {
    	        count = 0;
    	        sline_old = cpl_table_duplicate(sline_out);
    	        rchisq_old = rchisq_sline;
    	    }

    	    // Add a component to improve the chi-squared
    	    if (iter_i < iter_max - 1) {
    	        cpl_msg_info(cpl_func, " Adding components to improve the "
    	            "chi-squared...");

    	        double accept = 3.0;
    	        double vel_step = 5.0e3;
    	        cpl_table *rspec_res = cpl_table_duplicate(rspec_out);
    	        cpl_table_unselect_all(rspec_res);
    	        cpl_table_erase_invalid(rspec_res);
    	        cpl_table_copy_data_double(rspec_res, "FLUX", 
    	            cpl_table_get_data_double(rspec_res, "NORM"));
    	        cpl_table_copy_data_double(rspec_res, "FLUXERR", 
    	            cpl_table_get_data_double(rspec_res, "NORMERR"));
    	        cpl_table_subtract_columns(rspec_res, "FLUX", "FIT");
    	        cpl_table_add_scalar(rspec_res, "FLUX", 1.0);
    	        cpl_table *sline_temp = NULL;
    	        espda_util_func(
		    sline_temp = espda_line_detect(rspec_res, rspec_res,
						   accept), 
    	            LINE_DETECT, 1);
    	        cpl_table *sline_res = NULL;
    	        espda_util_safe(sline_temp, sline_res = espda_struct_extract(
    	            sline_temp, 1, "PEAK", CPL_GREATER_THAN, 0.0));
    	        espda_util_safe(sline_res, espda_struct_size(sline_res, 
    	            sline_res_num, 0));
    	        espda_util_safe(sline_res, espda_util_func(espda_line_clean(
    	            sline_res, vel_step), LINE_CLEAN, 0));

    	        if (sline_res != NULL && cpl_table_get_nrow(sline_res) > 0) {
    	            espda_struct_sort(sline_res, 1, "PEAK", FALSE);
    	            cpl_table *sline_res_1st = cpl_table_extract(sline_res, 0,
								 1);
    	            cpl_size res_minpos;
    	            cpl_table_get_column_minpos(rspec_res, "FLUX", &res_minpos);
    	            cpl_table_set_double(
			sline_res_1st, "PEAK", 0, 
			cpl_table_get_double(rspec_res, "NORM",res_minpos,
					     NULL));
    	            cpl_table_set_double(
			sline_res_1st, "WAVEL", 0, 
			cpl_table_get_double(rspec_res, "WAVEL", res_minpos, 
					     NULL));                

    	            // Find to which transition the new component is associated
    	            double wavel_1st = cpl_table_get_double(sline_res_1st,
							    "WAVEL", 0, NULL);
    	            cpl_table *sline_match = cpl_table_duplicate(sline_out);
    	            cpl_table_subtract_scalar(sline_match, "WAVEL", wavel_1st);
    	            cpl_table_multiply_columns(sline_match, "WAVEL", "WAVEL");
    	            cpl_size row_match;
    	            cpl_table_get_column_minpos(sline_match, "WAVEL",
						&row_match);

    	            // Prepare the Voigt parameters for the new transition
    	            espda_struct_initcol(sline_res_1st, "LINEID",
					 CPL_TYPE_STRING);
    	            espda_struct_initcol(sline_res_1st, "MULTID",
					 CPL_TYPE_STRING);
    	            espda_struct_initcol(sline_res_1st, "SYSTID",
					 CPL_TYPE_STRING);
    	            cpl_table_set_string(
			sline_res_1st, "LINEID", 0, 
    	                cpl_table_get_string(sline_out, "LINEID", row_match));
    	            cpl_table_set_string(
			sline_res_1st, "MULTID", 0, 
    	                cpl_table_get_string(sline_out, "MULTID", row_match));
    	            cpl_table_set_string(
			sline_res_1st, "SYSTID", 0, 
    	                cpl_table_get_string(sline_out, "SYSTID", row_match));

    	            cpl_table *sline_res_pre = NULL;
    	            char *par_range_new = "1.0e-4,10.0,22.0,2.0,100.0,0.0,0.0";
    	            espda_voigt_prep(sline_res_1st, ion, hwidth, z_em, 
				     par_range_new, &sline_res_pre);
    	            espda_line_syst(sline_res_pre);
    	            cpl_table_set_int(sline_res_pre, "ORDER", 0, order + 1);

    	            // Group ID is set manually (espda_voigt_prep does not)
    	            cpl_table_set_int(
			sline_res_pre, "GRPID", 0, 
    	                cpl_table_get_int(sline_out, "GRPID", row_match, NULL));
    	            
    	            cpl_table_insert(sline_out, sline_res_pre, 
				     cpl_table_get_nrow(sline_out));
    	            cpl_table_fill_column_window_double(
		        sline_out, "COLDENERR", 0, 
    	                cpl_table_get_nrow(sline_out), INV_DOUBLE);
    	            sline_ext = cpl_table_duplicate(sline_out);
		}
    	        else{
    	            iter_max = iter_i;
    	        }

    	        espda_util_safe(sline_ext,
				espda_syst_adjust(sline_ext, ion, 1000, 5e-5));
    	        espda_struct_sort(sline_ext, 1, "ORDER", FALSE);
    	        espda_util_safe(sline_ext,
				espda_syst_constr(sline_ext, ion_red));
    	    }
    	    else {
    	        sline_ext = cpl_table_duplicate(sline_out);
    	    }
    		espda_struct_sort(rspec_out, 1, "PIXSIZE", FALSE);
    	}
    	cpl_table_delete(sline_ext);
    	cpl_array_delete(rchisq_out);
    	cpl_array_delete(iter_out);
    	cpl_table_delete(fspec_chisq);

	// Export products into frames
    	if (fspec_out != NULL) {
    	    cpl_table_duplicate_column(fspec_out, "FITRES", fspec_out, "FLUX");
    	    cpl_table_subtract_columns(fspec_out, "FITRES", "FIT");
    	    cpl_table_erase_column(fspec_out, "FITRES");
    	} 
    	cpl_table_duplicate_column(rspec_out, "FITRES", rspec_out, "FLUX");
    	cpl_table_subtract_columns(rspec_out, "FITRES", "FIT");
    	cpl_table_erase_column(rspec_out, "FITRES");

    	char *fspec_tag = NULL; 
    	char *rspec_tag = NULL; 
    	char *sline_tag  = NULL;
    	char *fline_tag  = FLINE_IDEN;
    	if (fit_flag == 0) {
    	    fspec_tag = FSPEC_GUESS;
    	    rspec_tag = RSPEC_GUESS;
    	    if (syst_num != 0) {
    	        sline_tag = SLINE_GUESS;
    	    }
    	    else {
    	        sline_tag = FLINE_GUESS;
    	    }
    	}
    	else {
    	    fspec_tag =  FSPEC_VOIGT;
    	    rspec_tag =  RSPEC_VOIGT;
    	    if (syst_num != 0) {
    	        sline_tag = SLINE_VOIGT;
    	    }
    	    else {
    	        sline_tag = FLINE_VOIGT;
    	    }
    	}
    	espda_struct_sort(sline_out, 2, "USE", TRUE, "WAVEL", FALSE);
    	cpl_msg_info(cpl_func, "*** Export products into frames ***");

    	// Create property lists
    	cpl_propertylist *list  = cpl_propertylist_new();
    	if (!isinf(rchisq_sline) && !isnan(rchisq_sline)) {
    	    cpl_propertylist_append_double(list,  QC_LINE_RCHISQ, rchisq_sline);
    	}
    	espda_struct_sort(sline_out, 1, "SYSTID", TRUE);
    	cpl_array *syst_z_arr = cpl_array_new(1, CPL_TYPE_DOUBLE);
	cpl_table_dump(sline_out, 0, 10, NULL);
    	espda_util_safe(sline_out, syst_z_arr = espda_syst_ave(sline_out, 1));
    	espda_struct_sort(sline_out, 1, "WAVEL", FALSE);
    	syst_z = cpl_array_get_double(syst_z_arr, 0, NULL);
    	cpl_array_delete(syst_z_arr);
    
    	cpl_propertylist_append_string(list, "ESO QC SYSTID", syst_id);
    	cpl_propertylist_append_double(list, "ESO DAS SYSTZ", syst_z);
    	cpl_propertylist *flist = cpl_propertylist_duplicate(list);
    	cpl_propertylist *rlist = cpl_propertylist_duplicate(list);
    	cpl_propertylist *fline_list = cpl_propertylist_duplicate(list);

    	// Create file names
    	espda_frame_name(fspec_name, fspec_tag);
    	espda_frame_name(rspec_name, rspec_tag);
    	espda_frame_name(sline_name, sline_tag);
    	espda_frame_name(fline_name, fline_tag);

    	// Save frames
    	espda_dfs_call(0,
	    espda_frame_export(frameset, parlist, fspec_out,
			       flist, "espda_fit_line", fspec_tag, fspec_name));
    	espda_dfs_call(0,
	    espda_frame_export(frameset, parlist, rspec_out, rlist,
			       "espda_fit_line", rspec_tag, rspec_name));
    	espda_dfs_call(0,
	    espda_frame_export(frameset, parlist, sline_out,  list, 
			       "espda_fit_line", sline_tag, sline_name));
    	espda_dfs_call(0,
	    espda_frame_export(frameset, parlist, fline, fline_list, 
			       "espda_fit_line", fline_tag, fline_name));
    }
    // If the table is empty
    else {  
        cpl_msg_warning(cpl_func, "System discarded! Terminating.");
    }

    cpl_table_delete(fspec_cont);
    cpl_table_delete(rspec_cont);
    cpl_table_delete(ion);
    cpl_table_delete(voigt_func);
    cpl_msg_info(cpl_func, "*** Success!! ***");

    return (int)cpl_error_get_code();

}

/**@}*/
