/* $Id: espda_iden_syst.c 288165 2024-03-28 09:07:49Z amodigli $
 *
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: amodigli $
 * $Date: 2024-03-28 09:07:49 +0000 (Thu, 28 Mar 2024) $
 * $Revision: 288165 $
 * $Name:  $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>

#include "espda_check.h"
#include "espda_dfs.h"
#include "espda_frame.h"
#include "espda_param.h"
#include "espda_redsh.h"
#include "espda_syst.h"
#include "espda_wrap.h"

/*-----------------------------------------------------------------------------
                             Plugin registration
 -----------------------------------------------------------------------------*/

int cpl_plugin_get_info(cpl_pluginlist * list);

/*-----------------------------------------------------------------------------
                            Private function prototypes
 -----------------------------------------------------------------------------*/

static int espda_iden_syst_create(cpl_plugin *);
static int espda_iden_syst_exec(cpl_plugin *);
static int espda_iden_syst_destroy(cpl_plugin *);
static int espda_iden_syst(cpl_frameset *, const cpl_parameterlist *);

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static char espda_iden_syst_description[] =
"This recipe identifies absorption systems by searching for redshift "
"coincidences among lines and selecting the candidate systems with a proper "
"set of rules.\n"
"\n"
"INPUT FRAMES                       FORMAT  # EXPECTED  TAGS\n"
"List of absorption lines           LINE    1           FLINE_REM\n"
"List of ionic transitions          ION     1           QSO_ION\n"
"Reduced list of ionic transitions  ION     0 or 1      QSO_ION_RED\n"
"\n"
"OUTPUT FRAMES                      FORMAT  # PRODUCED  TAGS          NOTES\n"
"Individual absorption systems      LINE    0 to N      SLINE_IDEN\n"
"List of identified lines           LINE    0 or 1      FLINE_IDEN\n"
"List of non-identified lines       LINE    0 or 1      FLINE_REJ\n"
"Reduced list of ionic transitions  ION     0 or 1      QSO_ION_RED   produced "
"only if QSO_ION_RED is not provided in input"
;

/*-----------------------------------------------------------------------------
                                Function code
 -----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module.
  @param    list    the plugin list
  @return   0 if everything is ok, 1 otherwise
  @note     Only this function is exported

  Create the recipe instance and make it available to the application using the
  interface.
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist *list)
{
    cpl_recipe *recipe = cpl_calloc(1, sizeof *recipe );
    cpl_plugin *plugin = &recipe->interface;

    if (cpl_plugin_init(plugin, CPL_PLUGIN_API, ESPDA_BINARY_VERSION,
                        CPL_PLUGIN_TYPE_RECIPE, "espda_iden_syst", "Identify "
                        "absorption system in a spectrum",
                        espda_iden_syst_description, "Guido Cupani",
                        PACKAGE_BUGREPORT, espda_dfs_license(),
                        espda_iden_syst_create, espda_iden_syst_exec,
                        espda_iden_syst_destroy)) {
        cpl_msg_error(cpl_func, "Plugin initialization failed");
        (void)cpl_error_set_where(cpl_func);
        return 1;
    }

    if (cpl_pluginlist_append(list, plugin)) {
        cpl_msg_error(cpl_func, "Error adding plugin to list");
        (void)cpl_error_set_where(cpl_func);
        return 1;
    }

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options
  @param    plugin  the plugin
  @return   0 if everything is ok

  Defining the command-line/configuration parameters for the recipe.
 */
/*----------------------------------------------------------------------------*/
static int espda_iden_syst_create(cpl_plugin *plugin)
{
    cpl_recipe    *recipe;
    cpl_parameter *param;

    /* Do not create the recipe if an error code is already set */
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());
        return (int)cpl_error_get_code();
    }

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    /* Create the parameters list in the cpl_recipe object */
    recipe->parameters = cpl_parameterlist_new();
    if (recipe->parameters == NULL) {
        cpl_msg_error(cpl_func, "Parameter list allocation failed");
        cpl_ensure_code(0, (int)CPL_ERROR_ILLEGAL_OUTPUT);
    }

    /* Fill the parameters list */
    /*
    param = cpl_parameter_new_value("espda.espda_iden_syst.grp_width",
                CPL_TYPE_DOUBLE, "Width for redshift grouping [nm]: [1.0e-2, "
                "1.0]", "espda.espda_iden_syst", 2.0e-2);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "grp-width");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);
    */

    param = cpl_parameter_new_value("espda.espda_iden_syst.mult_dz",
				    CPL_TYPE_DOUBLE,
				    "Minimum redshift difference between two candidate components of a "
            "doublet for the identification to be accepted: [1.0e-5, 1.0e-3]",
				    "espda.espda_iden_syst",
				    5e-5);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "mult-dz");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_iden_syst.add_dv",
				    CPL_TYPE_DOUBLE,
				    "Minimum velocity difference between velocity components of the "
            "same ionic species (e.g. CIV) two decide they belong to different "
            "absorption systems.: [1.0e-5, 1.0]",
				    "espda.espda_iden_syst", 1.0e-3);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "add-dv");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_iden_syst.grp_num",
				    CPL_TYPE_INT,
				    "Minimum number of lines to define a group: [1, 10]",
				    "espda.espda_iden_syst",
				    2);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "grp-num");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_iden_syst.syst_dv",
				    CPL_TYPE_DOUBLE,
				    "Minimum velocity difference between different doublets/multiplets "
            "(e.g. SiIV and CIV, or MgII and FeII) to group them in the same "
            "system [km/s]: [0.0, 1000.0]",
				    "espda.espda_iden_syst",
				    0.0);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "syst-dv");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_iden_syst.peak_thres",
				    CPL_TYPE_DOUBLE,
				    "Peak threshold for starting the "
				    "identification (normalized to the "
				    "continuum): [0.0, 1.0]",
				    "espda.espda_iden_syst",
				    0.9);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "peak-thres");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_iden_syst.vis_rej",
				    CPL_TYPE_STRING,
				    "List of visually selected systems",
				    "espda.espda_iden_syst",
				    "");
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "vis-rej");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_iden_syst_exec(cpl_plugin *plugin)
{
    int             recipe_status;

    cpl_recipe     *recipe;
    cpl_errorstate  initial_errorstate = cpl_errorstate_get();

    /* Return immediately if an error code is already set */
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());
        return (int)cpl_error_get_code();
    }

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    /* Verify parameter and frame lists */
    if (recipe->parameters == NULL) {
        cpl_msg_error(cpl_func, "Recipe invoked with NULL parameter list");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }
    if (recipe->frames == NULL) {
        cpl_msg_error(cpl_func, "Recipe invoked with NULL frame set");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Invoke the recipe */
    recipe_status = espda_iden_syst(recipe->frames, recipe->parameters);

    /* Ensure DFS-compliance of the products */
    if (cpl_dfs_update_product_header(recipe->frames)) {
        if (!recipe_status) recipe_status = (int)cpl_error_get_code();
    }

    if (!cpl_errorstate_is_equal(initial_errorstate)) {
        /* Dump the error history since recipe execution start.
           At this point the recipe cannot recover from the error */
        cpl_errorstate_dump(initial_errorstate, CPL_FALSE, NULL);
    }

    return recipe_status;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_iden_syst_destroy(cpl_plugin *plugin)
{
    cpl_recipe * recipe;

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    cpl_parameterlist_delete(recipe->parameters);

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    frameset   the frames list
  @param    parlist    the parameters list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_iden_syst(cpl_frameset *frameset, const cpl_parameterlist
    *parlist) {
    extern int v_level;
    //extern char syst_id1, syst_id2;

    /* check for invalid input */
    if(espda_files_dont_exist(frameset) != CPL_ERROR_NONE) {
    	return CPL_ERROR_BAD_FILE_FORMAT;
    }

/*
 * Frames are loaded and checked.
 */


    cpl_msg_info(cpl_func, "*** Load frames ***");
    espda_dfs_call(0, espda_dfs_groups(frameset));
    cpl_recipeconfig *cfg = cpl_recipeconfig_new();
    char *cfg_tag = FLINE_REM;
    cpl_recipeconfig_set_tag  (cfg, cfg_tag, 1, 1);
    cpl_recipeconfig_set_input(cfg, cfg_tag, QSO_ION,     1, 1);
    cpl_recipeconfig_set_input(cfg, cfg_tag, QSO_ION_RED, 0, 1);
    espda_util_func(espda_wrap_check(frameset, cfg, cfg_tag), WRAP_CHECK, 1);
    cpl_table *fline_pre = espda_frame_load(frameset, cfg, cfg_tag, -1);
    cpl_table *ion       = espda_frame_load(frameset, cfg, cfg_tag,  0);
    cpl_table *ion_red   = espda_frame_load(frameset, cfg, cfg_tag,  1);
    cpl_recipeconfig_delete(cfg);
    espda_line_syst(fline_pre);

/*
 * Parameters are loaded and checked.
 */

    cpl_msg_info(cpl_func, "*** Load parameters ***");
    /*
    espda_param_double("grp-width",  grp_width,
                       cpl_parameterlist_get_first_const(parlist), 1.0e-2, 1.0);
    */
    espda_param_double("mult-dz", mult_dz,
                       cpl_parameterlist_get_first_const(parlist), 1.0e-5, 1.0e-3);
    espda_param_double("add-dv", add_dv,
                       cpl_parameterlist_get_next_const(parlist), 1.0e-5, 1.0);
    espda_param_int   ("grp-num",    grp_num,
                       cpl_parameterlist_get_next_const(parlist), 1,      10);
    espda_param_double("syst-dv", syst_dv,
                       cpl_parameterlist_get_next_const(parlist), 0.0, 1000.0);
    espda_param_double("peak-thres", peak_thres,
                       cpl_parameterlist_get_next_const (parlist), 0.0,    1.0);
    char *vis_rej = (char *)malloc(sizeof(char) * LEN_STRING);
    //char *vis_rej; //[LEN_STRING];
    espda_param_load(parlist, IDEN_SYST_VIS_REJ,   &vis_rej);


/*
 * Redshift is loaded from the list of lines.
 */

    double zem;
    espda_frame_desc(frameset, FLINE_REM, DAS_Z_EM, &zem);

/*
 *
 */

    cpl_array *rej_value = espda_param_extract(vis_rej, ",", 0, 1,
        CPL_TYPE_STRING);

/*
 *
 */

    cpl_table_fill_column_window_int(fline_pre, "GRPID", 0,
        cpl_table_get_nrow(fline_pre), INV_INT);
    double par_range_value[7];
    par_range_value[0] = 1.0e-4;
    par_range_value[1] = 10.0;
    par_range_value[2] = 22.0;
    par_range_value[3] = 2.0;
    par_range_value[4] = 100.0;
    par_range_value[5] = 0.0;
    par_range_value[6] = 0.0;
    espda_util_func(espda_line_group(fline_pre, ion, par_range_value),
        LINE_GROUP, 0);

/*
 *
 */

    //int ion_red_flag = 0;
    if (ion_red == NULL) {
    	cpl_table_unselect_all(ion);
    	cpl_table_or_selected_string(ion, "LINEID", CPL_EQUAL_TO, "NV_1238");
    	cpl_table_or_selected_string(ion, "LINEID", CPL_EQUAL_TO, "NV_1242");
    	cpl_table_or_selected_string(ion, "LINEID", CPL_EQUAL_TO, "SiII_1260");
    	cpl_table_or_selected_string(ion, "LINEID", CPL_EQUAL_TO, "OI_1302");
    	cpl_table_or_selected_string(ion, "LINEID", CPL_EQUAL_TO, "SiII_1304");
    	cpl_table_or_selected_string(ion, "LINEID", CPL_EQUAL_TO, "CII_1334");
    	cpl_table_or_selected_string(ion, "LINEID", CPL_EQUAL_TO, "SiIV_1393");
    	cpl_table_or_selected_string(ion, "LINEID", CPL_EQUAL_TO, "SiIV_1402");
    	cpl_table_or_selected_string(ion, "LINEID", CPL_EQUAL_TO, "SiII_1526");
    	cpl_table_or_selected_string(ion, "LINEID", CPL_EQUAL_TO, "CIV_1548");
    	cpl_table_or_selected_string(ion, "LINEID", CPL_EQUAL_TO, "CIV_1550");
    	cpl_table_or_selected_string(ion, "LINEID", CPL_EQUAL_TO, "FeII_2344");
    	cpl_table_or_selected_string(ion, "LINEID", CPL_EQUAL_TO, "FeII_2382");
    	cpl_table_or_selected_string(ion, "LINEID", CPL_EQUAL_TO, "FeII_2586");
    	cpl_table_or_selected_string(ion, "LINEID", CPL_EQUAL_TO, "FeII_2600");
    	cpl_table_or_selected_string(ion, "LINEID", CPL_EQUAL_TO, "MgII_2796");
    	cpl_table_or_selected_string(ion, "LINEID", CPL_EQUAL_TO, "MgII_2803");
    	cpl_table_or_selected_string(ion, "LINEID", CPL_EQUAL_TO, "CaII_3934");
    	cpl_table_or_selected_string(ion, "LINEID", CPL_EQUAL_TO, "CaII_3969");
    	//cpl_table_or_selected_string(ion, "LINEID", CPL_EQUAL_TO, "NaI_5891");
    	//cpl_table_or_selected_string(ion, "LINEID", CPL_EQUAL_TO, "NaI_5897");
    	ion_red = cpl_table_extract_selected(ion);
        //ion_red_flag = 1;
        cpl_table_new_column(ion_red, "LINEID_MULT", CPL_TYPE_STRING);
        cpl_table_set_string(ion_red, "LINEID_MULT",  0, "NV_1242");
        cpl_table_set_string(ion_red, "LINEID_MULT",  1, "NV_1238");
        cpl_table_set_string(ion_red, "LINEID_MULT",  2, "SiII_1304,SiII_1526");
        cpl_table_set_string(ion_red, "LINEID_MULT",  3, "");
        cpl_table_set_string(ion_red, "LINEID_MULT",  4, "SiII_1260,SiII_1526");
        cpl_table_set_string(ion_red, "LINEID_MULT",  5, "");
        cpl_table_set_string(ion_red, "LINEID_MULT",  6, "SiIV_1402");
        cpl_table_set_string(ion_red, "LINEID_MULT",  7, "SiIV_1393");
        cpl_table_set_string(ion_red, "LINEID_MULT",  8, "SiII_1260,SiII_1304");
        cpl_table_set_string(ion_red, "LINEID_MULT",  9, "CIV_1550");
        cpl_table_set_string(ion_red, "LINEID_MULT", 10, "CIV_1548");
        cpl_table_set_string(ion_red, "LINEID_MULT", 11,
            "FeII_2382,FeII_2586,FeII_2600");
        cpl_table_set_string(ion_red, "LINEID_MULT", 12,
            "FeII_2344,FeII_2586,FeII_2600");
        cpl_table_set_string(ion_red, "LINEID_MULT", 13,
            "FeII_2344,FeII_2382,FeII_2600");
        cpl_table_set_string(ion_red, "LINEID_MULT", 14,
            "FeII_2344,FeII_2382,FeII_2586");
        cpl_table_set_string(ion_red, "LINEID_MULT", 15, "MgII_2803");
        cpl_table_set_string(ion_red, "LINEID_MULT", 16, "MgII_2796");
        cpl_table_set_string(ion_red, "LINEID_MULT", 17, "CaII_3969");
        cpl_table_set_string(ion_red, "LINEID_MULT", 18, "CaII_3934");
        //cpl_table_set_string(ion_red, "LINEID_MULT", 19, "NaI_5897");
        //cpl_table_set_string(ion_red, "LINEID_MULT", 20, "NaI_5891");
        cpl_table_new_column(ion_red, "MULTID", CPL_TYPE_STRING);
        cpl_table_set_string(ion_red, "MULTID",  0, "NV");
        cpl_table_set_string(ion_red, "MULTID",  1, "NV");
        cpl_table_set_string(ion_red, "MULTID",  2, "SiII");
        cpl_table_set_string(ion_red, "MULTID",  3, "");
        cpl_table_set_string(ion_red, "MULTID",  4, "SiII");
        cpl_table_set_string(ion_red, "MULTID",  5, "");
        cpl_table_set_string(ion_red, "MULTID",  6, "SiIV");
        cpl_table_set_string(ion_red, "MULTID",  7, "SiIV");
        cpl_table_set_string(ion_red, "MULTID",  8, "SiII");
        cpl_table_set_string(ion_red, "MULTID",  9, "CIV");
        cpl_table_set_string(ion_red, "MULTID", 10, "CIV");
        cpl_table_set_string(ion_red, "MULTID", 11, "FeII");
        cpl_table_set_string(ion_red, "MULTID", 12, "FeII");
        cpl_table_set_string(ion_red, "MULTID", 13, "FeII");
        cpl_table_set_string(ion_red, "MULTID", 14, "FeII");
        cpl_table_set_string(ion_red, "MULTID", 15, "MgII");
        cpl_table_set_string(ion_red, "MULTID", 16, "MgII");
        cpl_table_set_string(ion_red, "MULTID", 17, "CaII");
        cpl_table_set_string(ion_red, "MULTID", 18, "CaII");
        //cpl_table_set_string(ion_red, "MULTID", 19, "NaI");
        //cpl_table_set_string(ion_red, "MULTID", 20, "NaI");
    }
    espda_struct_initcol(ion, "MULTID", CPL_TYPE_STRING);
    for (int i = 0; i < cpl_table_get_nrow(ion); i++) {
        for (int r = 0; r < cpl_table_get_nrow(ion_red); r++) {
            char *found = strstr(cpl_table_get_string(ion, "LINEID", i),
                cpl_table_get_string(ion_red, "MULTID", r));
            if (found != NULL) {
                cpl_table_set_string(ion, "MULTID", i,
                    cpl_table_get_string(ion_red, "MULTID", r));
            }
        }
    }

/*
 *
 */

    cpl_msg_info(cpl_func, "*** Identify absorption systems ***");
    cpl_table *redsh = NULL;

    // Only lines in the red part are considered
    cpl_table *fline_red = espda_struct_extract(fline_pre, 1, "WAVEL",
        CPL_GREATER_THAN, WAVEL_LY_A * (1 + zem));

    // Only lines with peak lower than peak_thres are considered
    cpl_table *fline_strong = espda_struct_extract(fline_red, 1, "NORM",
        CPL_LESS_THAN, peak_thres);
    espda_util_func(redsh = espda_redsh_list(fline_strong, ion_red, zem),
        REDSH_LIST, 0);

    cpl_table *redsh2 = cpl_table_duplicate(redsh);
    espda_struct_erase(redsh2, 1, "LINEZ", CPL_LESS_THAN, 1.1);
    cpl_table *fline_iden = NULL;
    cpl_size cand_num;
    espda_util_func(fline_iden = espda_redsh_group(redsh, mult_dz,
        add_dv, &cand_num), REDSH_GROUP, 1);
    cpl_table_delete(redsh);


/*
 *
 */

    cpl_table *fline_rej = NULL;
    int conf_num = 0;
    if (cand_num > 0) {
        cpl_msg_info(cpl_func, "*** Confirm absorption systems ***");
        espda_util_func(espda_syst_conf(fline_iden, fline_pre, grp_num,
            cand_num, &conf_num, &fline_rej), SYST_CONF, 2);
        if (conf_num > 0 && syst_dv > 0.0 && fline_rej != NULL) {
            cpl_msg_info(cpl_func, "*** Add lines to absorption systems ***");
            espda_util_func(espda_syst_add(fline_iden, fline_rej, ion,
                cand_num, syst_dv, zem), SYST_ADD, 2);
            cpl_table *fline_add = NULL;
            espda_util_func(espda_syst_conf(fline_iden, fline_pre, grp_num,
                cand_num, &conf_num, &fline_add), SYST_CONF, 2);
            espda_struct_check(fline_add, cpl_table_insert(fline_rej, fline_add,
                cpl_table_get_nrow(fline_rej)));
            cpl_table_delete(fline_add);
            espda_struct_check(fline_rej, espda_struct_sort_old(fline_rej, "SYSTID", FALSE));
        }
    }
    else {
        fline_rej = cpl_table_duplicate(fline_pre);
    }

    cpl_table_delete(fline_pre);
    cpl_array *syst_z = NULL;
    if (fline_iden != NULL) {
        if (cpl_table_get_nrow(fline_iden) > 0) {
            syst_z = espda_syst_ave(fline_iden, cand_num);
	}
    }

/*
 *
 */

    //cpl_table_save(fline_iden, NULL, NULL, "fline_iden.fits", CPL_IO_CREATE);

    cpl_msg_info(cpl_func, "*** Put constraints on absorption systems ***");
//    cpl_table_dump(fline_iden, 0, 10, NULL);
//    espda_util_safe(fline_iden, espda_syst_adjust(fline_iden, ion,
//						  conf_num, mult_dz));

    espda_util_safe(fline_iden, espda_syst_constr(fline_iden, ion_red));
    cpl_table_delete(ion);


/*
 * Products are exported into frames.
 */

    cpl_msg_info(cpl_func, "*** Export products into frames ***");

    // Loop over systems
    for (int n = 0; n < cand_num; n++) {
        char syst_id[4];
        cpl_table *sline = NULL;
        espda_util_safe(fline_iden, sline = espda_syst_blend(fline_iden,
            syst_id));
	char syst_id_ext[5];
	sprintf(syst_id_ext, "%s ", syst_id);
	espda_util_safe(sline, espda_syst_constr(sline, ion_red));

	//cpl_table_dump(fline_iden, 0, 100, NULL);
	if (cpl_table_get_nrow(sline) > 0) {

	    int rej_flag = 0;
	    if (rej_value != NULL) {
	        for (int i = 0; i < cpl_array_get_size(rej_value); i++) {
	            if (!strcmp(syst_id, cpl_array_get_string(rej_value, i)) ||
			!strcmp(syst_id_ext,
                            cpl_array_get_string(rej_value, i))) {
                        rej_flag = 1;
	       	    }
                }
	    }

	    if (rej_flag == 1) {
	        cpl_msg_warning(cpl_func, "System %s will not be saved "
                    "because it was visually rejected.", syst_id);
		cpl_table *rem = espda_struct_extract(fline_iden, 1, "SYSTID",
						      CPL_EQUAL_TO, syst_id);
		espda_struct_erase(fline_iden, 1, "SYSTID", CPL_EQUAL_TO,
				   syst_id);
		cpl_table_fill_column_window_string(rem, "SYSTID", 0,
						    cpl_table_get_nrow(rem),
						    INV_STRING);
		cpl_table_insert(fline_rej, rem, 0);
            }
	    else {

	    // Create propertylist
            	cpl_propertylist *sline_list   = cpl_propertylist_new();
            	cpl_propertylist_append_string(sline_list, "ESO QC SYSTID",
            	    syst_id);
            	cpl_propertylist_append_double(sline_list, "ESO DAS SYSTZ",
            	    cpl_array_get_double(syst_z, n, NULL));

            	// Create file name
            	char sline_root[LEN_STRING] = SLINE_IDEN;
            	strcat(sline_root, "_");
            	strcat(sline_root, syst_id);
            	espda_frame_name(sline_name, sline_root);

            	// Save frame
            	espda_dfs_call(0, espda_frame_export(frameset, parlist, sline,
                    sline_list, "espda_iden_syst", SLINE_IDEN, sline_name));
            }
       	}
    }
    cpl_array_delete(syst_z);
    espda_util_safe(fline_iden, cpl_table_unselect_all(fline_iden));
    espda_struct_sort(fline_rej, 1, "WAVEL", FALSE);

    // Create property lists
    cpl_propertylist *list = cpl_propertylist_new();
    cpl_propertylist *fiden_list = cpl_propertylist_duplicate(list);
    cpl_propertylist *frej_list  = cpl_propertylist_duplicate(list);
    cpl_propertylist *ion_list   = cpl_propertylist_duplicate(list);
    cpl_propertylist_delete(list);

    // Create file names
    espda_frame_name(fiden_name, FLINE_IDEN);
    espda_frame_name(frej_name,  FLINE_REJ);
    espda_frame_name(ion_name,   QSO_ION_RED);


    // Save frames
    espda_dfs_call(0, espda_frame_export(frameset, parlist, fline_iden,
        fiden_list, "espda_iden_syst", FLINE_IDEN,  fiden_name));
    espda_dfs_call(0, espda_frame_export(frameset, parlist, fline_rej,
        frej_list,  "espda_iden_syst", FLINE_REJ,   frej_name));
    espda_dfs_call(0, espda_frame_export(frameset, parlist, ion_red,
        ion_list,   "espda_iden_syst", QSO_ION_RED, ion_name));

    cpl_msg_info(cpl_func, "*** Success!! ***");

    return (int)cpl_error_get_code();
}

/**@}*/
