/* $Id: espda_rv_synth.c 202524 2019-01-22 14:15:10 jonaydas $
 *
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: jonaydas $
 * $Date: 2019-01-16 17:15:10 +0000 (wed, 16 jan 2019) $
 * $Revision: 202524 $
 * $Name:  $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#include <string.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>
#include "cpl_plot.h"

#include "espda_synth_ccf.h"

#include "espda_dfs.h"
#include "espda_frame.h"
#include "espda_spec.h"
#include "espda_param.h"

/*-----------------------------------------------------------------------------
                             Plugin registration
 -----------------------------------------------------------------------------*/

int cpl_plugin_get_info(cpl_pluginlist * list);

/*-----------------------------------------------------------------------------
                            Private function prototypes
 -----------------------------------------------------------------------------*/

static int espda_rv_synth_create(cpl_plugin *);
static int espda_rv_synth_exec(cpl_plugin *);
static int espda_rv_synth_destroy(cpl_plugin *);
static int espda_rv_synth(cpl_frameset *, const cpl_parameterlist *);

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static char espda_rv_synth_description[] = 
"This recipe cross-correlates the normalized, merged, rebinned observed spectrum \n"   
"with a synthetic spectrum interpolated from a grid of synthetic spectra. The \n"
"computed cross-correlation function (CCF) is saved in two new columns in the \n" 
"output CPL FITS Table. The cross-correlation is performed within the radial \n"  
"velocity (RV) window (i.e. +- RV window). The CCF is fitted with a gaussian \n" 
"function to compute the RV of the star, and thus to correct the wavelength array \n"
"of the observed spectrum for the computed RV value. The corrected wavelength \n"
"array is saved in hte output CPL FITS Table.\n"
"\n"
"INPUT FRAMES                          FORMAT  # EXPECTED  TAGS          "
"NOTES: Relevant Table Columns:\n" 
"normalized merged rebinned spectrum \n" 
"and interpolated synthetic spectrum   SPEC    1           SMERG_SPEC    "
"WAVEL, FLUX, FLUXERR, SYNTH\n"
"\n"
"OUTPUT FRAMES                         FORMAT  # PRODUCED  TAGS          "
"NOTES: Relevant Table Columns:\n"
"normalized merged rebinned spectrum                                     "
"\n"
"and interpolated synthetic spectrum   SPEC    1           RMERG_SPEC    "
"WAVEL, FLUX, FLUXERR, SYNTH, WAVEC \n"
"and interpolated synthetic spectrum   CCF     1           CCF_SPEC      "
"XCCF, YCCF, YFIT\n"
;

/*-----------------------------------------------------------------------------
                                Function code
 -----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module. 
  @param    list    the plugin list
  @return   0 if everything is ok, 1 otherwise
  @note     Only this function is exported

  Create the recipe instance and make it available to the application using the 
  interface. 
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist *list)
{
    cpl_recipe *recipe = cpl_calloc(1, sizeof *recipe );
    cpl_plugin *plugin = &recipe->interface;

    if (cpl_plugin_init(plugin, CPL_PLUGIN_API, ESPDA_BINARY_VERSION,
                        CPL_PLUGIN_TYPE_RECIPE, "espda_rv_synth", 
                        "Cross-correlate observed spectrum with " 
                        "synthetic spectrum", 
                        espda_rv_synth_description, 
                        "Jonay Gonzalez Hernandez",
                        PACKAGE_BUGREPORT, espda_dfs_license(),
                        espda_rv_synth_create, espda_rv_synth_exec,
                        espda_rv_synth_destroy)) {    
        cpl_msg_error(cpl_func, "Plugin initialization failed");
        (void)cpl_error_set_where(cpl_func);                          
        return 1;                                               
    }                                                    

    if (cpl_pluginlist_append(list, plugin)) {                 
        cpl_msg_error(cpl_func, "Error adding plugin to list");
        (void)cpl_error_set_where(cpl_func);                         
        return 1;                                              
    }                                                          
    
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options    
  @param    plugin  the plugin
  @return   0 if everything is ok

  Defining the command-line/configuration parameters for the recipe.
 */
/*----------------------------------------------------------------------------*/
static int espda_rv_synth_create(cpl_plugin *plugin)
{
    cpl_recipe    *recipe;                                               
    cpl_parameter *param;
                                                                     
    /* Do not create the recipe if an error code is already set */     
    if (cpl_error_get_code() != CPL_ERROR_NONE) {                      
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());      
        return (int)cpl_error_get_code();                              
    }                                                                  
                                                                       
    if (plugin == NULL) {                                              
        cpl_msg_error(cpl_func, "Null plugin");                        
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                 
    }                                                                  
                                                                       
    /* Verify plugin type */                                           
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {       
        cpl_msg_error(cpl_func, "Plugin is not a recipe");             
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);              
    }                                                                  
                                                                       
    /* Get the recipe */                                               
    recipe = (cpl_recipe *)plugin;                                     
                                                                       
    /* Create the parameters list in the cpl_recipe object */          
    recipe->parameters = cpl_parameterlist_new();                      
    if (recipe->parameters == NULL) {                                  
        cpl_msg_error(cpl_func, "Parameter list allocation failed");   
        cpl_ensure_code(0, (int)CPL_ERROR_ILLEGAL_OUTPUT);             
    }                                                                  

    /* Fill the parameters list */
    /* Parameters of espda_synth_ccf */
    /* Parameter ccf_cen */
    param = cpl_parameter_new_value("espda.espda_rv_synth.ccf_cen",
                                    CPL_TYPE_DOUBLE, "Guess radial velocity [km/s]: "
                                    "[-850., +850.] ",
                                    "espda.espda_rv_synth", 0.);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "ccf-cen");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    /* Parameter ccf_win */
    param = cpl_parameter_new_value("espda.espda_rv_synth.ccf_win",
                                    CPL_TYPE_DOUBLE, "Radial velocity window [km/s]: "
                                    "[10., +1000.] ",
                                    "espda.espda_rv_synth", 350.);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "ccf-win");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    /* Parameter ccf_fitg */
    param = cpl_parameter_new_value("espda.espda_rv_synth.ccf_fitg",
                CPL_TYPE_BOOL, "Flag to use a gaussian function to fit the CCF "
                		"instead of a lorentzian function: "
		"TRUE, FALSE", "espda_rv_synth", "FALSE");
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "ccf-fitg");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    /* Parameter ccf_fitp */
    param = cpl_parameter_new_value("espda.espda_rv_synth.ccf_frv",
                CPL_TYPE_BOOL, "Flag to use RV derived by fitting the CCF to "
                		"correct the wavelengths instead of RV derived by "
                		"fitting the CCF core: "
		"TRUE, FALSE", "espda_rv_synth", "FALSE");
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "ccf-frv");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    /* Parameter syn_fct */
    param = cpl_parameter_new_value("espda.espda_rv_synth.syn_fct",
                                    CPL_TYPE_DOUBLE, "Synthesis multiplication factor: "
                                    "[0.01, +10.] ",
                                    "espda.espda_rv_synth", 1.);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "syn-fct");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_rv_synth_exec(cpl_plugin *plugin)
{
    int             recipe_status;

    cpl_recipe     *recipe;                                                   
    cpl_errorstate  initial_errorstate = cpl_errorstate_get();              
                                                                           
    /* Return immediately if an error code is already set */               
    if (cpl_error_get_code() != CPL_ERROR_NONE) {                          
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",    
                      cpl_func, __LINE__, cpl_error_get_where());          
        return (int)cpl_error_get_code();                                  
    }                                                                      
                                                                           
    if (plugin == NULL) {                                                  
        cpl_msg_error(cpl_func, "Null plugin");                            
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                     
    }                                                                      
                                                                           
    /* Verify plugin type */                                               
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {           
        cpl_msg_error(cpl_func, "Plugin is not a recipe");                 
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);                  
    }                                                                      
                                                                           
    /* Get the recipe */                                                   
    recipe = (cpl_recipe *)plugin;                                         
                                                                           
    /* Verify parameter and frame lists */                                 
    if (recipe->parameters == NULL) {                                      
        cpl_msg_error(cpl_func, "Recipe invoked with NULL parameter list");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                     
    }                                                                      
    if (recipe->frames == NULL) {                                          
        cpl_msg_error(cpl_func, "Recipe invoked with NULL frame set");     
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                     
    }                                                                      
                                                                           
    /* Invoke the recipe */                                                
    recipe_status = espda_rv_synth(recipe->frames, recipe->parameters);
                                                                           
    /* Ensure DFS-compliance of the products */                            
    if (cpl_dfs_update_product_header(recipe->frames)) {                   
        if (!recipe_status) recipe_status = (int)cpl_error_get_code();                         
    }                                                                      
                                                                           
    if (!cpl_errorstate_is_equal(initial_errorstate)) {                    
        /* Dump the error history since recipe execution start.            
           At this point the recipe cannot recover from the error */       
        cpl_errorstate_dump(initial_errorstate, CPL_FALSE, NULL);          
    }                                                                      
                                                                           
    return recipe_status;                                                  
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_rv_synth_destroy(cpl_plugin *plugin)
{
    cpl_recipe * recipe;                                          
                                                                  
    if (plugin == NULL) {                                         
        cpl_msg_error(cpl_func, "Null plugin");                   
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);            
    }                                                             
                                                                  
    /* Verify plugin type */                                      
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {  
        cpl_msg_error(cpl_func, "Plugin is not a recipe");        
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);         
    }                                                             
                                                                  
    /* Get the recipe */                                          
    recipe = (cpl_recipe *)plugin;                                
                                                                  
    cpl_parameterlist_delete(recipe->parameters);             
                                                                  
    return 0;                                                    
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    frameset   the frames list
  @param    parlist    the parameters list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_rv_synth(cpl_frameset            *frameset,
			              const cpl_parameterlist *paramlist)
{
    cpl_propertylist *proplist   = NULL;
    cpl_table        *smerg_tab   = NULL;
    /* check for invalid input */
    if(espda_files_dont_exist(frameset) != CPL_ERROR_NONE) {
    	return CPL_ERROR_BAD_FILE_FORMAT;
    }
    cpl_msg_info(cpl_func, "*** Starting RV Synth ***");

/*
 * Getting parameters from the list into variables 
 */

//    double rvccf    = cpl_parameter_get_double(
//                            cpl_parameterlist_find_const(paramlist,
//			   "espda.espda_rv_synth.ccf_cen"));
    espda_param_double("ccf-cen",         rvccf,
        cpl_parameterlist_get_first_const(paramlist), -850.,  850.);

//    double vwccf     = cpl_parameter_get_double(
//                            cpl_parameterlist_find_const(paramlist,
//			   "espda.espda_rv_synth.ccf_win"));
    espda_param_double("ccf-win",         vwccf,
        cpl_parameterlist_get_next_const (paramlist), 10.,  1000.);

    espda_param_bool  ("ccf-fitg",   ccf_fitg,
       cpl_parameterlist_get_next_const(paramlist));

    espda_param_bool  ("ccf-frv",   ccf_frv,
        cpl_parameterlist_get_next_const(paramlist));

    //int ccf_fitg, ccf_frv ;
    //espda_param_load(paramlist, RV_SYNTH_CCF_FITG, &ccf_fitg);
    //espda_param_load(paramlist, RV_SYNTH_CCF_FRV, &ccf_frv);
    //needs to define RV_SYNTH_CCF_* in espda_param.h

    espda_param_double("syn-fct",         sfct,
        cpl_parameterlist_get_next_const (paramlist), 0.001,  1000.);


/* 
 * Spectra are imported from frame
 */

    cpl_msg_info(cpl_func, "*** Load input FITS frames into tables ***");
    cpl_ensure_code(espda_frame_import(frameset, SMERG_SPEC, &smerg_tab) 
                    == CPL_ERROR_NONE, cpl_error_get_code());
    if (smerg_tab == NULL) {
        return (int)cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND, 
               "No input merge spectrum frame! Terminating.");
    }
    cpl_msg_info(cpl_func, "*** 1D NORMALIZED MERGED SPEC - SMERG_SPEC table LOADED***");

/*
 * DFS groups are set
 */

    cpl_msg_info(cpl_func, "*** Set groups of FITS frames ***");
    cpl_ensure_code(espda_dfs_groups(frameset) == CPL_ERROR_NONE,
                    cpl_error_get_code());

    /* extracting wave, flux and mask of the normalized merged spectrum */

    const double *wobsp = cpl_table_get_data_double_const (smerg_tab, "WAVEL");
    cpl_array *wobs_arr = cpl_array_new(cpl_table_get_nrow(smerg_tab), CPL_TYPE_DOUBLE);
    cpl_array_copy_data_double (wobs_arr, wobsp);
 
    const double *fobsp = cpl_table_get_data_double_const (smerg_tab, "FLUX");
    cpl_array *fobs_arr = cpl_array_new(cpl_table_get_nrow(smerg_tab), CPL_TYPE_DOUBLE);
    cpl_array_copy_data_double (fobs_arr, fobsp);
   
    const int *fmskp = cpl_table_get_data_int_const (smerg_tab, "MASK");
    cpl_array *fmsk_arr = cpl_array_new(cpl_table_get_nrow(smerg_tab), CPL_TYPE_INT);
    cpl_array_copy_data_int (fmsk_arr, fmskp);

    /* Creating mask */

    cpl_size npoint = cpl_array_get_size (wobs_arr);
    int cfmsk = 1;
    for (int j = 0; j < (int)npoint; j++) {
    	double cwav =  cpl_array_get_double (wobs_arr, j, NULL);
    	if (cwav > 675.){
    		cfmsk = 0;
    	} else {
    	    cfmsk = 1;
    	}
    	cpl_array_set_int (fmsk_arr, j, cfmsk);
    }
   
    /* Extracting flux of the interpolated synthetic spectrum */

    const double *fsynp = cpl_table_get_data_double_const (smerg_tab, "SYNTH");
    cpl_array *fsyn_arr = cpl_array_new(cpl_table_get_nrow(smerg_tab), CPL_TYPE_DOUBLE);
    cpl_array_copy_data_double (fsyn_arr, fsynp);

    /* Defining new wavelength of the observed spectrum */
 
    const double *wcorp = cpl_table_get_data_double_const (smerg_tab, "WAVEL");
    cpl_array *wcor_arr = cpl_array_new(cpl_table_get_nrow(smerg_tab), CPL_TYPE_DOUBLE);
    cpl_array_copy_data_double (wcor_arr, wcorp);

    /* Defining new array of the residuals observed - synthetic spectra */

    cpl_array *fres_arr = cpl_array_duplicate (fobs_arr);

    // We multiply the synthetic spectrum by the factor sfct

    cpl_array_multiply_scalar (fsyn_arr, sfct);

    /* Defining new array of the new synthetic spectrum multiplied by fct */

    cpl_array *fsyc_arr = cpl_array_duplicate (fsyn_arr);

    //cpl_array_subtract (fres_arr, fsyn_arr);

    /* Defining new arrays of the CCF */
    cpl_array *xccf_arr=NULL;
    cpl_array *yccf_arr=NULL;
    cpl_array *ygfit_arr=NULL;
    cpl_array *ygfcore_arr=NULL;

    /* Defining the remaining variables */

    double rvn = -9999. ;
    double rvn_err = -9999. ;
    int npar = 6 ;
    cpl_array *pccf_arr = cpl_array_new(npar, CPL_TYPE_DOUBLE);
    cpl_array *eccf_arr = cpl_array_new(npar, CPL_TYPE_DOUBLE);
    int npr = 3 ;
    cpl_array *pccfc_arr = cpl_array_new(npr, CPL_TYPE_DOUBLE);
    cpl_array *eccfc_arr = cpl_array_new(npr, CPL_TYPE_DOUBLE);

    /* Calling the function to cross-correlate observed and synthetic spectra */

    printf("Cross-correlating observed and synthetic spectra \n");

    printf("CCF fit flag %i %i \n", ccf_fitg, ccf_frv);

//    cpl_ensure_code(espda_ccf_fit_synth(wobs_arr, fobs_arr, fsyn_arr,
      cpl_ensure_code(espda_ccf_fit_synth(wobs_arr, fobs_arr, fsyc_arr,
    		  fmsk_arr, wcor_arr, fres_arr, &xccf_arr, &yccf_arr,
			  pccf_arr, eccf_arr, pccfc_arr, eccfc_arr,
			  &ygfit_arr, &ygfcore_arr, rvccf, vwccf, ccf_fitg,
			  ccf_frv) == CPL_ERROR_NONE, cpl_error_get_code());

    //int nxp = cpl_array_get_size(wcor_arr);
    //int nrp = cpl_array_get_size(fres_arr);
    //printf(" npt nrp : %i %i \n", nxp, nrp);

    cpl_array_delete(fobs_arr);
    cpl_array_delete(wobs_arr);
    cpl_array_delete(fsyn_arr);

    // We create CCF table
    cpl_size ccf_npt = cpl_array_get_size(xccf_arr);
    //int ccf_yp = cpl_array_get_size(yccf_arr);
    //int ccf_fp = cpl_array_get_size(ygfit_arr);
    //printf("ccf points %i %i %i \n", (int)ccf_npt, ccf_yp, ccf_fp);

    cpl_table *ccf_tab = cpl_table_new(ccf_npt);
    cpl_table_new_column(ccf_tab,"XCCF", CPL_TYPE_DOUBLE);
    cpl_table_new_column(ccf_tab,"YCCF", CPL_TYPE_DOUBLE);
    cpl_table_new_column(ccf_tab,"YFIT", CPL_TYPE_DOUBLE);
    cpl_table_new_column(ccf_tab,"YCORE", CPL_TYPE_DOUBLE);

    const double *xccfp = cpl_array_get_data_double_const (xccf_arr);
    const double *yccfp = cpl_array_get_data_double_const (yccf_arr);
    const double *ygfitp = cpl_array_get_data_double_const (ygfit_arr);
    const double *ygfcorep = cpl_array_get_data_double_const (ygfcore_arr);
    cpl_table_copy_data_double (ccf_tab, "XCCF", xccfp);
    cpl_table_copy_data_double (ccf_tab, "YCCF", yccfp);
    cpl_table_copy_data_double (ccf_tab, "YFIT", ygfitp);
    cpl_table_copy_data_double (ccf_tab, "YCORE", ygfcorep);

    cpl_propertylist* plist = cpl_propertylist_new();
    if (ccf_tab != NULL) {
    	cpl_ensure_code(espda_frame_export(frameset, paramlist, ccf_tab,
                         plist, "espda_rv_synth", CCF_SPEC,
                         "CCF_SPEC.fits")  == CPL_ERROR_NONE,
                         cpl_error_get_code());
    }
    else {
    	cpl_msg_warning(cpl_func, "No RV CCF frame! No frame exported.");
    }

    cpl_array_delete(xccf_arr);
    cpl_array_delete(yccf_arr);
    cpl_array_delete(ygfit_arr);
    cpl_array_delete(ygfcore_arr);

    /* Printing on screen the derived RV from fitting of the CCF */
    double rvnew = cpl_array_get_double (pccf_arr, 1, NULL);
    double rvnew_err = cpl_array_get_double (eccf_arr, 1, NULL);
    printf(" RV CCF [km/s] : %f \n", rvnew);
    printf(" Error RV CCF [km/s] : %f \n", rvnew_err);

    double rvnewc = cpl_array_get_double (pccfc_arr, 1, NULL);
    double rvnewc_err = cpl_array_get_double (eccfc_arr, 1, NULL);
    printf(" RV CCF Core[km/s] : %f \n", rvnewc);
    printf(" Error RV CCF Core [km/s] : %f \n", rvnewc_err);

    cpl_array_delete(eccf_arr);
    cpl_array_delete(pccf_arr);

    cpl_array_delete(eccfc_arr);
    cpl_array_delete(pccfc_arr);

    // We duplicate the table
    cpl_table *rmerg_tab = cpl_table_duplicate(smerg_tab);

    // Copy the output RV-corrected wavelength data into a Table DOUBLE column "WAVEC"
    cpl_table_new_column(rmerg_tab, "WAVEC", CPL_TYPE_DOUBLE);
    cpl_table_new_column(rmerg_tab, "SYNTC", CPL_TYPE_DOUBLE);
    cpl_table_new_column(rmerg_tab, "RESIC", CPL_TYPE_DOUBLE);

    const double *wcor = cpl_array_get_data_double_const (wcor_arr);
    cpl_table_copy_data_double (rmerg_tab, "WAVEC", wcor);
    const double *fres = cpl_array_get_data_double_const (fres_arr);
    cpl_table_copy_data_double (rmerg_tab, "RESIC", fres);
    const double *fsyc = cpl_array_get_data_double_const (fsyc_arr);
    cpl_table_copy_data_double (rmerg_tab, "SYNTC", fsyc);
    const int *fmsk = cpl_array_get_data_int_const (fmsk_arr);
    cpl_table_copy_data_int (rmerg_tab, "MASK", fmsk);

    cpl_array_delete(wcor_arr);
    cpl_array_delete(fres_arr);
    cpl_array_delete(fsyc_arr);
    cpl_array_delete(fmsk_arr);

/*
 * Products are exported into frames. 
 */ 

    if (ccf_frv == 1) {
    	rvn = rvnew ;
    	rvn_err = rvnew_err ;
    } else {
    	rvn = rvnewc ;
    	rvn_err = rvnewc_err ;
    }

    cpl_msg_info(cpl_func, "*** Export products into frames ***");

    cpl_table_and_selected_double (rmerg_tab, "WAVEL", CPL_LESS_THAN, 680.);
    cpl_table *rmerger_tab = cpl_table_extract_selected (rmerg_tab);
    cpl_table_unselect_all (rmerg_tab);

    proplist = cpl_propertylist_new();
 //   cpl_propertylist_append_double(proplist, "ESO DAS QC RVSYNTH SYNFCT", sfct);
    cpl_propertylist_append_double(proplist, "ESO DAS QC RVSYNTH RVCCF", rvnew);
    cpl_propertylist_append_double(proplist, "ESO DAS QC RVSYNTH RVECCF", rvnew_err);
    cpl_propertylist_append_double(proplist, "ESO DAS QC RVSYNTH RVCORE", rvnewc);
    cpl_propertylist_append_double(proplist, "ESO DAS QC RVSYNTH RVECORE", rvnewc_err);
    cpl_propertylist_append_double(proplist, "ESO DAS QC RVSYNTH RVFINAL", rvn);
    cpl_propertylist_append_double(proplist, "ESO DAS QC RVSYNTH RVEFINAL", rvn_err);
    double mmflx= cpl_table_get_column_mean(rmerger_tab, "FLUX") ;
    cpl_propertylist_append_double(proplist, "ESO DAS QC RVSYNTH MEAN MFLUX W680", mmflx);
    double msflx= cpl_table_get_column_mean(rmerger_tab, "SYNTC") ;
    cpl_propertylist_append_double(proplist, "ESO DAS QC RVSYNTH MEAN SYNTC W680", msflx);
    double mrflx= cpl_table_get_column_mean(rmerger_tab, "RESIC") ;
    cpl_propertylist_append_double(proplist, "ESO DAS QC RVSYNTH MEAN RESIC W680", mrflx);
    double mrstd= cpl_table_get_column_stdev(rmerger_tab, "RESIC") ;
    cpl_propertylist_append_double(proplist, "ESO DAS QC RVSYNTH STDEV RESIC W680", mrstd);
    if (rmerg_tab != NULL) {
     cpl_ensure_code(espda_frame_export(frameset, paramlist, rmerg_tab, 
                        proplist, "espda_rv_synth", RMERG_SPEC,
                        "RMERG_SPEC.fits")  == CPL_ERROR_NONE,
                        cpl_error_get_code());
    }
    else {
     cpl_msg_warning(cpl_func, "No RV CCF frame! No frame exported.");
    }

/*    
*Finishing the recipe
*/


    //We delete all the property lists to release and clean the memory
    //cpl_propertylist_delete(proplist);

    //We delete all the tables to release and clean the memory
    cpl_table_delete(rmerger_tab);
    cpl_table_delete(smerg_tab);

    cpl_msg_info(cpl_func, "*** Successful Recipe Execution!! ***");

    return (int)cpl_error_get_code();

}

/**@}*/
/*--------------------------------------------------------------------------*/
